package lib.comm;

import java.awt.Container;
import java.awt.Toolkit;
import java.awt.MediaTracker;
import java.awt.image.BufferedImage;
import java.awt.Image;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.net.MalformedURLException;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.List;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.UnsupportedEncodingException;

import lib.cfg.Constants;
import lib.exceptions.MapRequestException;

/**
 * @class lib.applets
 * 
 * Creates a module for communicating with the servlets using HTTP.
 * There are two types of requests that are implemented:
 * 	- Actions over images that return images
 * 	- Requests that give information about the current map
 * 
 * @author Tihomir Lazarov
 * @date   Jun 5, 2005
 */
public class MapRequest {
	Hashtable m_URLs		= null;
	ArrayList m_Cookies	= new ArrayList( ); // Stores all cookie strings
	
	private static final String DEFAULT_ENCODING	= "UTF-8";
	
	// URL types
	private static final int MAX_URL_TYPES			= 10;
	
	public static final int URL_TYPE_INIT			= 1;
	public static final int URL_TYPE_SHOW_MAP		= 2;
	public static final int URL_TYPE_ADD_AREA		= 3;
	public static final int URL_TYPE_UPDATE_AREA	= 4;
	public static final int URL_TYPE_REMOVE_AREA	= 5;
	public static final int URL_TYPE_INFO			= 6;
	public static final int URL_TYPE_SAVE_MAP		= 7;
	public static final int URL_TYPE_SEARCH_MAP	= 8;
	public static final int URL_TYPE_LOGIN			= 9;
	public static final int URL_TYPE_LOGOUT		= 10;
	
	/**
	 * Constructor of the class
	 * 
	 * @param urls - a hashtable of the different URL types
	 * 
	 * @throws MapRequestException
	 */
	public MapRequest( Hashtable urls )
		throws MapRequestException
	{
		m_URLs = urls;
		
		int iURLTypesCount = 0;
		
		// Check if all needed URLs are here
		Enumeration urlTypes = urls.keys( );
		while ( urlTypes.hasMoreElements( ) ) {
			int iURLType = ( ( Integer ) urlTypes.nextElement( ) ).intValue( );
			switch ( iURLType ) {
				case URL_TYPE_INIT:
				case URL_TYPE_SHOW_MAP:
				case URL_TYPE_ADD_AREA:
				case URL_TYPE_UPDATE_AREA:
				case URL_TYPE_REMOVE_AREA:
				case URL_TYPE_INFO:
				case URL_TYPE_SAVE_MAP:
				case URL_TYPE_SEARCH_MAP:
				case URL_TYPE_LOGIN:
				case URL_TYPE_LOGOUT:
					iURLTypesCount++;
					break;
				default:
					throw new MapRequestException( "Invalid URL type passed: " + iURLType, null );
			}
		}
		
		if ( iURLTypesCount != MAX_URL_TYPES ) {
			throw new MapRequestException( "Insuffucient URL types passed: " + iURLTypesCount, null );
		}
	}
	
	/**
	 * Does a request to view the current map
	 * 
	 * @param params - key-value hashtable with CGI query parameters
	 * 
	 * @return an image of type BufferedImage
	 * 
	 * @throws MapRequestException
	 */
	public BufferedImage searchMapRequest( Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String )m_URLs.get( new Integer( URL_TYPE_SEARCH_MAP ) );
		if ( params != null ) {
				szURL += '?' + buildCGIParams( params );
			}
		return mapRequest( szURL );
	}
	
	/**
	 * Makes a zoom request
	 * 
	 * @param szZoomEvent - Constants.EVENT_ON_ZOOM* constants
	 * @param params - CGI parameters
	 * 
	 * @return the result image 
	 * 
	 * @throws MapRequestException
	 */
	public BufferedImage zoomMapRequest( String szZoomEvent, Hashtable params )
		throws MapRequestException
	{
		String szURL 	= ( String )m_URLs.get( new Integer( URL_TYPE_SHOW_MAP ) );
		szURL += "?" + Constants.PARAM_EVENT + "=" + szZoomEvent;
		if ( params != null ) { szURL += '&' + buildCGIParams( params ); }
		return mapRequest( szURL );
	}
	
	/**
	 * Makes a move request
	 * 
	 * @param szMoveEvent - Constants.EVENT_ON_MOVE_* constants
	 * @param params - CGI parameters
	 * 
	 * @return the result image 
	 * 
	 * @throws MapRequestException
	 */
	public BufferedImage moveMapRequest( String szMoveEvent, Hashtable params )
		throws MapRequestException
	{
		String szURL 	= ( String )m_URLs.get( new Integer( URL_TYPE_SHOW_MAP ) );
		szURL += "?" + Constants.PARAM_EVENT + "=" + szMoveEvent;
		if ( params != null ) { szURL += '&' + buildCGIParams( params ); }
		return mapRequest( szURL );
	}
	
	/**
	 * Makes a zoom (in/out) + move request
	 * 
	 * @param szEvent - Constants.EVENT_ON_ZOOM_IN/OUT_MOVE* constants
	 * 
	 * @return the result image 
	 * 
	 * @throws MapRequestException
	 */
	public BufferedImage combinedMapRequest( String szEvent, Hashtable params )
		throws MapRequestException
	{
		String szURL 	= ( String )m_URLs.get( new Integer( URL_TYPE_SHOW_MAP ) );
		szURL += "?" + Constants.PARAM_EVENT + "=" + szEvent;
		if ( params != null ) { szURL += '&' + buildCGIParams( params ); }
		return mapRequest( szURL );
	}	
	
	/**
	 * Initializes a map by requesting it for the first time
	 * 
	 * @return the returned map image
	 * 
	 * @throws MapRequestException
	 */
	public BufferedImage initMapRequest( int iViewerWidth, int iViewerHeight, int iMapID )
		throws MapRequestException
	{
		String szURL = ( String )m_URLs.get( new Integer( URL_TYPE_INIT ) );
		szURL += '?' + Constants.PARAM_VIEWER_WIDTH  + "=" + iViewerWidth + "&" + 
		               Constants.PARAM_VIEWER_HEIGHT + "=" + iViewerHeight + "&" + 
		               Constants.PARAM_MAP_ID + "=" + iMapID + "&" +  
		               Constants.PARAM_RESET_MAP + "=1";
		return mapRequest( szURL );
	}
	
	/**
	 * Makes a request that returns misc information about the map and the 
	 * environment
	 * 
	 * @param iInfoType - the type of information to get
	 * @param params - key-value CGI parameters. Can be null
	 * 
	 * @return an object of type MapRequestInfo
	 * 
	 * @throws MapRequestException
	 */
	public MapRequestInfo infoMapRequest( int iInfoType, Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String )m_URLs.get( new Integer( URL_TYPE_INFO ) );
		szURL += '?' + Constants.PARAM_INFO_TYPE + '=' + iInfoType;
		if ( params != null ) {
			szURL += '&' + buildCGIParams( params );
		}
		return infoRequest( szURL );
	}
	
	/**
	 * Saves the map in the database
	 * 
	 * @param params - key-value CGI parameters. Can be null
	 * 
	 * @return an object of type BufferedImage
	 * 
	 * @throws MapRequestException
	 */
	public BufferedImage saveMap( Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String )m_URLs.get( new Integer( URL_TYPE_SAVE_MAP ) );
		if ( params != null ) {
			szURL += '?' + buildCGIParams( params );
		}
		return mapRequest( szURL );
	}	
	
	/**
	 * Adds an area to a map
	 * 
	 * @param params - the parameters that must be passed to the 
	 *                 URL of type URL_TYPE_ADD_AREA
	 * @return the result image after adding the area
	 */
	public BufferedImage addArea( Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String ) m_URLs.get( new Integer( URL_TYPE_ADD_AREA ) );
		szURL 		+= '?' + buildCGIParams( params );
		return mapRequest( szURL );
	}
	
	/**
	 * Updates an area that's on the map
	 * 
	 * @param params - the parameters that must be passed to the 
	 *                 URL of type URL_TYPE_UPDATE_AREA
	 * @return the result image after updating the area
	 */
	public BufferedImage updateArea( Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String ) m_URLs.get( new Integer( URL_TYPE_UPDATE_AREA ) );
		szURL 		+= '?' + buildCGIParams( params );
		return mapRequest( szURL );
	}
	
	/**
	 * Removes an area from the map
	 * 
	 * @param params - the parameters that must be passed to the 
	 *                 URL of type URL_TYPE_REMOVE_AREA
	 * @return the result image after removing the area
	 */
	public BufferedImage removeArea( Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String ) m_URLs.get( new Integer( URL_TYPE_REMOVE_AREA ) );
		szURL 		+= '?' + buildCGIParams( params );
		return mapRequest( szURL );
	}
	
	/**
	 * Creates a login request passing the authentication parameters
	 * to the server
	 * @param params - the parameters needed for user authentication
	 * 
	 * @return an info request
	 */
	public MapRequestInfo loginRequest( Hashtable params )
		throws MapRequestException
	{
		String szURL = ( String ) m_URLs.get( new Integer( URL_TYPE_LOGIN ) );
		szURL 		+= '?' + buildCGIParams( params );
		MapRequestInfo tmp = infoRequest( szURL );
		return infoMapRequest( Constants.INFO_TYPE_COMMON, null );
	}
	
	/**
	 * Logs a user out
	 * 
	 * @return an info request
	 */
	public MapRequestInfo logoutRequest( )
		throws MapRequestException
	{
		String szURL = ( String ) m_URLs.get( new Integer( URL_TYPE_LOGOUT ) );
		MapRequestInfo tmp = infoRequest( szURL );
		return infoMapRequest( Constants.INFO_TYPE_COMMON, null );
	}
	
	/**
	 * Sends a request that returns serialized MapRequestInfo object
	 * 
	 * @param szURL - the URL for the request
	 * 
	 * @return the MapRequestInfo object
	 * 
	 * @throws MapRequestException
	 */
	private MapRequestInfo infoRequest( String szURL )
		throws MapRequestException
	{
		try {
			System.out.println( "INFO REQUEST: " + szURL );
			URL url 						= new URL( szURL );
			URLConnection conn 		= url.openConnection( );
			addCookiesToConnection( conn );
			ObjectInputStream objIn = new ObjectInputStream( conn.getInputStream( ) );
			MapRequestInfo info 		= ( MapRequestInfo )objIn.readObject( );
			System.out.println( "Info request response: " + info.toString( ) );
			return info;
		} catch ( MalformedURLException e ) {
			throw new MapRequestException( "Malformed exception while trying URL: " + szURL + ": " + 
			                               e.getMessage( ), e );
		} catch ( IOException e ) {
			throw new MapRequestException( "I/O exception while reading from URL: " + 
			                               szURL + ": " + e.getMessage( ), e );
		} catch ( ClassNotFoundException e ) {
			throw new MapRequestException( "Invalid response: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Builds a CGI query string 
	 * 	param1=value1&param2=value2&...
	 * from a hash of parameters. The values are escaped
	 * using the default encoding for the module
	 * 
	 * @param params - the key-value parameters
	 * 
	 * @return the CGI query string
	 */
	private String buildCGIParams( Hashtable params )
		throws MapRequestException
	{
		try {
			Enumeration keys 	= params.keys( );
			String szParams 	= "";
			while ( keys.hasMoreElements( ) ) {
				String szName 	= ( String ) keys.nextElement( );
				String szValue	= ( String ) params.get( szName );
				szParams += szName + '=' + URLEncoder.encode( szValue, DEFAULT_ENCODING ) + '&';
			}
			if ( szParams.charAt( szParams.length( ) - 1 ) == '&' ) {
				szParams = szParams.substring( 0, szParams.length( ) - 1 );
			}
			return szParams;
		} catch ( UnsupportedEncodingException e ) {
			throw new MapRequestException( "Unsupported exception: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Adding all available cookies to the connection
	 * 
	 * @param conn - the connection
	 */
	private void addCookiesToConnection( URLConnection conn )
	{
		for ( int i = 0; i < m_Cookies.size( ); i++ ) {
			String szCookieValue = ( String )m_Cookies.get( i );
			System.out.println( "=== Sending a cookie: " + szCookieValue );
			conn.setRequestProperty( "Cookie", szCookieValue );		
		}
	}
	
	/**
	 * Gets the http content from a given URL and stores the cookies
	 * from the connection
	 * 
	 * @param szURL - the URL to fetch the content from
	 * 
	 * @return an array of byte elements
	 */
	private byte[] getHTTPContent( String szURL )
		throws MalformedURLException, IOException
	{
		URL url 						= new URL( szURL );		
		HttpURLConnection conn 	= ( HttpURLConnection ) url.openConnection( );;
		addCookiesToConnection( conn );
		
		// Get all headers from the response and search for 'Set-Cookie' ones
		Map fields 	= conn.getHeaderFields( );
		Set keys 	= fields.keySet( );
		Iterator it = keys.iterator( );

		while ( it.hasNext() ) {
			String szFieldName = ( String )it.next();
			if ( szFieldName == null ) continue;
			if ( szFieldName.toLowerCase().equals( "set-cookie" ) ) {
				// Yep, that's a cookie. Add it to the list of cookies
				List values = ( List )fields.get( szFieldName );
				for ( int i = 0; i < values.size( ); i++ ) {
					String szCookie = ( String )values.get( i );
					System.out.println( "=== Received a cookie: " + szCookie );
					m_Cookies.add( szCookie );
				}
			}
		}

		// Finally read the content from the response and put it into an array of bytes
		ByteArrayOutputStream bytes 	= new ByteArrayOutputStream( );
		BufferedInputStream in  		= new BufferedInputStream( conn.getInputStream() );
		while ( true ) {
			int b = in.read( );
			if ( b == -1 ) break;
			bytes.write( b );
		}
		
		return bytes.toByteArray();		
	}
	
	/**
	 * Makes a request that is supposed to return an image/*
	 * content type. The content is converted to an image
	 * and returned 
	 * 
	 * @param szURL - the absolute URL
	 * 
	 * @return buffered image type of object
	 * 
	 * @throws MapRequestException
	 */
	private BufferedImage mapRequest( String szURL )
		throws MapRequestException
	{
		BufferedImage answerImage = null;
		
		try {
			System.out.println( "MAP REQUEST: " + szURL );
			Toolkit toolkit 				= Toolkit.getDefaultToolkit();
			Image image 					= toolkit.createImage( getHTTPContent( szURL ) );
			MediaTracker mediaTracker	= new MediaTracker( new Container( ) );
			mediaTracker.addImage( image, 0 );
			mediaTracker.waitForAll( );
			
			if ( mediaTracker.getErrorsAny( ) != null ) {
				throw new MapRequestException( "Failed loading image from URL: " + szURL, null );
			}
			
			// Copy the image into the buffered image
			int iImageWidth	= image.getWidth( null );
			int iImageHeight	= image.getHeight( null );
			
			answerImage = new BufferedImage( iImageWidth, iImageHeight, 
			                                 BufferedImage.TYPE_INT_RGB );
			answerImage.getGraphics( ).drawImage( image, 0, 0, null );
		} catch ( IOException e ) {
			throw new MapRequestException( "I/O exception while reading URL: " + szURL + ": " + e.getMessage( ), e );
		} catch ( InterruptedException e )	{
			throw new MapRequestException( "Interrupted exception while reading URL: " + szURL + ": " + e.getMessage( ), e );
		}

		return answerImage;
	}
}
