package lib.map.base;

import java.awt.image.*;
import java.util.ArrayList;

import lib.cfg.Constants;
import lib.cfg.Config;
import lib.map.base.cfg.MapConfig;
import lib.db.DBManager;
import lib.comm.MapDefinition;
import lib.exceptions.MapStateException;
import lib.exceptions.NullMapException;
import lib.exceptions.MapConfigException;

/**
 * @class  Map
 * 
 * Defines and implements the basic operations that can be done over a
 * map: zoom in/out, translate. 
 * 
 * @author Tihomir Lazarov <tihomir@lycos.com>
 * @date   2004.08.13
 */
public class Map {
	/* --- Class members --------------------------------------------------- */
	private MappedImage		m_Map							= null;
	private int 				m_iViewerWidth          = 0;
	private int 				m_iViewerHeight     	   = 0;
	private Navigator 		m_Navigator     			= null;
	private Rectangle 		m_LastRegion				= null;
	private int 				m_iLastImageWidth		   = 0;
	private int 				m_iLastImageHeight      = 0;
	private int					m_iTranslationOffsetX	= 0;
	private int					m_iTranslationOffsetY	= 0;
	private int             m_iLastZoomLevel			= 0;
	private ArrayList			m_ZoomDefinitions			= null;
	private int             m_iMaxZoomWidth			= 0;
	private int             m_iMaxZoomHeight			= 0;
	private int             m_iMapID						= 0;
	private DBManager			m_DBManager					= null;
	private Config				m_Config						= null;
	private MapDefinition	m_MapDefinition			= new MapDefinition( );
	
	/* --- Constants ------------------------------------------------------- */
	public static final int TRANS_DIR_LEFT    		= 0;
	public static final int TRANS_DIR_RIGHT   		= 1;
	public static final int TRANS_DIR_UP      		= 2;
	public static final int TRANS_DIR_DOWN   		 	= 3;
	public static final int ZOOM_IN           		= 0;
	public static final int ZOOM_OUT          		= 1;

	/* --- Public methods -------------------------------------------------- */
	
	public Map( ) 
	{
	}
	
	/**
	 * Constructs a MapState class
	 * 
	 * @param manager	- a reference to a DBManager object instance
	 * @param config	- a reference to a Config object instance
	 * @param iMapID	- the ID of the map that must be loaded
	 * 
	 * @throws FileNotFoundException, IOException
	 */
	public void load( DBManager manager, Config config, int iMapID )
		throws MapStateException
	{
		int iViewerWidth   = Integer.parseInt( config.getValue( Constants.PARAM_VIEWER_WIDTH ) );
		int iViewerHeight  = Integer.parseInt( config.getValue( Constants.PARAM_VIEWER_HEIGHT ) );
		
		load( manager, config, iMapID, iViewerWidth, iViewerHeight );
	}
	
	/**
	 * Loads a given map using a given viewer dimensions
	 * 
	 * @param manager
	 * @param config
	 * @param iMapID
	 * @param iViewerWidth
	 * @param iViewerHeight
	 * 
	 * @throws MapStateException
	 */
	public void load( DBManager manager, Config config, int iMapID, int iViewerWidth, int iViewerHeight )
		throws MapStateException
	{
		m_DBManager 		= manager;
		m_Config				= config;
		m_iViewerWidth		= iViewerWidth;
		m_iViewerHeight	= iViewerHeight;
		
		System.out.println( "Loaded a map within a viewer window " + m_iViewerWidth + "x" + m_iViewerHeight );
		
		try {
			load( iMapID ); // the m_iMapID is set here in case the iMapID <= 0
		} catch ( MapConfigException e ) {
			throw new MapStateException( "Map reader exception: " + e.getMessage( ), e );
		}
		
		/* Zoom out to the min zoom level */
		m_iLastZoomLevel = 0;
		
		loadZoom( m_iLastZoomLevel );
	}
	
	/**
	 * Returns the current zoom level
	 * @return the value of the current zoom level
	 */
	public int getCurrentZoomLevel( )
	{
		return m_iLastZoomLevel;
	}
	
	/**
	 * Gets the number of zoom levels
	 * 
	 * @return the number of zoom levels
	 */
	public int getZoomLevels( ) 
	{
		return m_ZoomDefinitions.size( );
	}
	
	/**
	 * Zooms to a specified zoom level
	 * 
	 * @param iZoomLevel
	 * @throws MapStateException
	 */
	public void zoomTo( int iZoomLevel ) throws MapStateException {
		/* Check if the zoom level is out of range */
		if ( iZoomLevel < 0 || iZoomLevel >= m_ZoomDefinitions.size( ) ) {
			return;
		}
		
		m_iLastZoomLevel = iZoomLevel;
		loadZoom( m_iLastZoomLevel );
	}

	/**
	 * Makes a zoom in/out. Loads a new image and re-calculates the
	 * viewer window position.
	 * 
	 * @param typeOfZoom - how to zoom: ZOOM_IN or ZOOM_OUT
	 * 
	 * @throws MapStateException
	 */
	public void zoom( int typeOfZoom ) 
		throws MapStateException 
	{
		if ( !isZoomable( typeOfZoom ) ) {
			return;
		}
		
		switch ( typeOfZoom ) {
			case ZOOM_IN:
				m_iLastZoomLevel++; 
				break;
			case ZOOM_OUT:
				m_iLastZoomLevel--;
				break;
			default:
				throw new MapStateException( "Invalid type of zoom supplied '" + 
				                             typeOfZoom + "'",
				                             null );
		}
		
		loadZoom( m_iLastZoomLevel );
	}
	
	/**
	 * 
	 * @param typeOfZoom - ZOOM_IN or ZOOM_OUT
	 * 
	 * @return 'true' if the image can be zoomed with the typeOfZoom,
	 * 		  'false' otherwise 
	 * 
	 * @throws MapStateException
	 */
	public boolean isZoomable( int typeOfZoom ) 
		throws MapStateException 
	{
		switch ( typeOfZoom ) {
			case ZOOM_IN:
				return ( m_iLastZoomLevel < m_ZoomDefinitions.size( ) - 1 ); 
			case ZOOM_OUT:
				return ( m_iLastZoomLevel > 0 ); 
			default:
				throw new MapStateException( "Invalid type of zoom supplied '" + 
				                             typeOfZoom + "'",
				                             null );
		}
	}
	
	/**
	 * Translates the viewer window to some direction
	 * 
	 * @param direction - the direction to translate to
	 * 
	 * @throws MapStateException
	 */
	public void translate( int direction ) 
		throws MapStateException 
	{
		if ( !isTranslateable( direction ) ) {
			return;
		}
		
		int iOffsetX = 0;
		int iOffsetY = 0;
		
		switch ( direction ) {
			case TRANS_DIR_LEFT:
				iOffsetX = -m_iTranslationOffsetX;
				break;
			case TRANS_DIR_RIGHT:
				iOffsetX = m_iTranslationOffsetX;
				break;
			case TRANS_DIR_UP:
				iOffsetY = -m_iTranslationOffsetY;
				break;
			case TRANS_DIR_DOWN:
				iOffsetY = m_iTranslationOffsetY;
				break;
			default:
				throw new MapStateException( "Invalid translation direction '" + 
				                              direction + "'",
				                              null );
		}
		
		m_Navigator.translateNavigatorRegion( m_LastRegion, 
		                                      m_iLastImageWidth, 
		                                      m_iLastImageHeight, 
		                                      iOffsetX, iOffsetY );
	}
	
	/**
	 * Translates the viewer window with a given offsetLoading
	 * 
	 * @param iOffsetX
	 * @param iOffsetY
	 * 
	 * @throws MapStateException
	 */
	public void translateWithOffset( int iOffsetX, int iOffsetY ) 
		throws MapStateException 
	{
		System.out.println( "Translating to offset (" + iOffsetX + ", " + iOffsetY + "), " + 
		                    "last image dimentions (" + m_iLastImageWidth + ", " + m_iLastImageHeight + "), " + 
		                    "last region " + m_LastRegion.toString() );
		m_LastRegion = m_Navigator.translateNavigatorRegion( m_LastRegion, 
													  	  m_iLastImageWidth, 
													     m_iLastImageHeight, 
													     iOffsetX, iOffsetY );
		System.out.println( "After translating the region is: " + m_LastRegion.toString() );													     
	}
	
	public void translateTo( int iRealX, int iRealY ) throws MapStateException {
		Rectangle temp = new Rectangle( iRealX, iRealY, 1, 1 );
		temp = scaleToZoomLevel( temp );
		m_LastRegion = m_Navigator.translateNavigatorRegion( m_LastRegion, 
		                                                     temp.x, temp.y );
	}

	/**
	 * Checks if the viewer window is translateable to some direction
	 * 
	 * @param direction - the direction to check
	 * 
	 * @return 'true' if it can be translated more to the specified direction
	 * 
	 * @throws MapStateException
	 */	
	public boolean isTranslateable( int direction ) throws MapStateException 
	{
		switch ( direction ) {
			case TRANS_DIR_LEFT:
				return m_Navigator.isTranslateableLeft( );
			case TRANS_DIR_RIGHT:
				return m_Navigator.isTranslateableRight( );
			case TRANS_DIR_UP:
				return m_Navigator.isTranslateableUp( );
			case TRANS_DIR_DOWN:
				return m_Navigator.isTranslateableDown( );
			default:
				throw new MapStateException( "Invalid translation direction '" + 
													   direction + "'",
													   null );
		}
	}
	
	
	/**
	 * Gets the part of the image that should be displayed from the
	 * viewer that has dimentions m_iViewerWidth, m_iViewerHeight
	 * 
	 * @return a reference to a BufferedImage
	 * 
	 * @throws NullMapImageException
	 */
	public BufferedImage getViewableRegionSnapshot( ) 
		throws NullMapException
	{
		/* TODO: all brackets on new line */
		/* Check if we're invoked too early */
		if ( m_Map == null || m_LastRegion == null || m_Navigator == null ) {
			String szMessage = "";
			if ( m_Map == null        ) { szMessage = "The BigImage object is not initialized"; }
			if ( m_LastRegion == null ) { szMessage = "The LastRegion object is not initialized"; }
			if ( m_Navigator == null  ) { szMessage = "The Navigator object is not initialized"; }
			throw new NullMapException( "Class not initialized properly: " + szMessage );
		}
		
		BufferedImage part = m_Map.crop( m_LastRegion.x, m_LastRegion.y, 
		                                 m_LastRegion.width, m_LastRegion.height );
		                                 
		if ( part == null ) {
			throw new NullMapException( "Failed to crop the image" );
		}
		
		/* If the image is smaller than the dimentions of the current viewer window 
		 * stretch it to fit it. This must not be an usual case though...
		 */
		if ( m_LastRegion.width < m_iViewerWidth || m_LastRegion.height < m_iViewerHeight ) {
			BufferedImage temp = new BufferedImage( m_iViewerWidth, m_iViewerHeight, BufferedImage.TYPE_INT_RGB );
			temp.getGraphics( ).drawImage( 
				part.getScaledInstance( m_iViewerWidth, m_iViewerHeight, BufferedImage.SCALE_SMOOTH ),
				0, 0,
				null
			);
		}
		
		return part;
	}
	
	/**
	 * Returns the current map definition (the information for the
	 * map in the database)
	 * 
	 * @return an object of type MapDefinition
	 */
	public MapDefinition getMapDefinition( )
	{
		return m_MapDefinition;
	}
	
	/**
	 * Gets a full view of the image in the current zoom level
	 * 
	 * @return a reference to a BufferedImage
	 * 
	 * @throws NullMapImageException
	 */
	public BufferedImage getFullImage( ) 
		throws NullMapException
	{
		/* Check if we're invoked too early */
		if ( m_Map == null || m_LastRegion == null || m_Navigator == null ) {
			String szMessage = "";
			if ( m_Map == null        ) { szMessage = "The BigImage object is not initialized"; }
			if ( m_LastRegion == null ) { szMessage = "The LastRegion object is not initialized"; }
			if ( m_Navigator == null  ) { szMessage = "The Navigator object is not initialized"; }
			throw new NullMapException( "Class not initialized properly: " + szMessage );
		}
	
		BufferedImage part = m_Map.crop( 0, 0, 
										         m_Map.getWidth(), m_Map.getHeight( ) );
	                                 
		if ( part == null ) {
			throw new NullMapException( "Failed to crop the image" );
		}
	
		return part;
	}	
	
	/**
	 * Gets the current X position of the viewer related
	 * to the dimentions of the original image
	 * 
	 * @return the X position
	 */
	public Rectangle scaleToZoomLevel( Rectangle fullSizeRectangle ) 
	{
		Rectangle scaled     = new Rectangle( fullSizeRectangle );
		int iThisZoomWidth	= getWidth( );
		int iThisZoomHeight  = getHeight( );
		int iMaxZoomWidth		= m_iMaxZoomWidth;
		int iMaxZoomHeight	= m_iMaxZoomHeight;
		float fWCoef         = ( float )iThisZoomWidth  / iMaxZoomWidth;
		float fHCoef			= ( float )iThisZoomHeight / iMaxZoomHeight;
		
		scaled.x 		*= fWCoef;
		scaled.y 		*= fHCoef;
		scaled.width 	*= fWCoef;
		scaled.height 	*= fHCoef;

		return scaled;
	}

	/**
	 * Scales a given rectangle that has relative coordinates 
	 * to the currently zoomed image to the real size image's
	 * dimentions and coordinates
	 * 
	 * @param scaledToZoomRectangle
	 * 
	 * @return the scaled rectangle
	 */
	public Rectangle scaleToRealSize( Rectangle scaledToZoomRectangle ) 
	{
		Rectangle real			= new Rectangle( scaledToZoomRectangle );
		int iThisZoomWidth	= getWidth( );
		int iThisZoomHeight  = getHeight( );
		int iMaxZoomWidth		= m_iMaxZoomWidth;
		int iMaxZoomHeight	= m_iMaxZoomHeight;
		float fWCoef         = ( float )iThisZoomWidth  / iMaxZoomWidth;
		float fHCoef			= ( float )iThisZoomHeight / iMaxZoomHeight;
		
		real.x 		/= fWCoef;
		real.y 		/= fHCoef;
		real.width 	/= fWCoef;
		real.height /= fHCoef;

		return real;
	}
	
	/**
	 * Returns the original image width
	 * 
	 * @return the width
	 */
	public int getWidth( ) 
	{
		return m_Map.getWidth( );
	}
	
	/**
	 * Returns the original image height
	 * 
	 * @return the height
	 */
	public int getHeight( ) 
	{
		return m_Map.getHeight( );
	}
	
	/**
	 * Gets the original (maximum zoomed) image width
	 * 
	 * @return the image width
	 */
	public int getMaxZoomWidth( ) 
	{
		return m_iMaxZoomWidth;
	}
	
	/**
	 * Gets the original (maximum zoomed) image height
	 * 
	 * @return the image height
	 */
	public int getMaxZoomHeight( ) 
	{
		return m_iMaxZoomHeight;
	}
	
	/**
	 * Returns the current viewer region rectangle 
	 * 
	 * @return the rectangle
	 */
	public Rectangle getViewerRegion( ) 
	{
		return new Rectangle( m_LastRegion );
	}
		
	/* --- Private methods ------------------------------------------------- */
	/**
	 * Loads the information for each of the zoom levels for a map.
	 * Each of the zoom levels is described with an XML file that's
	 * passed to the BigImage class.
	 * 
	 * @param iMapID - the ID of the map to load
	 *  
	 * @throws MapConfigException
	 */
	private void load( int iMapID ) 
		throws MapConfigException 
	{
		MapConfig config 	= new MapConfig( m_DBManager );
		
		if ( iMapID <= 0 ) {
			iMapID = config.getDefaultMapID( );
		}
		
		m_MapDefinition	= config.getMapByID( iMapID );
		m_ZoomDefinitions = config.getZoomDefinitions( iMapID );
		if ( m_ZoomDefinitions == null ) {
			throw new MapConfigException( "There are no maps in the database for MapID [" + iMapID + "]", null );
		}
		Zoom maxZoom      = ( Zoom )m_ZoomDefinitions.get( m_ZoomDefinitions.size( ) - 1 );
		m_iMaxZoomWidth	= maxZoom.width;
		m_iMaxZoomHeight	= maxZoom.height;
		m_iMapID				= iMapID;
	}
	
	/**
	 * Loads the big image that's defined in the array of zoom definitions
	 * for the index pointed by the m_iLastZoomLevel.
	 * 
	 * @param iLevel - zoom level
	 * 
	 * @throws MapStateException
	 */
	private void loadZoom( int iLevel ) 
		throws MapStateException
	{
		if ( iLevel < 0 || iLevel >= m_ZoomDefinitions.size( ) ) {
			throw new MapStateException( "Zoom level out of bounds: " + iLevel, null );
		}
		
		Zoom zoomDefinition = ( Zoom )m_ZoomDefinitions.get( iLevel );
		
		m_Map  				= new MappedImage( m_Config.getValue( Constants.PARAM_MAPS_DIRECTORY ),
		                                  m_iMapID,
													 zoomDefinition.width, zoomDefinition.height,
		                                  zoomDefinition.hparts, zoomDefinition.vparts,
		                                  zoomDefinition.srcPrefix, zoomDefinition.srcSuffix );
		int iImageWidth	= m_Map.getWidth( );
		int iImageHeight  = m_Map.getHeight( );
		
		if ( m_Navigator == null ) {	
			m_Navigator 	= new Navigator( );
			m_LastRegion   = m_Navigator.resetNavigatorRegion( m_iViewerWidth, m_iViewerHeight, iImageWidth, iImageHeight );
		} else {
			if ( ( m_LastRegion.width < m_iViewerWidth || m_LastRegion.height < m_iViewerHeight ) && 
			     ( iImageWidth >= m_iViewerWidth || iImageHeight >= m_iViewerHeight ) ) {
				m_LastRegion   = m_Navigator.resetNavigatorRegion( m_iViewerWidth, m_iViewerHeight, iImageWidth, iImageHeight );
			} else {
				m_LastRegion	= m_Navigator.zoomNavigatorRegion( m_LastRegion, m_iLastImageWidth, m_iLastImageHeight, iImageWidth, iImageHeight );
			}
		}

		m_iLastImageWidth  		= iImageWidth;
		m_iLastImageHeight 		= iImageHeight;
		
		m_iTranslationOffsetX 	= ( m_iViewerWidth  * Constants.TRANSLATION_OFFSET_PERCENTS ) / 100;//zoomDefinition.offsetX;
		m_iTranslationOffsetY 	= ( m_iViewerHeight * Constants.TRANSLATION_OFFSET_PERCENTS ) / 100;//zoomDefinition.offsetY;
	}
	
	/**
	 * Returns the current map ID
	 * 
	 * @return the current map ID
	 */
	public int getMapID( ) 
	{
		return m_iMapID;
	}
	
	/**
	 * Returns the size of the current zoom definition.
	 * This fix is applied to be used for RoadRecognition.
	 * When scanning the big image (see lib.map.ext.*) and
	 * when reaching a pixel from another image part it
	 * takes time to load the next image, but if processing
	 * the large image by each sub-image (zoom image) it
	 * will be quicker.
	 * 
	 * @return the size of the "sub image" part
	 */
	public Rectangle getSubImageSize( ) throws NullMapException
	{
		return getViewerRegion( );
	}
}
