package lib.map.base;

import java.awt.*;
import java.awt.image.*;
import java.io.*;

import lib.map.dynamic.cfg.DynamicMapConfig;
import lib.map.base.cfg.MapConfig;
import lib.map.dynamic.Area;
import lib.map.dynamic.Event;
import lib.encoder.ImageEncoder;
import lib.db.DBManager;
import lib.exceptions.MapConfigException;
import lib.exceptions.DynamicMapConfigException;
import lib.cfg.Constants;


/**
 * @class  MappedImage
 * 
 * Represents an image as a matrix of small image parts. The basic operation
 * with the big image is showing a part of it. Instead of loading the whole
 * big image, only the small parts from the matrix that contain the region
 * to be viewed are glued together. 
 * 
 * @author Tihomir Lazarov <tihomir@lycos.com>
 * @date   2004.08.13
 */
public class MappedImage {
	/* --- Class members ---------------------------------------------------- */
	private int m_iMapID;			  /** The ID of the map to be loaded        */
	private int m_iWidth;			  /** Big image width                       */
	private int m_iHeight;			  /** Big image height                      */
	private int m_iHParts; 			  /** Number of horizontal parts the
	                         			   big image is divided into             */
	private int m_iVParts; 			  /** Number of vertical parts the big image 
	                                    is divided into                       */
	private int m_iSinglePartWidth; /** The width of a single part of image   */
	private int m_iSinglePartHeight;/** The height of a single part of image  */
	private String[][] m_szFilesMap;/** The filenames of each part of image   */
	
	/* --- Constants -------------------------------------------------------- */
	private static int PRECISION_COEF   = 100; /* To avoid float number calcs */
	private static int HPARTS  			= 15;  /* Number of horizontal parts  */
	private static int VPARTS 				= 12;  /* Number of vertical parts    */
	
	/**
	 * Loads the information about the big image from an XML file.
	 * The information contains the size of the whole image and
	 * the chunks it is divided into.
	 *
	 * @param szBaseDir			- base files directory
	 * @param iMapID				- the ID of the map that will be loaded. 0 if there
	 * 								  won't be loaded a MapID
	 * @param iWidth				- image width
	 * @param iHeight				- image height
	 * @param iHParts				- number of horizontal parts the image is 
	 *                           divided into
	 * @param iVParts				- number of vertical parts
	 * @param szSrcPrefix		- part file name prefix
	 * @param szSrcSuffix		- part file name suffix 
	 */
	public MappedImage( String szBaseDir, 
	                 int iMapID,
	                 int iWidth, int iHeight, 
	                 int iHParts, int iVParts,
	                 String szSrcPrefix, String szSrcSuffix ) {
		/* Extract the width and height of the image */
		m_iWidth  = iWidth;
		m_iHeight = iHeight;
		
		m_iMapID = iMapID;
		
		/* Extract the number of horizontal and vertical parts the image 
		 * is divided into 
		 */
		m_iHParts = iHParts;
		m_iVParts = iVParts;
		
		/* Store the files for each image part */
		m_szFilesMap = new String[ m_iVParts ][ m_iHParts ];
		for ( int j = 0; j < m_iVParts; j++ ) {
			for ( int i = 0; i < m_iHParts; i++ ) {
				m_szFilesMap[ j ][ i ] = szBaseDir + "/" + m_iMapID + "/" +  
				                         generateZoomPartFileName( szSrcPrefix, 
				                                                   szSrcSuffix, 
				                                                   iWidth, 
				                                                   iHeight, 
				                                                   i, 
				                                                   j );
			}
		}
		
		/* Calculate the dimentions of a single part */
		m_iSinglePartWidth  = PRECISION_COEF * m_iWidth  / m_iHParts;
		m_iSinglePartHeight = PRECISION_COEF * m_iHeight / m_iVParts;
	}
	
	/**
	 * Generates a file name out of some prefix string, the dimentions
	 * of an image and a suffix. The file name also includes the row and
	 * column of the small image (the big image contains many little images
	 * with file names like the generated below).
	 * 
	 * @param szPrefix	- file name prefix (e.g. zoom_2_)
	 * @param szSuffix	- file name suffix (e.g. .jpg)
	 * @param iWidth		- image width
	 * @param iHeight		- image height
	 * @param iHPart		- the sequence number of the 
	 * 						  horizontal part (starting from 0)
	 * @param iVPart		- the sequence number of the vertical part
	 * 
	 * @return the generated file name
	 */
	public static String generateZoomPartFileName( String szPrefix, 
	                                           String szSuffix, 
	                                           int iWidth, 
	                                           int iHeight, 
	                                           int iHPart, 
	                                           int iVPart ) {
		return szPrefix + 
				 iWidth + "x" + iHeight + 
				 "_" + 
				 iHPart + "_" + iVPart + 
				 szSuffix;
	}
	
	/**
	 * Returns the full image width
	 * 
	 * @return the image width
	 */
	public int getWidth( ) {
		return m_iWidth;
	}

	/**
	 * Returns the full image height
	 * 
	 * @return the image height
	 */	
	public int getHeight( ) {
		return m_iHeight;
	}
	
	/**
	 * Crops a part of the big image. Estimates which of the small parts to 
	 * join together and cuts off the part that's needed.
	 * 
	 * @param x
	 * @param y
	 * @param iWidth
	 * @param iHeight
	 * 
	 * @return a reference to an image object
	 */
	public BufferedImage crop( int x, int y, int iWidth, int iHeight ) {
		
		/* First check for out of bounds conditions */
		if ( x + iWidth >= m_iWidth ) {
			x = m_iWidth - iWidth;
			if ( x <= 0 ) {
				x      = 0;
				iWidth = m_iWidth;
			}
		}
		
		if ( y + iHeight >= m_iHeight ) {
			y = m_iHeight - iHeight;
			if ( y <= 0 ) {
				y       = 0;
				iHeight = m_iHeight;
			}
		}
		
		/* Calculate the row/column where the "y"/"x" position resides to 
		 * determine which is the left-top part that should be
		 * used
		 */
		int iStartColumn 	= ( x * PRECISION_COEF ) / m_iSinglePartWidth;
		int iStartRow 	  	= ( y * PRECISION_COEF ) / m_iSinglePartHeight;
		int iEndColumn   	= ( ( x + iWidth  - 1 ) * PRECISION_COEF ) / m_iSinglePartWidth;
		int iEndRow      	= ( ( y + iHeight - 1 ) * PRECISION_COEF ) / m_iSinglePartHeight;
		
		int iRegionWidth	= ( ( iEndColumn - iStartColumn + 1 ) * m_iSinglePartWidth  ) / PRECISION_COEF;
		int iRegionHeight = ( ( iEndRow    - iStartRow    + 1 ) * m_iSinglePartHeight ) / PRECISION_COEF;
		
		int iRelativeX			= x - ( iStartColumn * m_iSinglePartWidth  ) / PRECISION_COEF;
		int iRelativeY    	= y - ( iStartRow    * m_iSinglePartHeight ) / PRECISION_COEF;
		int iRelativeWidth 	= iRelativeX + iWidth  > iRegionWidth  ? iRegionWidth  - iRelativeX : iWidth;
		int iRelativeHeight  = iRelativeY + iHeight > iRegionHeight ? iRegionHeight - iRelativeY : iHeight;
		
		/* Create an image where the part images will be glued together */ 
		BufferedImage imageRegion = new BufferedImage( iRegionWidth, 
		                                               iRegionHeight, 
		                                               BufferedImage.TYPE_INT_RGB );
		
		/* Load all images that are needed for the producing of the cropped image */
		MediaTracker mediaTracker = new MediaTracker( new Container( ) );
		Graphics terrain          = imageRegion.getGraphics( );
		int iNumberOfParts        = ( iEndColumn - iStartColumn + 1 ) * ( iEndRow - iStartRow + 1 );
		Image[] partImages		  = new Image[ iNumberOfParts ];
		int iImageIndex           = 0;
		int i = 0;
		int j = 0;
		
		for ( i = iStartColumn; i <= iEndColumn; i++ ) {
			for ( j = iStartRow; j <= iEndRow; j++ ) {
				partImages[ iImageIndex ] = Toolkit.getDefaultToolkit( ).getImage( m_szFilesMap[ j ][ i ] );
				if ( partImages[ iImageIndex ] == null ) {
					throw new NullPointerException( "Failed to open \"" + m_szFilesMap[ j ][ i ] + "\"" );
				}
				mediaTracker.addImage( partImages[ iImageIndex ], iImageIndex );
				iImageIndex++;
			}
		}
		
		/* Wait for the images to load */
		try {
			mediaTracker.waitForAll( );
			if ( mediaTracker.isErrorAny() == true ) {
				return null;
			}
		} catch ( InterruptedException e ) {
			return null;
		} catch ( NullPointerException e ) {
			return null;
		}
		
		/* Glue the images */
		iImageIndex = 0;
		int iPartXPos	= 0;
		int iPartYPos  = 0;
		for ( i = iStartColumn; i <= iEndColumn; i++ ) {
			iPartYPos    = 0;
			for ( j = iStartRow; j <= iEndRow; j++ ) {
				terrain.drawImage( partImages[ iImageIndex++ ], iPartXPos, iPartYPos, null );
				iPartYPos  += m_iSinglePartHeight / PRECISION_COEF;
			}
			iPartXPos	+= m_iSinglePartWidth / PRECISION_COEF;
		}
		
		/* Now cut off the part we want */
		//return imageRegion.getSubimage( iRelativeX, iRelativeY, iRelativeWidth, iRelativeHeight );
		BufferedImage result = imageRegion.getSubimage( iRelativeX, iRelativeY, iRelativeWidth, iRelativeHeight );
		return result;
	}
	
	/**
	 * Creates a mapped image and stores its definition in an XML file
	 * 
	 * @param manager - the DB manager
	 * @param szMapName - the name of the map
	 * @param szMapDescription - the description of the map
	 * @param lScale - the scale of the map - the distance in meters for 1px of the map
	 * @param bDefault - is default map
	 * @param szInputFileName
	 * @param szOutputDirectory
	 * @param iMaxZoomLevel
	  
	 * @throws FileNotFoundException
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws IllegalArgumentException
	 */
	public static void create( DBManager manager,
										String szMapName,
										String szMapDescription,
										long   lScale, 
										boolean bDefault,
	                           String szInputFileName, 
										String szOutputDirectory,
										int iMaxZoomLevel ) 
				throws FileNotFoundException, 
						 InterruptedException, 
						 IOException,
						 IllegalArgumentException {
				
		/* Check if the max zoom level is valid */
		if ( !( iMaxZoomLevel > 0 ) ) {
			throw new IllegalArgumentException( "The max zoom level should be a positive number" );
		}
		
		//_( "Creating a mapped image. Loading the original image [" + szInputFileName + "]..." );

		/* Try to open the input file */
		MediaTracker mediaTracker = new MediaTracker( new Container( ) );
		Image image               = Toolkit.getDefaultToolkit( ).getImage( szInputFileName );
		
		if ( image == null ) {
			throw new FileNotFoundException( "Failed getting image \"" + szInputFileName + "\"" );
		}
		
		mediaTracker.addImage( image, 0 );
		mediaTracker.waitForAll( );
		
		/* Check if the input file can be opened for reading */		
		if ( mediaTracker.getErrorsAny( ) != null ) {
			throw new FileNotFoundException( "Cannot open \"" + szInputFileName + "\" for reading" );
		}
		
		/* Get image width and height */		
		int iOriginalWidth	= image.getWidth( null );
		int iOriginalHeight	= image.getHeight( null );
		
		/* Open the XML file */
		MapConfig mapConfig = new MapConfig( manager );
		
		try {
			int iMapID = mapConfig.addMapInstance( szMapName, szMapDescription, lScale, bDefault );
			
			/* Try to create a directory in the m_szOutputDirectory */
			File inputFile 		= new File ( szInputFileName );
			String szDirectory   = szOutputDirectory + "/" + iMapID + "/";
			File outputDirectory = new File( szDirectory );

			/* Create the directory if it doesn't exist */
			if ( outputDirectory.exists( ) == false && outputDirectory.mkdirs( ) == false ) {
				throw new IOException( "Failed to create directory \"" + szDirectory + "\"" );
			}
			
			/* Add default areas */
			DynamicMapConfig config = new DynamicMapConfig( manager, iMapID );
			config.clearAreas( );
	
			Area area = new Area( 0, 0, iOriginalWidth, iOriginalHeight, 0 );
			area.setPolicy( Area.POLICY_UNREMOVABLE );
			area.addEvent( new Event( Constants.EVENT_ON_MOVE_DOWN		, Constants.ACTION_MOVE_DOWN ) );
			area.addEvent( new Event( Constants.EVENT_ON_MOVE_LEFT		, Constants.ACTION_MOVE_LEFT ) );
			area.addEvent( new Event( Constants.EVENT_ON_MOVE_UP			, Constants.ACTION_MOVE_UP ) );
			area.addEvent( new Event( Constants.EVENT_ON_MOVE_RIGHT		, Constants.ACTION_MOVE_RIGHT ) );
			area.addEvent( new Event( Constants.EVENT_ON_ZOOM_IN			, Constants.ACTION_ZOOM_IN ) );
			area.addEvent( new Event( Constants.EVENT_ON_ZOOM_OUT			, Constants.ACTION_ZOOM_OUT ) );
			area.addEvent( new Event( Constants.EVENT_ON_ZOOM_IN_MOVE	, Constants.ACTION_ZOOM_IN_MOVE ) );
			area.addEvent( new Event( Constants.EVENT_ON_ZOOM_OUT_MOVE  , Constants.ACTION_ZOOM_OUT_MOVE ) );
			area.addEvent( new Event( Constants.EVENT_ON_ZOOM_TO        , Constants.ACTION_ZOOM_TO ) );
			config.addArea( area );
	
			for ( int i = 0; i < iMaxZoomLevel; i++ ) {
				String szSrcPrefix	= "zoom_" + i + "_";
				String szSrcSuffix	= ".jpg";
				int iWidth        	= ( int )( ( iOriginalWidth  / ( float )iMaxZoomLevel ) * ( i + 1 ) );
				int iHeight       	= ( int )( ( iOriginalHeight / ( float )iMaxZoomLevel ) * ( i + 1 ) );
				
				/*
				_( "Genreating images for zoom level #" + i + 
	            ". The image dimentions on this zoom level are ( " + iWidth + "x" + iHeight + " )" );
	         */
	
				int iHParts				= HPARTS;
				int iVParts				= VPARTS;
				int iOffsetRectWidth = 50;
				int iOffsetRectHeight= 50;
				int iOffsetX			= iWidth  / iOffsetRectWidth  == 0 ? 1 : iWidth  / iOffsetRectWidth;
				int iOffsetY			= iHeight / iOffsetRectHeight == 0 ? 1 : iHeight / iOffsetRectHeight;
				
				//_( "The matrix the current image is represented as is ( " + iHParts + "x" + iVParts + " )" );
	
				Zoom zoomDefinition = new Zoom( );
				zoomDefinition.zoomLevel		= i;
				zoomDefinition.srcPrefix		= szSrcPrefix;
				zoomDefinition.srcSuffix		= szSrcSuffix;
				zoomDefinition.hparts			= iHParts;
				zoomDefinition.vparts			= iVParts;
				zoomDefinition.width				= iWidth;
				zoomDefinition.height			= iHeight;
				zoomDefinition.offsetX			= iOffsetX;
				zoomDefinition.offsetY			= iOffsetY;
				mapConfig.addZoom( iMapID, zoomDefinition );
	
				int iSinglePartWidth  = iWidth  / iHParts;
				int iSinglePartHeight = iHeight / iVParts;
				int iSingleOriginalPartWidth  = iOriginalWidth  / iHParts;
				int iSingleOriginalPartHeight = iOriginalHeight / iVParts;
				
				System.gc( );
	
				/* Now divide the image into m_iHParts horizontal and m_iVParts
				 * vertical parts and store them on the FS
				 */
				for ( int iHPart = 0; iHPart < iHParts; iHPart++ ) {
					for ( int iVPart = 0; iVPart < iVParts; iVPart++ ) {
						
						String szSpaces = i < 10 ? "  " : ( i < 100 ? " " : "" );
						__( "Zoom level " + szSpaces + ( i + 1 ) + "/" + iMaxZoomLevel,
						    iHPart  * iVParts + iVPart + 1, 
						    iHParts * iVParts );
			
						/* Generate the file name where the image will be saved to */
						String szFileName 	= MappedImage.generateZoomPartFileName( 
																		szSrcPrefix, 
																		szSrcSuffix, 
																		iWidth, 
																		iHeight, 
																		iHPart, 
																		iVPart );
			
						int iPartX = iHPart * iSingleOriginalPartWidth;
						int iPartY = iVPart * iSingleOriginalPartHeight;
						
						//_( "*** Creating a buffered image ( " + iSingleOriginalPartWidth + "x" + iSingleOriginalPartHeight + " )" );
			
						/* Make a zoomed image */ 
						BufferedImage originalPart = new BufferedImage( iSingleOriginalPartWidth, 
																					   iSingleOriginalPartHeight, 
																					   BufferedImage.TYPE_INT_RGB );
						
						/*															   
						_( "*** Putting the part of the original image that contains this region ( " + 
						        iPartX + ", " + iPartY + ", " + 
						        iSingleOriginalPartWidth + ", "+ iSingleOriginalPartHeight + " )" );
						*/
						
						originalPart.getGraphics( ).drawImage( image, 
															  0, 0, 
															  iSingleOriginalPartWidth, 
															  iSingleOriginalPartHeight, 
															  iPartX, iPartY, 
															  iPartX + iSingleOriginalPartWidth, 
															  iPartY + iSingleOriginalPartHeight, 
															  null );
	
						//_( "*** Reserving memory for a single part image ( " + iSinglePartWidth + "x" + iSinglePartHeight + " )" );														  
						BufferedImage partImage = new BufferedImage( iSinglePartWidth, iSinglePartHeight, BufferedImage.TYPE_INT_RGB );
			
						//_( "*** Drawing the part image..." );
						/* Get the sub image that will be stored into this part */					
						partImage.getGraphics( ).drawImage(
							originalPart,
							0, 0,
							iSinglePartWidth, iSinglePartHeight,  
							0, 0, 
							iSingleOriginalPartWidth, 
							iSingleOriginalPartHeight, 
							null );
							
						//_( "*** Encoding the image..." );
			
						/* Encode and save to FS */
						ImageEncoder enc = new ImageEncoder( );
						enc.JPEGEncode( partImage, szDirectory + "/" + szFileName );
						
						//_( "*** Done" );
			
						partImage.flush( );
						System.gc( );
					}
				}
				
				//_( "Done with this zoom level" );
	
				System.gc( );
			}
		} catch ( MapConfigException e ) {
			_( "MAP CONFIG EXCEPTION: " + e.getMessage( ) );	
		} catch ( DynamicMapConfigException e ) {
			_( "DYNAMIC MAP CONFIG EXCEPTION: " + e.getMessage( ) );
		}
		_( "Finished" );
	}
	
	private static void _( String szMessage ) {
		System.out.println( szMessage );
	}
	
	private static void __( String szMessage, int iPassed, int iAll ) {
		int i;
		int iPercent 			= ( iPassed * 100 ) / iAll;
		int iNumberOfPoints  = 40;
		int iThesePoints 		= ( iNumberOfPoints * iPercent ) / 100;

		for ( i = 0; i < 256; i++ ) {
			System.out.print( "\b" );
		}
		System.out.print( szMessage + " " );
		for ( i = 0; i < iThesePoints; i++ ) {
			System.out.print( "=" ); 
		}
		for ( ; i < iNumberOfPoints; i++ ) {
			System.out.print( "-" );
		}
		String szSpaces = iPercent < 10 ? "  " : ( iPercent < 100 ? " " : "" );
		System.out.print( "| " + szSpaces + iPercent + "%" );
		System.out.flush( );
	}
}
