package lib.map.base;

/**
 * @class lib.map.base.Navigator
 * 
 * Manages the transitions of the viewer window
 * 
 * @author Tihomir Lazarov
 */
public class Navigator {
	
	private boolean m_bTranslateableLeft   = true;
	private boolean m_bTranslateableRight  = true;
	private boolean m_bTranslateableUp     = true;
	private boolean m_bTranslateableDown	= true;
	
	public Navigator( ) {
	}
	
	/**
	 * Calculates the navigator region based on the last navigator region and 
	 * the image dimentions.
	 * 
	 * @param lastNavRegion		- the last navigator region 
	 * @param iOldImageWidth 	- the width of the former zoom
	 * @param iOldImageHeight	- the height of the former zoom
	 * @param iNewImageWidth	- the width of the current zoom
	 * @param iNewImageHeight 	- the height of the current zoom
	 * 
	 * @return a reference to the new navigator region
	 */
	public Rectangle zoomNavigatorRegion( Rectangle lastNavRegion, 
	                                      int iOldImageWidth, 
	                                      int iOldImageHeight, 
	                                      int iNewImageWidth, 
	                                      int iNewImageHeight ) {
		Rectangle newRegion = lastNavRegion;
		
		float fCenterXPos = lastNavRegion.x + ( float )lastNavRegion.width  / 2;
		float fCenterYPos = lastNavRegion.y + ( float )lastNavRegion.height / 2;
		
		newRegion.x	= Math.round( ( fCenterXPos * iNewImageWidth  ) / iOldImageWidth  - ( float )lastNavRegion.width  / 2 );
		newRegion.y = Math.round( ( fCenterYPos * iNewImageHeight ) / iOldImageHeight - ( float )lastNavRegion.height / 2 );
		
		return putInBounds( newRegion, iNewImageWidth, iNewImageHeight );
	}
	
	/**
	 * Translates the viewer rectangle with some horizontal and vertical offset
	 * 
	 * @param lastNavRegion
	 * @param iImageWidth
	 * @param iImageHeight
	 * @param iOffsetX
	 * @param iOffsetY
	 * 
	 * @return the new viewer rectangle (or navigator region)
	 */
	public Rectangle translateNavigatorRegion ( 
									Rectangle lastNavRegion, 
									int iImageWidth, 
									int iImageHeight, 
									int iOffsetX, 
									int iOffsetY ) {
		Rectangle newRegion = lastNavRegion;
		
		newRegion.x += iOffsetX;
		newRegion.y += iOffsetY;
		
		return this.putInBounds( newRegion, iImageWidth, iImageHeight );
	}
	
	/**
	 * Translates the given region to some exact position
	 * 
	 * @param lastNavRegion
	 * @param x
	 * @param y
	 * 
	 * @return the translated rectangle
	 */
	public Rectangle translateNavigatorRegion( Rectangle lastNavRegion, int x, int y ) {
		lastNavRegion.x = x;
		lastNavRegion.y = y;
		return lastNavRegion;
	}
	
	/**
	 * Centers the viewer position
	 * 
	 * @param iViewerWidth
	 * @param iViewerHeight
	 * @param iImageWidth
	 * @param iImageHeight
	 * 
	 * @return the centered region
	 */
	public Rectangle resetNavigatorRegion( int iViewerWidth, int iViewerHeight, int iImageWidth, int iImageHeight ) {
		Rectangle region = new Rectangle( );
		
		/* Check for "out-of-bounds" condition */
		if ( iViewerWidth > iImageWidth ) {
			iViewerWidth = iImageWidth;
		}
		if ( iViewerHeight > iImageHeight ) {
			iViewerHeight = iImageHeight;
		}
		
		/* Center the viewer region */
		region.x      = ( iImageWidth  - iViewerWidth  ) / 2;
		region.y      = ( iImageHeight - iViewerHeight ) / 2;
		region.width  = iViewerWidth;
		region.height = iViewerHeight;
		
		return region;
	}
	
	/**
	 * Checks if the region is in the image bounds
	 * 
	 * @param region
	 * @param iWidth
	 * @param iHeight
	 * 
	 * @return the region that has fixed margins if they are out of bounds
	 */
	private Rectangle putInBounds( Rectangle region, int iWidth, int iHeight ) {
		m_bTranslateableLeft   	= true;
		m_bTranslateableRight  	= true;
		m_bTranslateableUp		= true;
		m_bTranslateableDown	  	= true;
		
		/* If no more translation to the left */
		if ( region.x < 0 ) {
			m_bTranslateableLeft = false;
			region.x 				= 0;
		}
		
		/* If no more translation to the top */
		if ( region.y < 0 ) {
			m_bTranslateableUp   = false;
			region.y 				= 0;
		}
		
		/* If the translation to the right cannot be made */
		if ( region.x + region.width >= iWidth ) {
			m_bTranslateableRight 	= false;
			region.x 					= iWidth - region.width;
			
			/* If the viewer window is bigger than the image */
			if ( region.x < 0 ) {
				m_bTranslateableLeft	= false;
				region.x     			= 0;
				region.width 			= iWidth;
			}
		}
		
		/* If the translation to the bottom cannot be made */
		if ( region.y + region.height >= iHeight ) {
			m_bTranslateableDown	= false;
			region.y 				= iHeight - region.height;
			
			/* If the viewer window's height is bigger than the image */
			if ( region.y < 0 ) {
				m_bTranslateableUp	= false;
				region.y      			= 0;
				region.height 			= iHeight; 
			}
		}
		
		return region;
	}
	
	/**
	 * Checks if the current viewer window is translateable to the left
	 * 
	 * @return boolean value
	 */
	public boolean isTranslateableLeft( ) {
		return m_bTranslateableLeft;
	}
	
	/**
	 * Checks if the current viewer window is translateable to the right
	 * 
	 * @return boolean value
	 */
	public boolean isTranslateableRight( ) {
		return m_bTranslateableRight;
	}
	
	/**
	 * Checks if the current viewer window is translateable to the top
	 * 
	 * @return boolean value
	 */
	public boolean isTranslateableUp( ) {
		return m_bTranslateableUp;
	}
	
	/**
	 * Checks if the current viewer window is translateable to the bottom
	 * 
	 * @return boolean value
	 */
	public boolean isTranslateableDown( ) {
		return m_bTranslateableDown;
	}
}
