package lib.map.base.cfg;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.util.ArrayList;

import lib.db.DBManager;
import lib.map.base.Zoom;
import lib.comm.MapDefinition;
import lib.map.ext.MapSelector;
import lib.exceptions.DBManagerException;
import lib.exceptions.MapConfigException;

public class MapConfig {
	
	private DBManager m_Manager = null;

	/**
	 * Class constructor 
	 * 
	 * @param manager			- database manager
	 * @param szName 			- map name
	 * @param szDescription - map descritpion
	 */
	public MapConfig( DBManager manager ) 
	{
		m_Manager = manager;
	}
	
	/**
	 * Adds a new map instance of a map without adding the zooms and
	 * returns an ID of the added map
	 * 
	 * @param szName - the map name
	 * @param szDescription - the map description
	 * @param lScale - the map scale (meters per 1px)
	 * @param bDefault - if the map must be set as default
	 * 
	 * @return the ID of the added map
	 * 
	 * @throws MapConfigException
	 */
	public int addMapInstance( String szName, String szDescription, 
	                           long lScale, boolean bDefault )
		throws MapConfigException
	{
		int iMapID;
		try {
			m_Manager.execute( "LOCK TABLES maps WRITE" );
			PreparedStatement statement = 
				m_Manager.getPreparedStatement( 
					"INSERT INTO maps ( Name, Description, Scale, IsDefault ) VALUES( ?, ?, ?, ? )" );
			statement.setString	( 1, szName );
			statement.setString	( 2, szDescription );
			statement.setLong		( 3, lScale );
			statement.setInt 		( 4, bDefault ? 1 : 0 );
			if ( bDefault ) {
				// Clear the last default map
				m_Manager.execute( "UPDATE maps SET IsDefault = 0 WHERE IsDefault = 1" );
			}
			statement.executeUpdate( );
			ResultSet lastInsertIDResult = m_Manager.select( "SELECT LAST_INSERT_ID( ) AS LID" );
			lastInsertIDResult.first( );
			iMapID = lastInsertIDResult.getInt( "LID" );
			m_Manager.execute( "UNLOCK TABLES" );
		} catch ( DBManagerException e ) {
			throw new MapConfigException( "DBManagerException while adding a map instance: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new MapConfigException( "SQLException while adding a map instance: " + e.getMessage( ), e );
		}
		
		return iMapID;
	}
	
	
	/**
	 * Gets the map information by a given map ID
	 * 
	 * @param iMapID - the ID of the map
	 * 
	 * @return a non-null map definition object
	 */
	public MapDefinition getMapByID( int iMapID )
	{
		MapSelector selector = new MapSelector( m_Manager );
		ArrayList maps   	   = selector.getMaps( iMapID );
		
		if ( maps == null ) return new MapDefinition( );
		
		return ( MapDefinition ) maps.get( 0 );
	}
	
	/**
	 * Gets a map instance by name
	 * 
	 * @param szName
	 * 
	 * @return the map ID with the given name or -1 if it's not found
	 * 
	 * @throws MapConfigException - in case of SQL error or
	 *         more than one map with this name exists
	 */
	public int getMapIDByName( String szName )
		throws MapConfigException 
	{
		int iMapID = -1;
		try {
			PreparedStatement statement = m_Manager.getPreparedStatement( "SELECT ID FROM maps WHERE Name = ?" );
			statement.setString( 1, szName );
			ResultSet mapSet = statement.executeQuery( );
			if ( mapSet.first() ) { // Check if there are more than 1 map with this name
				if ( mapSet.getFetchSize( ) == 1 ) {
					iMapID = mapSet.getInt( "ID" ); 
				} else {
					throw new MapConfigException( "Duplicated name found", null );
				}
			} else { // There's no such map
				return -1;
			}
		} catch ( DBManagerException e ) {
			throw new MapConfigException( "DBManagerException: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new MapConfigException( "SQLException while getting map ID by name: " + e.getMessage( ), e );
		}
		
		return iMapID;
	}
	
	/**
	 * Gets the map ID of the map that has IsDefault flag 1
	 * 
	 * @return the map ID
	 * 
	 * @throws MapConfigException
	 */
	public int getDefaultMapID( ) 
		throws MapConfigException
	{
		try {
			ResultSet mapSet = m_Manager.select( "SELECT ID FROM maps WHERE IsDefault = 1" );
			if ( mapSet.first( ) ) {
				return mapSet.getInt( "ID" );				
			} else {
				throw new MapConfigException( "There is no default map in the datbase", null );
			}
		} catch ( DBManagerException e ) {
			throw new MapConfigException( "DB Manager Exception: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new MapConfigException( "SQL Exception while getting the default map ID: " + e.getMessage( ), e );
		}
	}

	/**
	 * 
	 * @param szName
	 * @return
	 * @throws MapConfigException
	 */
	public ArrayList getZoomDefinitions( int iMapID )
		throws MapConfigException 
	{
		ArrayList zoomDefinitions = new ArrayList( );
		try {
			ResultSet mapSet = m_Manager.select( "SELECT * FROM zoom_levels WHERE MapID = " + iMapID );
			while ( mapSet.next() ) { // Check if there are more than 1 map with this name
				Zoom zoom = new Zoom( );
				zoom.zoomLevel = mapSet.getInt( "Level" );
				zoom.srcPrefix	= mapSet.getString( "SrcPrefix" );
				zoom.srcSuffix	= mapSet.getString( "SrcSuffix" );
				zoom.hparts		= mapSet.getInt( "HParts" );
				zoom.vparts		= mapSet.getInt( "VParts" );
				zoom.width		= mapSet.getInt( "Width" );
				zoom.height		= mapSet.getInt( "Height" );
				zoom.offsetX	= mapSet.getInt( "OffsetX" );
				zoom.offsetY	= mapSet.getInt( "OffsetY" );
				zoomDefinitions.add( zoom );
			}
		} catch ( DBManagerException e ) {
			throw new MapConfigException( "DBManagerException: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new MapConfigException( "SQLException while getting the zoom defitions: " + e.getMessage( ), e );
		}
		
		return zoomDefinitions.size( ) == 0 ? null : zoomDefinitions;
	}
	
	/**
	 * Deletes all zooms for a map
	 * 
	 * @param iMapID - the map ID
	 * 
	 * @throws MapConfigException
	 */
	public void clearZooms( int iMapID ) 
		throws MapConfigException
	{
		try {
			m_Manager.execute( "DELETE FROM zoom_levels WHERE MapID = " + iMapID );
		} catch ( DBManagerException e ) {
			throw new MapConfigException( "DBManagerException: " + e.getMessage( ), e );
		}
	}	
	

	/**
	 * Adds a "<Zoom ...>" XML tag that describes the files that contain
	 * the image parts of a zoom level X
	 * 
	 * @param zoomDefinition - a structure of type ZoomDefinition with
	 *                         initialized members
	 */
	public void addZoom( int iMapID, Zoom zoomDefinition )
		throws MapConfigException 
	{
		try {
			PreparedStatement statement = 
				m_Manager.getPreparedStatement( 
					"INSERT INTO zoom_levels ( MapID, Level, SrcPrefix, SrcSuffix, HParts, VParts, OffsetX, OffsetY, Width, Height ) " +
				                    "VALUES ( ?    , ?    , ?        , ?        , ?     , ?     , ?      , ?      , ?    , ?      ) " );
			int iIndex = 1;
			statement.setInt   ( iIndex++, iMapID                    );
			statement.setInt   ( iIndex++, zoomDefinition.zoomLevel 	);		
			statement.setString( iIndex++, zoomDefinition.srcPrefix 	);
			statement.setString( iIndex++, zoomDefinition.srcSuffix 	);
			statement.setInt   ( iIndex++, zoomDefinition.hparts 		);
			statement.setInt   ( iIndex++, zoomDefinition.vparts 		);
			statement.setInt   ( iIndex++, zoomDefinition.offsetX 	);
			statement.setInt   ( iIndex++, zoomDefinition.offsetY 	);
			statement.setInt   ( iIndex++, zoomDefinition.width 		);
			statement.setInt   ( iIndex++, zoomDefinition.height 		);
			
			statement.executeUpdate( );
			
		} catch ( DBManagerException e ) {
			throw new MapConfigException( "DBManagerException: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new MapConfigException( "SQLException: " + e.getMessage( ), e );
		}
	}
}
