package lib.map.dynamic;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Enumeration;

import lib.map.base.Rectangle;

/**
 * @class lib.map.dynamic.Area
 * 
 * Describes a rectangle area that can be in a "z-index" relation
 * with other areas. The object may have a number of events with
 * actions attached to them. 
 * 
 * @author Tihomir Lazarov
 * @date   Aug 16, 2004
 */
public class Area {
	private Rectangle m_Rect	= null;
	private int m_iZIndex		= 0;
	private int m_iPolicy		= 0;
	private Hashtable m_Info	= null;
	private ArrayList m_Events	= null;
	private long m_lUniqueID	= 0;
	
	public static final int POLICY_REMOVABLE   = 0;
	public static final int POLICY_UNREMOVABLE = 1;
	
	/**
	 * Area class constructor 
	 * 
	 * @param x			- the X position of the area
	 * @param y			- the Y position of the area
	 * @param width	- the width of the rectangle
	 * @param height	- the height of the rectangle
	 * @param iZIndex	- the Z position of the rectangle. "0" is the 
	 *                  area that stays at the bottom
	 */
	public Area( int x, int y, int width, int height, int iZIndex ) {
		m_Rect		= new Rectangle( x, y, width, height );
		m_iZIndex 	= iZIndex;
		m_iPolicy   = POLICY_REMOVABLE; 
	
		m_Info 	= new Hashtable( );
		m_Events = new ArrayList( );
		
		setUniqueID( );
	}
	
	/**
	 * Area class constructor. Constructs an area by rectangle 
	 *
	 * @param rect    The rectangle position and dimentions
	 * @param iZIndex	- the Z position of the rectangle. "0" is the 
	 *                  area that stays at the bottom
	 */
	public Area( Rectangle rect, int iZIndex ) {
		new Area( rect.x, rect.y, rect.width, rect.height, iZIndex ); 
	}
	
	
	/**
	 * Sets some area information. The information is stored
	 * into a hash.
	 * 
	 * @param szName	- the name of the pair of information	
	 * @param szValue	- the value of the information
	 */
	public void addInfo( String szName, String szValue ) 
	{
		m_Info.put( szName, szValue );
	}
	
	/**
	 * Adds an object of type 'Event' to the current area
	 * 
	 * @param event - the event
	 */
	public void addEvent( Event event ) 
	{
		// Check if this event already exists and replace it
		// otherwise add it
		for ( int i = 0; i < m_Events.size( ); i++ ) {
			Event existingEvent = ( Event ) m_Events.get( i );
			if ( existingEvent.getIdentifier( ).equals( event.getIdentifier( ) ) ) {
				m_Events.remove( i );
			}
		}
		
		m_Events.add( event );
	}
	
	/**
	 * Returns the current area unique ID
	 * 
	 * @return the unique area ID
	 */
	public long getUniqueID( )
	{
		return m_lUniqueID;
	}
	
	/**
	 * Returns the list of events available for the current
	 * area
	 * 
	 * @return an ArrayList of events
	 */
	public ArrayList getEvents( ) {
		return ( ArrayList )m_Events.clone( );
	}
	
	/**
	 * Gets the whole hashtable of information
	 * 
	 * @return the information hashtable
	 */
	public Hashtable getInfo( ) {
		return ( Hashtable )m_Info.clone( );
	}
	
	/**
	 * Sets the area policy 
	 * 
	 * @param iPolicy - the policy type
	 */
	public void setPolicy( int iPolicy ) 
	{
		m_iPolicy = iPolicy;
	}
	
	/**
	 * Returns the current policy
	 * 
	 * @return the policy type
	 */
	public int getPolicy( ) 
	{
		return m_iPolicy;
	}
	
	/**
	 * Finds an Event object in the list of current events
	 * by a given event identifier
	 * 
	 * @param szEventIdentifier
	 * 
	 * @return The event with the given identifier
	 */
	public Event findEvent( String szEventIdentifier ) {
		for ( int i = 0; i < m_Events.size( ); i++ ) {
			Event event     = ( Event )m_Events.get( i );
			if ( event.hasIdentifier( szEventIdentifier ) == true ) {
				 return event;
			}
		}
		
		return null;
	}
	
	/**
	 * Gets the rectangular region where the area is located
	 * 
	 * @return a region of type Rectangle
	 */
	public Rectangle getRectangle( ) {
		return new Rectangle( m_Rect );
	}
	
	/**
	 * Returns the 'Z-Index' of the area
	 * 
	 * @return the Z Index
	 */
	public int getZIndex( ) {
		return m_iZIndex;
	}
	
	/**
	 * Returns the 'Z-Index' of the area
	 * 
	 * @return the Z Index
	 */
	public boolean setZIndex( int iZIndex ) {
		if ( iZIndex >= 0 ) {
			m_iZIndex = iZIndex;
			return true;
		} 
		return false;
	}	
	
	/**
	 * Returns the value of information pair with a given name
	 * 
	 * @param szName - the information pair name
	 * 
	 * @return the information pair value or null if it's not matched
	 */
	public String getInfo( String szName ) {
		return ( String )m_Info.get( szName );
	}

	/**
	 * Checks if there's an information pair (Name, Value)
	 * 
	 * @param szName	- the name of the pair
	 * @param szValue	- the value of the pair
	 * 
	 * @return true if there's exactly the same information
	 */	
	public boolean hasInfo( String szName, String szValue ) {
		return m_Info.containsKey( szName ) == true &&
		       ( ( String )m_Info.get( szName ) ).equals( szValue );
	}
	
	/**
	 * Returns the value of inofrmation pair which value
	 * contains a given string and which name matches a given name
	 * 
	 * @param szName	- the name of the information pair
	 * @param szMatch - the string to check if it's contained
	 *                  inside the value of the matched name pair
	 * @param bCaseSensitive - case sensitive search or not
	 * 
	 * @return if the info is matched returns true. Otherwise - false
	 */
	public boolean hasInfoLike( String szName, String szMatch, boolean bCaseSensitive ) {
		if ( m_Info.containsKey( szName ) == true ) {
			String szValue = ( String )m_Info.get( szName );
			if ( bCaseSensitive ) {
				return szValue.indexOf( szMatch ) != -1;
			} else {
				return szValue.toLowerCase( ).indexOf( szMatch.toLowerCase( ) ) != -1;
			}
		}
		return false;
	}
	
	/**
	 * Returns a debugging information about the area
	 */
	public String toString( ) {
		String szInfo = "Area rectangle( " + m_Rect.x + ", " + m_Rect.y + ", " + 
		                                     m_Rect.width + ", " + m_Rect.height + " )" + 
		                 " Z Index = " + m_iZIndex;
		szInfo += "\nInformation:\n";
		
		Enumeration infoKeys = m_Info.keys( );
		while ( infoKeys.hasMoreElements( ) ) {
			String szKey   = ( String )infoKeys.nextElement( );
			String szValue = ( String )m_Info.get( szKey );
			
			szInfo += "\t[" + szKey + "] = [" + szValue + "]\n"; 
		}
		
		szInfo += "Events:\n";
		for ( int i = 0; i < m_Events.size( ); i++ ) {
			Event event = ( Event )m_Events.get( i );
			szInfo += "\t[" + event.getIdentifier( ) + "]" + 
                   " = [" + event.getAction( ) + "]\n";
		}
		
		return szInfo;
	}

	/**
	 * Generates a unique ID for the area
	 */
	private void setUniqueID( )
	{
		m_lUniqueID = ( long )( Math.random( ) * 1000 ) + System.currentTimeMillis( ) * 1000;
	}
	

}
