package lib.map.dynamic;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Enumeration;
import sun.misc.Compare;
import sun.misc.Sort;

import lib.map.base.Rectangle;
import lib.exceptions.AreaException;


/**
 * @class lib.map.dynamic.AreaManager
 * 
 * Manages the following operations for a number of 'Areas':
 * - Adding of area objects to a list
 * - Finding an area by a (x, y) position
 * - Finding an area by a given information
 * - Finding an area by a given event identifier
 * - Finding an area by unique ID
 * 
 * @author Tihomir Lazarov
 * @date   Aug 16, 2004
 */
public class AreaManager {
	
	private ArrayList m_Areas = null; // The array that holds the area objects
	
	/**
	 * Constructs an AreaManager object
	 */
	public AreaManager( ) {
		m_Areas = new ArrayList( );
	}

	/**
	 * Adds an area to the list
	 * 
	 * @param area - the area to add
	 * 
	 * @throws AreaException - when the zIndex and coordinates overlap with
	 *                         existing areas
	 */	
	public void addArea( Area area ) 
		throws AreaException 
	{
		/* Find all areas that contain the four points of the rectangle
		 * of the currently added area. If there's even a single area
		 * that overlaps with the current area and is on the same ZIndex
		 * there should be an exception
		 */
		int i;
		Rectangle rect            = area.getRectangle( );
		ArrayList allMatchedAreas = findAreasByRectangle( rect );

		if ( allMatchedAreas != null ) {
			for ( i = 0; i < allMatchedAreas.size( ); i++ ) {
				Area matchedArea = ( Area )allMatchedAreas.get( i );
				if ( matchedArea.getZIndex( ) == area.getZIndex( ) ) {
					throw new AreaException( "Trying to add an overlapping area:\n" + 
													 area.toString( ) + "\n\n" + 
													 "The area that already exists and was " + 
													 "about to be overlapped is:\n" + 
													 matchedArea.toString( ) );
				}
			}
		}
		
		m_Areas.add( area );
	}
	
	/**
	 * Adds an area to the list. Automatically gets the zIndex
	 * that will be 0 or 1 point higher than the max zIndex of
	 * the area that overlaps with the coordinates of the current
	 * area. 
	 * 
	 * @param area - the area to add
	 * 
	 * @throws AreaException
	 */	
	public void pushArea( Area area ) 
		throws AreaException 
	{
		int i;
		Rectangle rect            = area.getRectangle( );
		ArrayList allMatchedAreas = findAreasByRectangle( rect );
		int iZIndex               = 0;
		
		if ( allMatchedAreas != null ) {
			System.out.println( "pushArea: Trying to add an area that overlaps with " + allMatchedAreas.size() + " area(s)");
			/* Find the greatest zindex */
			for ( i = 0; i < allMatchedAreas.size( ); i++ ) {
				Area matchedArea = ( Area )allMatchedAreas.get( i );
				int iThisZIndex  = matchedArea.getZIndex( );
				if ( iThisZIndex >= iZIndex ) {
					iZIndex = iThisZIndex;
				}
			}
			iZIndex++;
		} else {
			System.out.println( "No overlapping area found for rectangle: " + rect.toString( ) );
		}
		
		System.out.println( "pushArea: zIndex estimated: " + iZIndex );
		
		area.setZIndex( iZIndex );
		m_Areas.add( area );
	}
	
	/**
	 * Removes the topmost area found on position (x, y)
	 * 
	 * @param x - the X position
	 * @param y - the Y position
	 */
	public void removeAreaByPosition( int x, int y ) 
	{
		ArrayList areas = findAreasByPosition( x, y );
		if ( areas != null ) {
			System.out.println( "Found area to remove" );
			Area area = ( Area )areas.get( areas.size( ) - 1 );
			if ( area.getPolicy( ) == Area.POLICY_REMOVABLE ) {
				System.out.println( "Removed area" );
				m_Areas.remove( area );
			} else {
				System.out.println( "Area is NOT removable: " + area.toString( ) );
			}
		} else {
			System.out.println( "No area found for removal" );
		}
	}
	
	/**
	 * Returns the whole list of areas
	 * 
	 * @return an ArrayList of Area objects
	 */
	public ArrayList getAreas( ) 
	{
		return ( ArrayList )m_Areas.clone( ); 	
	}
	
	/**
	 * Removes all areas from the list
	 */
	public void clear( )
	{
		m_Areas.clear( );
	}

	/**
	 * Finds areas by a (x, y) position. The given
	 * (x, y) point should be inside or on the border of the
	 * rectangle, described by the areas
	 * 
	 * @param x - the X position
	 * @param y - the Y position
	 * 
	 * @return a sorted ArrayList of Area objects in ascending order by zIndex 
	 *         that contain the given (x, y) point. If no area matches
	 *         a 'null' is returned
	 */	
	public ArrayList findAreasByPosition( int x, int y ) 
	{
		ArrayList matchedAreas = new ArrayList( );
		
		for ( int i = 0; i < m_Areas.size( ); i++ ) {
			Area area = ( Area )m_Areas.get( i ); 
			if ( isInRectangle( x, y, area.getRectangle( ) ) ) {
				matchedAreas.add( area );
			}
		}
		
		if ( matchedAreas.size( ) == 0 ) {
			return null;
		} else {
			return sortAreasByZIndex( matchedAreas );
		}
	}
	
	/**
	 * Finds all areas that have any of their corners inside
	 * a given rectangular area
	 * 
	 * @param r - the rectangular area
	 * 
	 * @return and ArrayList of all areas
	 */
	public ArrayList findAreasByRectangle( Rectangle r ) 
	{
		ArrayList areas 	= new ArrayList( );
		
		for ( int i = 0; i < m_Areas.size( ); i++ ) {
			Area area = ( Area )m_Areas.get( i );
			Rectangle areaRect = area.getRectangle( );
			
			if ( haveCommonParts( areaRect, r ) ) {
				areas.add( area );
			}
		}
		return areas.size( ) == 0 ? null : areas;
	}
	
	/**
	 * Removes the area that has a given unique ID
	 * 
	 * @param lUniqueID
	 */
	public void removeAreaByUniqueID( long lUniqueID ) 
	{
		for ( int i = 0; i < m_Areas.size( ); i++ ) {
			Area area = ( Area )m_Areas.get( i );
		
			if ( area.getUniqueID( ) == lUniqueID ) {
				m_Areas.remove( i );
				break;
			}
		}
	}

	/**
	 * Searches for an area with a given unique ID. Does a
	 * linear search by unique ID as this operation happens
	 * rarely in the time and it's not a ROI if we're using a
	 * hashtable of all unique IDs.
	 * 
	 * @param lUniqueID - the unique ID to search for
	 * 
	 * @return the area with the given unique ID or null if no
	 * 		  such area is found
	 */	
	public Area findAreaByUniqueID( long lUniqueID )
	{
		for ( int i = 0; i < m_Areas.size( ); i++ ) {
			Area area = ( Area )m_Areas.get( i );
			if ( area.getUniqueID( ) == lUniqueID ) {
				return area;
			}
		}
		return null;
	}
	
	/**
	 * Finds areas by given criteria. The criteria is a hash of 
	 * ( name, value ) pairs which is matched against the information 
	 * stored inside the Area object. The area that matches at least
	 * one of the criteria is added to the list of matched areas.
	 * The flag 'exact match' shows if the value of the information
	 * pair should be exactly matched or only to check if the given
	 * value in the criteria is contained within any value in the
	 * information pairs for the Area.
	 * 
	 * @param criteria		 - a hash table with criteria. The criteria
	 *                         are pairs of Name, Value, where the Name
	 *                         is exactly matched against the name of a 
	 *                         an information pair in the Area.
	 * @param bExactMatch	 - if the value should be exactly matched
	 * @param bCaseSensitive - if the search to be case sensitive. If
	 *                         the bExactMatch is true, this argument is
	 *                         ignored
	 * 
	 * @return an ArrayList of matched areas. Otherwise returns 'null'
	 */
	public ArrayList findAreasByInfo( Hashtable criteria,
	                                  boolean bExactMatch,
	                                  boolean bCaseSensitive ) 
	{
		ArrayList matchedAreas = new ArrayList( );
		
		for ( int i = 0; i < m_Areas.size( ); i++ ) {
			Area area = ( Area )m_Areas.get( i );
			
			Enumeration keys = criteria.keys( );
			while ( keys.hasMoreElements( ) ) {
				String szKey   = ( String )keys.nextElement( );
				String szValue = ( String )criteria.get( szKey );
				
				/* Check if at least one element from the crietria is matched */
				if ( bExactMatch == true ) {
					if ( area.hasInfo( szKey, szValue ) == true ) {
						matchedAreas.add( area );
						break;
					}
				} else {
					if ( area.hasInfoLike( szKey, szValue, bCaseSensitive ) == true ) {
						matchedAreas.add( area );
						break;
					}
				}
			}
		}
		
		if ( matchedAreas.size( ) == 0 ) {
			return null;
		} else {
			return sortAreasByZIndex( matchedAreas );
		}
	}
	
	/**
	 * Finds an area by event identifier. The topmost area that matches
	 * the given event identifier is returned
	 * 
	 * @param x	- the X position to search areas for
	 * @param y	- the Y position to search areas for
	 * @param szEventIdentifier - the event identifier to match
	 * 
	 * @return the area that matches the given identifier at position (x, y)
	 *         or null otherwise
	 */
	public Area findAreaByEventIdentifier( int x, int y, String szEventIdentifier ) 
	{
		ArrayList areasOnSpot = findAreasByPosition( x, y );
		
		if ( areasOnSpot != null ) {
			System.out.println( "Found " + areasOnSpot.size( ) +" areas on spot (" + x + ", " + y + "). " + 
			                    "Checking for event identifier [" + szEventIdentifier + "]" );
	
			/* Here we assume that the areasOnSpot are sorted by ZIndex starting
			 * from ZIndex 0.
			 */
			for ( int i = areasOnSpot.size( ) - 1; i >= 0; i-- ) {
				Area area 	= ( Area )areasOnSpot.get( i );
				/*
				 System.out.println( "Checking if area matches identifier [" + 
				                     szEventIdentifier + "], area [" + area.toString() + "]" );
				*/
				Event event	= area.findEvent( szEventIdentifier ); 
				if ( event != null ) {
					return area;
				}
			}
		} else {
			System.out.println( "No areas found on spot (" + x + ", " + y + "), identifier [" + szEventIdentifier + "]" );
		}
		
		return null;
	}
	
	/* --- Private methods ------------------------------------------------- */
	/**
	 * Sorts a list of areas by zIndex
	 * 
	 * @param areas - an ArrayList of Area objects
	 * 
	 * @return a sorted by zIndex ArrayList of Area objects
	 */
	private ArrayList sortAreasByZIndex( ArrayList areas ) 
	{
		AreasComparatorByZIndex c = new AreasComparatorByZIndex( );
		Object[] areasArray  = areas.toArray( );
		
		/* Make a quick sort in the areas array using the 
		 * AreasComparatorByZIndex to compare pair of elements
		 */
		Sort.quicksort( areasArray, c );
		
		/* Return the array of areas into an ArrayList object */
		areas.clear( );
		for ( int i = 0; i < areasArray.length; i++ ) {
			areas.add( areasArray[ i ] );
		}
		return areas;
	}
	
	/**
	 * Checks if a point is placed in a given rectangular area
	 * 
	 * @param x		- point X coordinate
	 * @param y		- point Y coordinate
	 * @param rect	- rectangle
	 * 
	 * @return true if the point (x, y) is located inside the rectangle or
	 *         on some border
	 */
	private boolean isInRectangle( int x, int y, Rectangle rect ) 
	{
		return ( x >= rect.x && x <= rect.x + rect.width && 
				   y >= rect.y && y <= rect.y + rect.height );
	}
	
	/**
	 * Checks if two rectangles overlap each other
	 * 
	 * @param r1 - the first rectangle
	 * @param r2 - the second rectangle
	 * 
	 * @return bolean value
	 */
	private boolean haveCommonParts( Rectangle r1, Rectangle r2 ) 
	{
		if ( crossesRectangle( r1.x, r1.y, r1.x + r1.width, r1.y, r2 ) ||  /* Top line  */
			  crossesRectangle( r1.x, r1.y, r1.x, r1.y + r1.height, r2 ) || /* Left line */
			  crossesRectangle( r1.x + r1.width, r1.y, r1.x + r1.width, r1.y + r1.height, r2 ) || /* Right line */
			  crossesRectangle( r1.x, r1.y + r1.height, r1.x + r1.width, r1.y + r1.height, r2 ) || /* Bottom line */
			  insideRectangle( r1, r2 ) ||
			  insideRectangle( r2, r1 )
		   ) {
			return true;
		}
		return false;
	}
	
	/**
	 * Checks if a given line crosses a given rectangle
	 * 
	 * @param x1 - line start X position
	 * @param y1 - line start Y position
	 * @param x2 - line end X position
	 * @param y2 - line end Y position
	 * @param rect - the rectangle
	 * 
	 * @return boolean value
	 */
	private boolean crossesRectangle( int x1, int y1, int x2, int y2, Rectangle rect ) {
		
		if ( y1 == y2 ) { /* Horizontal line */
			if ( x1 > x2 ) {
				int temp = x1;
				x1 		= x2;
				x2 		= temp;
			}
			/* Check if the line is between the rectangle's top and bottom lines */
			if ( rect.y <= y1 && rect.y + rect.height >= y1 ) {
				/* Check if the line's ends are 
				 * o From the both sides of the left rectangle wall
				 *   or
				 * o From the both sides of the right rectangle wall
				 */
				if ( ( x1 <= rect.x && x2 >= rect.x ) ||
				     ( x1 <= rect.x + rect.width && x2 >= rect.x + rect.width ) ||
				     ( x1 >= rect.x && x2 <= rect.x + rect.width ) ) {
					return true; 
				}
			} 
		} else if ( x1 == x2 ) { /* Vertical line */
			if ( y1 > y2 ) {
				int temp	= y1;
				y1			= y2;
				y2			= temp;
			}
			/* Check if the line is between the left and 
			 * right sides of the rectangle
			 */
			if ( rect.x <= x1 && rect.x + rect.width >= x1 ) {
				/* Check if the line's ends are 
				 * o From the both sides of the top rectangle wall
				 *   or
				 * o From the both sides of the bottom rectangle wall
				 *   or
				 * o In the inside of the rectangle
				 */
				if ( ( y1 <= rect.y && y2 >= rect.y ) || 
				     ( y1 <= rect.y + rect.height && y2 >= rect.y + rect.height ) || 
				     ( y1 >= rect.y && y2 <= rect.y + rect.height ) ) {
					return true;
				}
			}
		}
		
		return false;
	}
	
	/**
	 * Checks if r1 is inside r2
	 * 
	 * @param r1	- first rectangular area
	 * @param r2	- second rectangular area
	 * 
	 * @return if r1 is inside r2 returns true. Otherwise returns false
	 */
	private boolean insideRectangle( Rectangle r1, Rectangle r2 )
	{
		return ( 
			r1.x >= r2.x && r1.x + r1.width <= r2.x + r2.width &&
			r1.y >= r2.y && r1.y + r1.height <= r2.y + r2.height 
		);
	}

	
	/* --- Private classes ------------------------------------------------- */
	/**
	 * @class lib.map.dynamic.AreaManager.AreasComparatorByZIndex
	 * 
	 * This class is used for sorting a list of areas by ZIndex
	 * 
	 * @author Tihomir Lazarov
	 * @date   Aug 29, 2004
	 */
	private class AreasComparatorByZIndex implements Compare {
		/**
		 * Compares two area objects' zIndex properties
		 * 
		 * @param object1 - the first area
		 * @param object2 - the second area
		 * 
		 * @return -1 if the first zIndex < second zIndex,
		 *          0 if they are equal and 1 if the first
		 *          zIndex is bigger
		 */		
		public int doCompare( Object object1, Object object2 ) {
			Area area1 = ( Area )object1;
			Area area2 = ( Area )object2;
			if ( area1.getZIndex() < area2.getZIndex( ) ) {
				return -1;
			}
			if ( area1.getZIndex( ) > area2.getZIndex( ) ) {
				return 1;
			}
			return 0;
		}
	}
	
}
