package lib.map.dynamic;

import java.util.ArrayList;
import java.util.Hashtable;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Stroke;

import lib.cfg.Constants;
import lib.cfg.Config;
import lib.comm.MapDefinition;
import lib.db.DBManager;
import lib.map.base.Map;
import lib.map.base.Rectangle;
import lib.map.dynamic.Icons.Icon;
import lib.map.dynamic.cfg.DynamicMapConfig;
import lib.map.dynamic.cfg.DynamicMapEnvironment;
import lib.road.ext.DBRoadGraph;
import lib.road.ConnectionPoint;
import lib.road.RoadGraph;
import lib.road.RoadPoint;
import lib.exceptions.DynamicMapException;
import lib.exceptions.MapStateException;
import lib.exceptions.AreaException;
import lib.exceptions.NullMapException;
import lib.exceptions.DynamicMapConfigException;
import lib.exceptions.DBRoadGraphException;

/**
 * @class lib.map.dynamic.DynamicMap
 * 
 * An extention ot the lib.map.base.Map that holds rectangular
 * locations on thepushArea map called Areas. The navigaion is managed
 * through events over these areas 
 * 
 * @author Tihomir Lazarov
 * @date   Aug 15, 2004
 */
public class DynamicMap extends Map {
	
	int 			m_iViewerWidth					= 0;
	int 			m_iViewerHeight				= 0;
	AreaManager m_AreaManager					= null;
	DBManager	m_DBManager						= null;
	Config		m_Config							= null;
	Icons			m_Icons							= null;
	int			m_iMapID							= 0;
	RoadGraph   m_RoadGraph						= null;
	
	// Parameters that are used when applying search criteria over a snapshot
	boolean 		m_bShowGraph					= false;
	String		m_szSearchQuery				= "";
	ArrayList   m_AttrToSearchIn				= null;
	ArrayList   m_SearchIconIDs				= null;
	Hashtable	m_ShortestPathCache			= null;
	long 			m_lAreaID1						= 0;
	long			m_lAreaID2						= 0;
	
	/**
	 * Creates a DynamicMap object
	 * 
	 * @param env - DynamicMapEnvironment object instance 
	 * @param iMapID - the ID of the map to load (-1 loads the default map)
	 * 
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws MapStateException
	 */
	public DynamicMap( DynamicMapEnvironment env, int iMapID ) 
		throws DynamicMapException 
	{
		init( env, iMapID );
	}
	
   /**
	  * Returns the current map environment
	  */
	public DynamicMapEnvironment getEnvironment( )
	{
		return new DynamicMapEnvironment( m_DBManager, m_Config, m_Icons, m_iViewerWidth, m_iViewerHeight );
	}
	
	/**
	 * Loads a map and the areas that are over it
	 * 
	 * @param szDynamicMapDefinition - the XML file that contains the 
	 * 										  dynamic map definition
	 * 
	 * @throws DynamicMapException
	 */
	public void load( int iMapID ) 
		throws DynamicMapException
	{
		try {
			super.load( m_DBManager, m_Config, iMapID, m_iViewerWidth, m_iViewerHeight );
			m_iMapID = getMapID();
			
			DynamicMapConfig config = new DynamicMapConfig( m_DBManager, m_iMapID );
			
			// Load the list of areas on the map
			ArrayList areas = config.getAreas( );
			m_AreaManager.clear( );
			System.out.println( "Loading areas for MapID " + iMapID );
			for ( int i = 0; i < areas.size( ); i++ ) {
				Area area = ( Area )areas.get( i );
				System.out.println( "******************\nLoading area: " + area.toString( ) );
				m_AreaManager.addArea( area );
			}
			System.out.println( "Areas loaded" );
			
			// Load the graph of roads on the map
			m_ShortestPathCache 		= new Hashtable( );
			DBRoadGraph dbRoadGraph = new DBRoadGraph( m_DBManager );
			
			// There can be many road graphs on the map. Currently we will be using
			// just one of them
			ArrayList roadGraphs		= dbRoadGraph.load( m_iMapID );
			if ( roadGraphs != null ) {
				m_RoadGraph	= ( RoadGraph ) roadGraphs.get( 0 );
			}
		} catch ( MapStateException e ) {
			throw new DynamicMapException( "Map state exception: " + e.getMessage( ), e );
		} catch ( AreaException e ) {
			throw new DynamicMapException( "Area exception: " + e.getMessage( ), e );
		} catch ( DynamicMapConfigException e ) {
			throw new DynamicMapException( "Dynamic map config exception: " + e.getMessage( ), e );
		} catch ( DBRoadGraphException e ) {
			throw new DynamicMapException( "Road graph exception: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Gets all areas on the map
	 * 
	 * @return an ArrayList of Area objects
	 */
	public ArrayList getAreas( ) 
	{
		return m_AreaManager.getAreas( );
	}
	
	/**
	 * Sets the query to search for. This search is applied
	 * before returning the map that is currently displayed
	 * in the viewer. This query is passed to the area manager
	 * which returns the areas that match this query and the
	 * list of attributes to search in 
	 * 
	 * @param szQuery - the query string 
	 * @param attrToSearchIn - the ArrayList of string attributes to search in 
	 */
	public void searchByInfo( String szQuery, ArrayList attrToSearchIn ) 
	{
		m_szSearchQuery 	= szQuery;
		m_AttrToSearchIn	= new ArrayList( attrToSearchIn );
	}
	
	/**
	 * Resets the search query
	 */
	public void resetSearchQuery( )
	{
		m_szSearchQuery 	= "";
		m_AttrToSearchIn	= null;
	}
	
	
	/**
	 * Sets search by an ArrayList of integer values that
	 * represent icon IDs
	 * 
	 * @param iconIDs - the icon IDs ArrayList 
	 */
	public void searchByIcons( ArrayList iconIDs )
	{
		m_SearchIconIDs = new ArrayList( iconIDs );
	}
	
	/**
	 * Resets the search by icons IDs
	 */
	public void resetSearchByIcons( )
	{
		m_SearchIconIDs = null;
	}
	
	/**
	 * Sets the flag for showing or not the graph when returing
	 * a snapshot
	 * 
	 * @param bShowGraph - true to show the graph, false otherwise
	 */	
	public void showGraph( boolean bShowGraph )
	{
		m_bShowGraph = bShowGraph;
	}
	
	/**
	 * Schedules two area IDs for later search. A search for a shortest
	 * path between those two areas is made.
	 * 
	 * @param lAreaID1 - one of the area IDs
	 * @param lAreaID2 - the other area ID
	 */
	public void searchShortestPath( long lAreaID1, long lAreaID2 )
	{
		m_lAreaID1	= lAreaID1;
		m_lAreaID2	= lAreaID2;
	}
	
	/**
	 * Resets the shortest path search criteria
	 */
	public void resetShortestPathSearch( )
	{
		m_ShortestPathCache.clear( );
		m_lAreaID1				= 0;
		m_lAreaID2				= 0;
	}
	
	
	/**
	 * Saves the current dynamic map state in an XML file
	 * 
	 * @throws DynamicMapException
	 */
	public void save( ) 
		throws DynamicMapException
	{
		try {
			DynamicMapConfig config = new DynamicMapConfig( m_DBManager, m_iMapID );
			config.clearAreas( );
			
			ArrayList areas = m_AreaManager.getAreas( );
			for ( int i = 0; i < areas.size( ); i++ ) {
				config.addArea( ( Area )areas.get( i ) );
			}
		} catch ( DynamicMapConfigException e ) {
			throw new DynamicMapException( "Dynamic map config exception: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Reloads the current dynamic map
	 * 
	 * @throws DynamicMapException
	 */
	public void reload( ) throws DynamicMapException
	{
		load( m_iMapID );
	}
	
	/**
	 * Checks if the map is zoomable in
	 * 
	 * @return boolean value
	 * 
	 * @throws DynamicMapException
	 */
	public boolean isZoomableIn( ) throws DynamicMapException 
	{
		try {
			return isZoomable( Map.ZOOM_IN );
		} catch ( MapStateException e ) {
			throw new DynamicMapException( e.getMessage( ), e );
		}
	}
	
	/**
	 * Checks if the map is zoomable out
	 * 
	 * @return boolean value
	 * 
	 * @throws DynamicMapException
	 */
	public boolean isZoomableOut( ) throws DynamicMapException 
	{
		try {
			return isZoomable( Map.ZOOM_OUT );
		} catch ( MapStateException e ) {
			throw new DynamicMapException( e.getMessage( ), e );
		}
	}
	
	/**
	 * Returns a snapshot of the current viewer window and draws
	 * the areas over the snapshot
	 * 
	 * @return an object of type Snapshot
	 * 
	 * @throws NullMapException
	 */
	public Snapshot getSnapshot( ) throws NullMapException 
	{
		BufferedImage image 				= getViewableRegionSnapshot( );
		
		/* Draw the areas that are visible */
		Rectangle viewer 			 		= getViewerRegion( );
		Rectangle viewerRealSize 		= scaleToRealSize( viewer );
		Snapshot snapshot 				= new Snapshot( );
		
		ArrayList allAreas 				= m_AreaManager.findAreasByRectangle( viewerRealSize );
		ArrayList filteredAreas 		= applyAreasSearchFilter( allAreas );
		
		if ( m_bShowGraph ) {
			drawGraph( image );
		} else {
			snapshot.roadLength = applyShortestPathSearch( image, allAreas );
		}
		
		if ( filteredAreas != null ) {
			snapshot.areas = new Area[ filteredAreas.size( ) ];
			for ( int i = 0; i < filteredAreas.size( ); i++ ) {
				Area area 				= ( Area )filteredAreas.get( i );
				image 					= drawArea( image, area );
				snapshot.areas[ i ] 	= area; 
			}
		}
		
		snapshot.image  = image;
		return snapshot;	
	}
	
 
	/**
	 * Searches for an area by its EventIdentifier within all
	 * areas that are on (x, y) position relative to the maximum
	 * zoom (real size)
	 * 
	 * @param x	- X position
	 * @param y - Y position
	 * @param szEventIdentifier - the event identifier to search for
	 * 
	 * @return the found Area or null if no area is found
	 */
	public Area findAreaByEventIdentifier( int x, int y, String szEventIdentifier ) 
	{
		Rectangle viewer = getViewerRegion( );
		viewer.x += x;
		viewer.y += y;
		viewer = scaleToRealSize( viewer );
		return m_AreaManager.findAreaByEventIdentifier( viewer.x, viewer.y, szEventIdentifier );
	}
	
	/**
	 * Removes an area by a given unique ID
	 * 
	 * @param lUniqueID - the ID to find the area by
	 */
	public void removeAreaByUniqueID( long lUniqueID ) 
	{
		m_AreaManager.removeAreaByUniqueID( lUniqueID );	
	}
	
	/**
	 * Adds an area to the dynamic map
	 * 
	 * @param area - the Area object instance that hodls the area deinition
	 * 
	 * @throws DynamicMapException
	 */
	public void pushArea( Area area ) throws DynamicMapException 
	{
		try {
			m_AreaManager.pushArea( area );
		} catch ( AreaException e ) {
			throw new DynamicMapException( "Failed adding an area: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Removes the topmost area that's located on real size (x, y)
	 * position and do not have a policy UNREMOVABLE
	 * 
	 * @param x - the X position
	 * @param y - the Y position
	 */
	public void removeAreaByZoomPosition( int x, int y ) 
	{
		Rectangle viewer = getViewerRegion( );
		viewer.x += x;
		viewer.y += y;
		Rectangle rect = scaleToRealSize( viewer );
		m_AreaManager.removeAreaByPosition( rect.x, rect.y );
	}

	/* --- Private methods ------------------------------------------------- */
	
	/**
	 * Initializes the dynamic map environment and loads a map
	 * 
	 * @param env - an object of type DynamicMapEnvironment
	 * @param iMapID - the ID of the map to load (-1 loads default map)
	 */
	private void init( DynamicMapEnvironment env, int iMapID )
		throws DynamicMapException
	{
		m_DBManager			= env.getDBManager( );
		m_Config				= env.getConfig( );
		m_Icons				= env.getIcons( );
		m_iViewerWidth		= env.getViewerWidth( );
		m_iViewerHeight	= env.getViewerHeight( );
		m_AreaManager		= new AreaManager( );
		
		// If the viewer dimensions are not passed, get 
		// them frmo the global configuration
		if ( m_iViewerWidth == 0 || m_iViewerHeight == 0 ) {
			m_iViewerWidth		= Integer.parseInt( env.getConfig( ).getValue( Constants.PARAM_VIEWER_WIDTH  ) );
			m_iViewerHeight 	= Integer.parseInt( env.getConfig( ).getValue( Constants.PARAM_VIEWER_HEIGHT ) );
			
			System.out.println( "No viewer dimensions found. " + 
			                    "Loading default (" + m_iViewerWidth + "x" + m_iViewerHeight + ")" );
		}
		
		System.out.println( "Viewer dimentions (" + m_iViewerWidth + "x" + m_iViewerHeight + ")" ); 
		
		/* Load the dynamic map definition */
		load( iMapID );
	}
	
	/**
	 * Adds an array list to a nother array list
	 * 
	 * @param dest - destination array list
	 * @param src  - source array list
	 */
	private ArrayList addArrayList( ArrayList dest, ArrayList src ) 
	{
		if ( src != null ) { 
			for ( int i = 0; i < src.size( ); i++ ) {
				dest.add( src.get( i ) );
			}
		}
		return dest;
	}
	
	/**
	 * Filters a given list of areas by applying the current search query
	 * 
	 * @param areas
	 * 
	 * @return the list of filtered areas or null if the list
	 *         of filtered areas produced an empty result
	 */
	private ArrayList applyAreasSearchFilter( ArrayList areas )
	{
		//	Apply the search criteria by query string
		ArrayList filteredAreas 		= new ArrayList( );
		ArrayList infoMatchingAreas 	= null;
		ArrayList iconMatchingAreas	= null;		
		boolean bHasSearchFilter		= m_szSearchQuery.equals( "" ) == false ||
		                                m_SearchIconIDs != null;
		boolean bSearchByQuery			= false;
		boolean bSearchByIcon			= false;
		Hashtable criteria = new Hashtable( );
				                                
		if ( m_szSearchQuery.equals( "" ) == false ) {
			bSearchByQuery = true;
			
			System.out.println( "APPLY SEARCH: Searching by query '" + m_szSearchQuery + "'..." );
			if ( m_AttrToSearchIn != null ) {
				for ( int i = 0; i < m_AttrToSearchIn.size( ); i++ ) {
					criteria.put( ( String ) m_AttrToSearchIn.get( i ), m_szSearchQuery );
				}
			}
			infoMatchingAreas = m_AreaManager.findAreasByInfo( criteria, false, false );
			if ( infoMatchingAreas != null ) {
				System.out.println( "APPLY SEARCH: Searching by query found " + infoMatchingAreas.size( ) + " results" );
			} else {
				System.out.println( "APPLY SEARCH: Searching by query did NOT found any results" );
			}
		} 
		
		if ( m_SearchIconIDs != null ) {
			bSearchByIcon = true;
			for ( int i = 0; i < m_SearchIconIDs.size( ); i++ ) {
				int iIconID = ( ( Integer ) m_SearchIconIDs.get( i ) ).intValue( );
				System.out.println( "APPLY SEARCH: Searching by icon ID " + iIconID + "..." );
				criteria.put( Constants.AREA_ATTRIBUTE_ICON_ID, "" + iIconID );
				iconMatchingAreas = m_AreaManager.findAreasByInfo( criteria, false, false );
				if ( iconMatchingAreas != null ) {
					System.out.println( "APPLY SEARCH: Searching by icon ID  matched " + 
											  iconMatchingAreas.size( ) + " results" );
				} else {
					System.out.println( "APPLY SEARCH: Search by icon ID did NOT gave any results" );
				}
			}
		} else {
			System.out.println( "APPLY SEARCH: No search by icon ID" );
		}
		
		// Merge the icon/info matching arrays with the elements that are
		// in both arrays if both search criteria were applied - by info and by icon.
		// Otherwise match only the array that was searched for
		for ( int i = 0; i < areas.size( ); i++ ) {
			Area regionArea = ( Area ) areas.get( i );
			if ( bSearchByQuery == true && bSearchByIcon == false && 
			     infoMatchingAreas != null && infoMatchingAreas.contains( regionArea ) ) {
				filteredAreas.add( regionArea );
			} else if ( bSearchByQuery == false && bSearchByIcon == true &&
			            iconMatchingAreas != null && iconMatchingAreas.contains( regionArea ) ) {
				filteredAreas.add( regionArea );            	
			} else if ( bSearchByQuery == true && bSearchByIcon == true &&
			            iconMatchingAreas != null && iconMatchingAreas.contains( regionArea ) &&
			            infoMatchingAreas != null && infoMatchingAreas.contains( regionArea ) ) {
				filteredAreas.add( regionArea );
			}
		}
		
		if ( ! bHasSearchFilter ) {
			// No search filter available
			filteredAreas = areas;
		}
		
		return filteredAreas.size( ) > 0 ? filteredAreas : null;		
	}
	
	/**
	 * Applies a shortest path search and draws the shortest path over the given
	 * image, representing the current viewer region. If the shotest path search
	 * is the same like the last shortest path search, a cached graph road is used
	 * without having to search again for the shortest path.
	 * 
	 * @param image - the viewer region to draw on (the image is modified!)
	 * @param allAreas - the list of all areas on the map
	 * 
	 * @return the road length in meters (using the map scale property)
	 */
	private long applyShortestPathSearch( BufferedImage image, ArrayList allAreas )
	{
		if ( m_RoadGraph == null ) {
			System.out.println( "SHORTEST PATH: Road graph is not initialized" );
			return 0;
		} 
		
		if ( m_lAreaID1 == 0 || m_lAreaID2 == 0 ) {
			System.out.println( "SHORTEST PATH: Not valid area IDs: " + 
			                    "Area 1 (" + m_lAreaID1 + "), Area 2 (" + m_lAreaID2 + ")" );
			return 0;
		} 
		
		System.out.println( "SHORTEST PATH: Applying shortest path search..." );
		
		// The last shortest path may be in the cache in the following form:
		//	 areaIDsString => array list of RoadPoint points
		String szCacheKey		= getShortestPathCacheKey( m_lAreaID1, m_lAreaID2 );  
		ArrayList lastRoad 	= ( ArrayList ) m_ShortestPathCache.get( szCacheKey );
		
		System.out.println( "SHORTEST PATH: Searching for old road for key: " + szCacheKey );
		
		// Check if there was a cache hit and draw the road. Otherwise
		// must search for the road again.
		if ( lastRoad == null ) {
			System.out.println( "SHORTEST PATH: No" );
			ArrayList road = getShortestPath( m_lAreaID1, m_lAreaID2 );
			
			System.out.println( "SHORTEST PATH: Found road between " + m_lAreaID1 + " and " + 
                             m_lAreaID2 );
			System.out.println( "SHORTEST PATH: Road size: " + road.size( ) + " points" );
			m_ShortestPathCache.put( szCacheKey, road );
			lastRoad = road; 
		} else {
			System.out.println( "SHORTEST PATH: Yes. Road size: " + lastRoad.size( ) + " points" );
		}
		
		// Draw the road over the map
		drawRoad( image, lastRoad, null );
		
		return calculateRealRoadLength( lastRoad );
	}
	
	/**
	 * Calculates the distance of a road using the map scale and
	 * a given array list of road points.
	 * 
	 * @param road - the array list of road points
	 * 
	 * @return the length of the road in meters
	 */
	private long calculateRealRoadLength( ArrayList road )
	{
		long lPixelsLength 	= road == null || road.size( ) <= 1 ? 0 :
		                         m_RoadGraph.getRoadLength( road );
		MapDefinition def		= getMapDefinition( );
		System.out.println( "CALCULATE ROAD LENGTH: " + 
		                    "Pixels: " + lPixelsLength + ", Scale: " + def.scale );
		
		return lPixelsLength * def.scale;
	}
	
	/**
	 * Draws the whole graph over a given image
	 * 
	 * @param image - the image to draw over
	 */
	private void drawGraph( BufferedImage image )
	{
		ArrayList roadPoints		= m_RoadGraph.getPoints( );

		/*		
		Graphics g = image.getGraphics( );
		g.setColor( Color.WHITE );
		g.fillRect( 0, 0, image.getWidth( ), image.getHeight( ) );
		*/
		
		Hashtable tester = new Hashtable( );
		// Test the road points first
		// Check if for each two points there is a connection from
		// both points to each other
		for ( int i = 0; i < roadPoints.size( ); i++ ) {
			RoadPoint p1 = ( RoadPoint ) roadPoints.get( i );
			for ( int j = 0; j < p1.connections.size( ); j++ ) {
				ConnectionPoint conn = ( ConnectionPoint ) p1.connections.get( j );
				RoadPoint p2 = m_RoadGraph.getPointByID( conn.id );
				
				boolean bFound = false;
				for ( int k = 0; k < p2.connections.size( ); k++ ) {
					conn = ( ConnectionPoint ) p2.connections.get( k );
					if ( conn.id == p1.id ) {
						bFound = true;
						break;
					}
				}
				if ( !bFound ) {
					System.out.println( "Corrupted graph!!!!" );
					System.out.println( p1.toString( ) );
					System.out.println( p2.toString( ) );
					System.out.println( "========================" );
				}
			}
		}
		
		for ( int i = 0; i < roadPoints.size( ); i++ ) {
			// Get the connection points for each of the road points
			// and draw a line between them
			RoadPoint point = ( RoadPoint ) roadPoints.get( i );
			for ( int j = 0; j < point.connections.size( ); j++ ) {
				ConnectionPoint conn = ( ConnectionPoint ) point.connections.get( j );
				ArrayList line = new ArrayList( );
				line.add( point );
				line.add( m_RoadGraph.getPointByID( conn.id ) );
				
				Color c = Color.CYAN;
				if ( tester.get( "" + conn.id ) == null ) {
					tester.put( conn.id + "", "" );
				}
				drawRoad( image, line, c );		
			}
		}
	}
	
	/**
	 * Draws a given road over a given image representing the
	 * currently viewed region.
	 * 
	 * @param image - the image to draw on
	 * @param road - an array list of RoadPoint objects that represent
	 *               a sequence of points forming a road
	 * @param c - the color of the road (used for test purposes). Can be null
	 */
	private void drawRoad( BufferedImage image, ArrayList road, Color c )
	{
		RoadPoint lastPoint 	= null;
		Graphics2D g			= ( Graphics2D ) image.getGraphics( );
		int iImageWidth		= image.getWidth( );
		int iImageHeight		= image.getHeight( );
		Rectangle viewer		= getViewerRegion( );
		Rectangle realViewer = scaleToRealSize( viewer );
		
		// Set the line stroke
		Stroke stroke			= new BasicStroke( 2f );
		g.setStroke( stroke );
		
		for ( int i = 0; i < road.size( ); i++ ) {
			RoadPoint point = ( RoadPoint ) road.get( i );
			// Draw a circle with a bright color to identify the road point
			// and connect to the last point
			if ( lastPoint != null ) {
				// Make sure that both points are inside the viewer region
				if ( realViewer.containsPoint( point.x, point.y ) == false ||
				     realViewer.containsPoint( lastPoint.x, lastPoint.y ) == false ) {
					lastPoint = null;
					continue;
				} // If point is inside

				// Convert the point coordinates to coordinates that will fit the current
				// zoom level and the current viewer image
				Rectangle r1 	= new Rectangle( point.x, point.y, 1, 1 );
				Rectangle r2 	= new Rectangle( lastPoint.x, lastPoint.y, 1, 1 );
				
				r1 				= scaleToZoomLevel( r1 );
				r2					= scaleToZoomLevel( r2 );
				r1.x				-= viewer.x;   r1.y				-= viewer.y;
				r2.x				-= viewer.x;	r2.y				-= viewer.y;
				
				Line2D.Double line 	= new Line2D.Double( r1.x, r1.y, r2.x, r2.y );
				g.setColor( c == null ? Color.CYAN : c );
				g.draw( stroke.createStrokedShape( line ) );
			} // If the last point is null
			lastPoint = point;
		}
	}
	
	/**
	 * Searches for a shortest road between two given areas. The problem
	 * this method solves actually searches for the shortest path between
	 * two points inside the areas. The points that are taken are the two
	 * closest to the center of the areas points.
	 * 
	 * @param lAreaID1 - one of the area
	 * @param lAreaID2 - the other area
	 * 
	 * @return an array list of RoadPoints identifying the shortest road.
	 *         If there's no shortest path an empty array list is returned
	 */
	private ArrayList getShortestPath( long lAreaID1, long lAreaID2 ) 
	{
		ArrayList roadPoints		= m_RoadGraph.getPoints( );
		
		RoadPoint p1 = getRoadSpot( roadPoints, lAreaID1 );
		// If there's no p1, we don't need to proceed further
		if ( p1 == null ) {
			System.out.println( "GET SHORTEST PATH: There's no road point beside area 1 (" + lAreaID1 + ")" );
			return new ArrayList( );
		} 
		
		RoadPoint p2 = getRoadSpot( roadPoints, lAreaID2 );
		if ( p2 == null ) {
			System.out.println( "GET SHORTEST PATH: There's no road point beside area 2 (" + lAreaID2 + ")" );
			return new ArrayList( );
		} 
		
		// We've got the two points. Now search for the shortest road
		ArrayList road = m_RoadGraph.findShortestRoad( p1, p2 );
		if ( road == null ) {
			System.out.println( "GET SHORTEST PATH: There's no road between given points: " + 
                             "(" + p1.x + "," + p1.y + ") <=> (" + p2.x + "," + p2.y + ")" );
			return new ArrayList( );
		}
		
		System.out.println( "GET SHORTEST PATH: Yes, there is a road" );
		
		// Road is found!!!
		return road;
	}
	
	/**
	 * Gets the road point that is closest to the center of a given area and
	 * within the area region
	 * 
	 * @param roadPoints - the available road points
	 * @param lAreaID - the area ID
	 * 
	 * @return the found road point or null if such point was not found
	 */
	private RoadPoint getRoadSpot( ArrayList roadPoints, long lAreaID )
	{
		RoadPoint closestRoadPoint = null;
		long lClosestDistance		= Integer.MAX_VALUE;
		
		if ( roadPoints == null ) {
			// The graph is empty
			System.out.println( "GET ROAD SPOT: The graph is empty" );
			return null;
		}
		
		// Get the area for the given area ID
		Area area = m_AreaManager.findAreaByUniqueID( lAreaID );
		if ( area == null ) {
			System.out.println( "GET ROAD SPOT: No such area ID " + lAreaID );
			return null;
		} 
		Rectangle r 		= area.getRectangle( );
		int iAreaCenterX	= r.x + r.width  / 2;
		int iAreaCenterY	= r.y + r.height / 2;
		
		for ( int i = 0; i < roadPoints.size( ); i++ ) {
			RoadPoint roadPoint = ( RoadPoint ) roadPoints.get( i );
			// Now calculate the distance to the center of the area and if 
			// it's shorter than the  currently shortest estimated one, 
			// remember the point
			long a = Math.abs( iAreaCenterX - roadPoint.x );
			long b = Math.abs( iAreaCenterY - roadPoint.y );
			long lDistance = Math.round( Math.sqrt( a * a + b * b ) );
			if ( lDistance < lClosestDistance ) {
				lClosestDistance = lDistance;
				closestRoadPoint = roadPoint;
			}
		}
		
		return closestRoadPoint;
	}
	
	/**
	 * Generates a cache key out of two area IDs. The key is a string:
	 *		<X coord1,Y coord1,Width1,Height1>-<X coord1,Y coord2,Width1,Height1> 
	 * where the first X,Y,W,H group  is the area with the smaller unique ID
	 * 
	 * @param lAreaID1 - one of the area IDs
	 * @param lAreaID2 - the other of the area IDs
	 * 
	 * @return the cache key string that can be used when searching withint the
	 *         shortest path cache
	 */
	private String getShortestPathCacheKey( long lAreaID1, long lAreaID2 )
	{
		long lID1	= Math.min( lAreaID1, lAreaID2 );
		long lID2	= Math.max( lAreaID1, lAreaID2 );
		Area area1	= m_AreaManager.findAreaByUniqueID( lID1 );
		Area area2 	= m_AreaManager.findAreaByUniqueID( lID2 );
		
		if ( area1 == null || area2 == null ) return "";
		
		Rectangle r1 = area1.getRectangle( );
		Rectangle r2 = area2.getRectangle( );
		
		return r1.x + "," + r1.y + "," + r1.width + "," + r2.height +
		       "-" + 
		       r2.x + "," + r2.y + "," + r2.width + "," + r2.height;
	}
	
	/**
	 * Draws an area over the map
	 * 
	 * @param dest - the bitmap image
	 * @param area - the area to draw
	 * 
	 * @return the result bitmap image
	 */
	private BufferedImage drawArea( BufferedImage dest, 
	                                Area area ) 
	{
		Rectangle rect = area.getRectangle( );
		
		/* Scale the area dimentions to the dimentions of the current zoom level */
		rect = scaleToZoomLevel( rect );
		
		Rectangle viewer = getViewerRegion( );

		int iWidth 		= rect.width; 
		int iHeight 	= rect.height;
		int iX 			= rect.x - viewer.x;
		int iY 			= rect.y - viewer.y;
		
		Graphics g = dest.getGraphics( );
		
		/* Draw the icon if any */
		String szIconID = area.getInfo( Constants.AREA_ATTRIBUTE_ICON_ID );
		if ( szIconID != null ) {
			int iIconID = Integer.parseInt( szIconID );
			Icon icon   = m_Icons.getIconByID( iIconID );
			Image img	= icon.getImage( );
			if ( icon != null ) {
				g.drawImage( img, iX, iY, iWidth, iHeight, null );
				// If the area has a submap - draw something distinctive
				String szSubMap = area.getInfo( Constants.AREA_ATTRIBUTE_SUB_MAP );
				if ( szSubMap != null && szSubMap.equals( "" ) == false ) {
					g.setColor( Color.WHITE );
					g.drawRect( iX, iY, iWidth, iHeight );
					g.setColor( Color.YELLOW );
					g.drawRect( iX + 1, iY + 1, iWidth, iHeight );
				}
			}
		}
		
		return dest;
	}
	
}
