package lib.map.dynamic;

import java.awt.MediaTracker;
import java.awt.Container;
import java.awt.Toolkit;
import java.awt.Image;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Hashtable;

import lib.db.DBManager;
import lib.cfg.Config;
import lib.cfg.Constants;
import lib.exceptions.IconsException;
import lib.exceptions.DBManagerException;

/**
 * @class lib.map.base.Icons
 * 
 * Icons list 
 * 
 * @author Tihomir Lazarov
 */
public class Icons {
	private ArrayList m_Icons 			= new ArrayList( );
	private Hashtable m_IconsByID 	= new Hashtable( );
	private ArrayList m_Groups			= new ArrayList( );
	private Hashtable m_GroupsByID	= new Hashtable( );
	
	/**
	 * Creates a list of icons
	 * 
	 * @param manager	- Database manager
	 * @param config	- Global configuration
	 * 
	 * @throws IconsException
	 */
	public Icons( DBManager manager, Config config )
		throws IconsException
	{
		MediaTracker tracker = new MediaTracker( new Container( ) );
		try {
			ResultSet records = manager.select( "SELECT * FROM icons" );
			int iImageIndex   = 0;
			while ( records.next( ) ) {
				String szFileName = records.getString( "FileName" );
				String szFilePath = config.getValue( Constants.PARAM_ICONS_DIRECTORY ) + 
                                "/" + 
                                szFileName;
				String szWebPath	= config.getValue( Constants.PARAM_ICONS_WEB_DIRECTORY ) + 
										  "/" + 
										  szFileName;
				Image img = Toolkit.getDefaultToolkit( ).createImage( szFilePath );
				
				tracker.addImage( img, iImageIndex++ );
				 
				Icon icon = new Icon(
					records.getInt( "ID" ),
					records.getString( "Description" ),
					szFilePath,
					szWebPath,
					img
				);
				
				m_Icons.add( icon );
				m_IconsByID.put( new Integer( icon.getID( ) ), icon );
			}
			tracker.waitForAll();
			
			// Get all groups and all relations between the icons and the groups
			records = manager.select( "SELECT * FROM category_groups" );
			while ( records.next( ) ) {
				int iGroupID 			= records.getInt( "ID" );
				String szGroupName	= records.getString( "Name" );
				if ( szGroupName == null ) continue;
				Group group = new Group( iGroupID, szGroupName );
				m_Groups.add( group );
				m_GroupsByID.put( new Integer( iGroupID ), group );
			}
			
			records = manager.select( "SELECT * FROM icons_cat_groups" );
			while ( records.next( ) ) {
				int iGroupID	= records.getInt( "GroupID" );
				int iIconID		= records.getInt( "IconID" );
				Icon icon		= ( Icon ) m_IconsByID.get( new Integer( iIconID ) );
				Group group		= ( Group ) m_GroupsByID.get( new Integer( iGroupID ) );
				if ( icon != null && group != null ) {
					group.addIcon( icon );
				}
			}
		} catch ( DBManagerException e ) {
			throw new IconsException( "DBManagerException: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new IconsException( "SQLException: " + e.getMessage( ), e );
		} catch ( InterruptedException e ) {
			throw new IconsException( "InterruptedException: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Returns the array of all available icons
	 * 
	 * @return the array of icons
	 */
	public ArrayList getIcons( )
	{
		return m_Icons;
	}

	/**
	 * Returns an icon object by a given ID. If there's
	 * no such icon ID - null is retkurned
	 * 
	 * @param iID - the ID of the icon to return 
	 * 
	 * @return null or the icon object with the given ID
	 */	
	public Icon getIconByID( int iID ) 
	{
		if ( m_IconsByID.containsKey( new Integer( iID ) ) ) {
			return ( Icon ) m_IconsByID.get( new Integer( iID ) );
		}
		return null;
	}
	
	/**
	 * Returns the icon groups that are currently available
	 *  
	 * @return an ArrayList of Group objects
	 */
	public ArrayList getGroups( )
	{
		return m_Groups;
	}

	/**
	 * Returns a group for a given ID
	 * 
	 * @param iID - the group ID to match
	 * 
	 * @return the Group object if matches the ID, or null otherwise
	 */	
	public Group getGroupByID( int iID )
	{
		if ( m_GroupsByID.containsKey( new Integer( iID ) ) ) {
			return ( Group ) m_GroupsByID.get( new Integer( iID ) );
		}
		return null;
	}
	
	/**
	 * @class lib.map.base.Icons.Icon
	 * 
	 * A simple Icon class that contains ID, Description, FilePath
	 * properties
	 * 
	 * @author Tihomir Lazarov
	 */
	public class Icon {
		int m_ID						= 0;
		String m_szDescription	= "";
		String m_szFilePath		= "";
		String m_szWebPath		= "";
		Image  m_Image				= null;
		
		public Icon( int iID, 
		             String szDescription, 
		             String szFilePath, 
		             String szWebPath, 
		             Image image ) 
		{
			m_ID					= iID;
			m_szDescription	= szDescription;
			m_szFilePath		= szFilePath;
			m_szWebPath			= szWebPath;
			m_Image				= image;
		}
		
		public int getID() 
		{
			return m_ID;
		}
		
		public String getDescription()
		{
			return m_szDescription;
		}
		
		public String getFilePath()
		{
			return m_szFilePath;
		}
		
		public String getWebPath( )
		{
			return m_szWebPath;
		}
		
		public Image getImage( )
		{
			return m_Image;
		}
	}
	
	public class Group
	{
		int m_iID			= 0;
		String m_szName	= "";
		ArrayList m_Icons	= null;
		
		public Group( int iID, String szName )
		{
			m_iID		= iID;
			m_szName	= szName;
			m_Icons	= new ArrayList( );
		}
		
		public String getName( ) 
		{
			return m_szName;
		}
		
		public int getID( )
		{
			return m_iID;
		}
		
		public ArrayList getIcons( )
		{
			return m_Icons;
		}
		
		public void addIcon( Icon icon )
		{
			m_Icons.add( icon );
		}
	}
}
