package lib.map.ext;

import java.awt.image.BufferedImage;
import java.awt.image.RenderedImage;
import java.awt.Graphics;
import java.awt.Color;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

import lib.map.base.colors.*;


/**
 * @class lib.map.base
 * 
 * Plays with image colors
 * 
 * @author Tihomir Lazarov
 * @date   May 14, 2005
 */
public class ColorManager {
	
	// The number of hue values 
	private static final int HUE_VALUES_COUNT	  			= 360 + 1;
	private static final int SATURATION_VALUES_COUNT	= 100 + 1;
	private static final int VALUE_VALUES_COUNT			= 100 + 1;
	
	public static final int DEFAULT_SATURATION         = 100;
	public static final int DEFAULT_VALUE              = 100;
	
	private BufferedImage     m_Source          			= null;
	private HistogramValue[]  m_HueHistogram 				= null;
	private HistogramValue[]  m_SaturationHistogram 	= null;
	private HistogramValue[]  m_ValueHistogram 			= null;
	private HistogramValue[]  m_DistanceHistogram		= null;	
	
	/**
	 * Manages image histogram
	 * 
	 * @param source
	 */
	public ColorManager( )
	{
	}
	
	/**
	 * Diminishes the number of colors in the image to a given value. Uses
	 * the hue only as a color value 
	 * 
	 * @param image - the image to diminish colors of
	 * @param colors - the number of colors to use (max 360)
	 */
	public void diminishColors( BufferedImage image, int iHueValues, int iSaturationValues, int iValueValues )
	{
		int iWidth 	= image.getWidth( );
		int iHeight = image.getHeight( );
		
		HSV hsv = new HSV( );
		
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				Color c = new Color( image.getRGB( i, j ) );
				
				hsv.toHSV( c.getRed( ), c.getGreen( ), c.getBlue( ) );
				
				hsv.hue 			= diminishColor( hsv.hue, HUE_VALUES_COUNT, iHueValues ); 
				hsv.saturation = diminishColor( hsv.saturation, SATURATION_VALUES_COUNT, iSaturationValues );
				hsv.value 		= diminishColor( hsv.value, VALUE_VALUES_COUNT, iValueValues );
				
				hsv.toRGB( hsv.hue, hsv.saturation, hsv.value );
				
				c = new Color( hsv.red, hsv.green, hsv.blue );
				image.setRGB( i, j, c.getRGB( ) );
			}
		}
	}
	
	public static long calculateDistance( Color color )
	{
		int r = color.getRed( );
		int g = color.getGreen( );
		int b = color.getBlue( );
		return Math.round( Math.sqrt( r * r + g * g + b * b ) );
	}
	
	private int diminishColor( int iValue, int iNumberOfValues, int iWantedColors )
	{
		float fGroups     = iNumberOfValues / ( float ) iWantedColors;
		int iRegionIndex 	= Math.round( ( iValue * ( float ) iWantedColors ) / iNumberOfValues );
		return Math.round( iRegionIndex * fGroups ); 
	}
	
	public void normalizeColors( BufferedImage image )
	{
		int iWidth	= image.getWidth( );
		int iHeight = image.getHeight( );
		
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				image.setRGB( i, j ,normalizeColor( new Color( image.getRGB( i, j ) ) ).getRGB( ) );
			}
		}
	}
	
	private Color normalizeColor( Color c )
	{
		int r = c.getRed( );
		int g = c.getGreen( );
		int b = c.getBlue( );
		
		int rn = ( 256 * r ) / ( r + g + b + 1 );
		int gn = ( 256 * g ) / ( r + g + b + 1 );
		int bn = ( 256 * b ) / ( r + g + b + 1 );
		return new Color( rn, gn, bn );
	}
	
	public void toYUV( BufferedImage image )
	{
		/*
			RGB -> YUV                                  | YUV -> RGB
			Y =  0.299*Red+0.587*Green+0.114*Blue       | Red   = Y+0.000*U+1.140*V
			U = -0.147*Red-0.289*Green+0.436*Blue       | Green = Y-0.396*U-0.581*V
			V =  0.615*Red-0.515*Green-0.100*Blue       | Blue  = Y+2.029*U+0.000*V
		*/	
		int iWidth	= image.getWidth( );
		int iHeight = image.getHeight( );
		
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				Color c = new Color( image.getRGB( i, j ) );
				int r = c.getRed( );
				int g = c.getGreen( );
				int b = c.getBlue( );
				int y = ( int )Math.round( 0.299 * r + 0.587 * g + 0.114 * b );
				int u = ( int )Math.round( 0.147 * r + 0.289 * g + 0.436 * b );
				int v = ( int )Math.round( 0.615 * r + 0.515 * g + 0.100 * b );
				c = new Color( y % 256, u % 256, v % 256 );
				image.setRGB( i, j, c.getRGB( ) );
			}
		}				
	}
	
	public void flattenHistogramColors( )
	{
		dumpHistogram( m_HueHistogram, "/tmp/hue1.png" );
		dumpHistogram( m_SaturationHistogram, "/tmp/saturation1.png" );
		dumpHistogram( m_ValueHistogram, "/tmp/value1.png" );
		
		flattenHistogramColors( m_HueHistogram );
		flattenHistogramColors( m_SaturationHistogram );
		flattenHistogramColors( m_ValueHistogram );
		
		dumpHistogram( m_HueHistogram, "/tmp/hue2.png" );
		dumpHistogram( m_SaturationHistogram, "/tmp/saturation2.png" );
		dumpHistogram( m_ValueHistogram, "/tmp/value2.png" );
	}
	
	/**
	 * Applies the last generated histogram over a passed image
	 * 
	 * @param dest - the image that will be applied over
	 */	
	public void applyHistogram( BufferedImage dest )
	{
		int iWidth 	= dest.getWidth( );
		int iHeight	= dest.getHeight( );
	
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				Color c = new Color( dest.getRGB( i, j ) );
				HSV hsv = new HSV( );
				hsv.toHSV( c.getRed( ), c.getGreen( ), c.getBlue( ) );
				if ( hsv.hue > 0 ) {
					int hue 			= m_HueHistogram[ hsv.hue ].value;
					int saturation = m_SaturationHistogram[ hsv.saturation ].value;
					int value		= m_ValueHistogram[ hsv.value ].value;
					hsv.toRGB( hue, saturation, value );
					dest.setRGB( i, j, new Color( hsv.red, hsv.green, hsv.blue ).getRGB() );
				}
			}
		}
	}
	
	/**
	 * Generates histogram from an image. 
	 * 
	 * Transforms the RGB colors to HSV and takes only the hue
	 * value. Creates an array of elements that are the
	 * number of occurances of the hue value. The array
	 * is as big as the number of different hsv values 
	 */
	public void generateHistogram( BufferedImage source )
	{
		int iWidth        = source.getWidth( );
		int iHeight       = source.getHeight( );
		
		m_HueHistogram 			= new HistogramValue[ HUE_VALUES_COUNT 			];
		m_SaturationHistogram	= new HistogramValue[ SATURATION_VALUES_COUNT 	];
		m_ValueHistogram        = new HistogramValue[ VALUE_VALUES_COUNT        ];

		// Initialize the hue histogram
		initializeHistogram( m_HueHistogram );
			
		// Initialize the saturation histogram
		initializeHistogram( m_SaturationHistogram );

		// Initialize the value histograam
		initializeHistogram( m_ValueHistogram );
		
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				Color c = new Color( source.getRGB( i, j ) );
				HSV hsv = new HSV( );
				hsv.toHSV( c.getRed( ), c.getGreen( ), c.getBlue( ) );

				m_HueHistogram[ hsv.hue ].value = hsv.hue;
				m_HueHistogram[ hsv.hue ].occurances++;
				
				m_SaturationHistogram[ hsv.saturation ].value = hsv.saturation;
				m_SaturationHistogram[ hsv.saturation ].occurances++;
				
				m_ValueHistogram[ hsv.value ].value = hsv.value;
				m_ValueHistogram[ hsv.value ].occurances++;
			}
		}
	}	
	
	/**
	 * Initializes an array of histogram values 
	 * 
	 * @param histogram - the array to initialize
	 * 
	 * @return the initialized array
	 */
	private void initializeHistogram( HistogramValue[] histogram )
	{
		for ( int i = 0; i < histogram.length; i++ ) {
			histogram[ i ] = new HistogramValue( );
			histogram[ i ].value      = 0;
			histogram[ i ].occurances = 0;
		}
	}
	
	/**
	 * Flattens the peaks in the histogram, i.e. every up and down
	 * is made flat using the middle value between the up and the down
	 * 
	 * @param histogram - the histogram to use
	 */
	private void flattenHistogramColors( HistogramValue[] histogram )
	{
		HistogramValue lastValue		= histogram[ 0 ];
		HistogramValue lastMaxValue 	= lastValue;
		int iPeakStartIndex	= -1;
		int iPeakTopIndex		= -1;
		for ( int i = 1; i < histogram.length; i++ ) {
			HistogramValue currentValue = histogram[ i ];
			
			if ( lastValue.occurances < currentValue.occurances ) { // We're going up
				lastMaxValue = currentValue;
			
				if ( iPeakStartIndex == -1 ) {
					iPeakStartIndex 	= i - 1;
				}
			
				if ( iPeakTopIndex != -1 ) {
					for ( int j = iPeakTopIndex; j <= i; j++ ) {
						histogram[ j ] = lastMaxValue;
						iPeakTopIndex = -1;
					}
				}
			
			} else if ( lastValue.occurances > currentValue.occurances ) { // We're going down
				if ( iPeakTopIndex == -1 ) {
					iPeakTopIndex = i - 1; 
				}
			
				if ( iPeakStartIndex != -1 ) {
					for ( int j = iPeakStartIndex; j <= i; j++ ) {
						histogram[ j ] = lastMaxValue;
					}
					iPeakStartIndex = -1;
				}
			}
		
			lastValue = currentValue;
		}
	}
	
	private void dumpHistogram( HistogramValue[] histogram, String szFileName )
	{
		int iWidth  = histogram.length;
		int iHeight = 100;
		
		int iRectangleWidth 	= iWidth / histogram.length;
		BufferedImage image 	= new BufferedImage( iWidth, iHeight, BufferedImage.TYPE_INT_RGB );
		int iMaxValue 			= -1;
		
		// Find the max value of occurances. It wlil be the 100% height metric.
		// For each value draw a black rectangle on the appropriate position
		// with height - the number of occurances of the value calculated
		// against the maximum occurances count
		for ( int i = 0; i < histogram.length; i++ ) {
			if ( iMaxValue < histogram[ i ].occurances ) {
				iMaxValue = histogram[ i ].occurances;
			}
		}
		
		if ( iMaxValue == 0 ) { iMaxValue = 1; } // To avoid division by zero
		
		// Draw a white background
		Graphics g = image.getGraphics( );
		g.setColor( Color.WHITE );
		g.fillRect( 0, 0, iWidth, iHeight );
		g.setColor( Color.BLACK );
		
		for ( int i = 0; i < histogram.length; i++ ) {
			int iValueOccurances = histogram[ i ].occurances;
			int iRectangleHeight = Math.round( ( iValueOccurances / ( float )iMaxValue ) * iHeight );
			int iXPosition       = i * iRectangleWidth;
			g.fillRect( iXPosition, iHeight - iRectangleHeight, iRectangleWidth, iRectangleHeight );
		}
		
		// Save the image into a file
		RenderedImage rendImage = image;
		try {
			File file 		= new File( szFileName );
			String szExt	= szFileName.substring( szFileName.lastIndexOf( '.' ) + 1 );
			ImageIO.write( rendImage, szExt, file );
		} catch ( IOException e ) {
			System.out.println( "I/O Exception: " + e.getMessage( ) );
		}
	}
	
	/*
	 * @class lib.map.base.HistogramValue
	 * 
	 * A single element from the array the histogram is built of
	 * 
	 * @author Tihomir Lazarov
	 * @date   May 14, 2005
	 */
	private class HistogramValue
	{
		public int occurances 	= 0;
		public int value			= 0;
	}
}  
