package lib.map.ext;

import java.awt.image.BufferedImage;

import lib.map.base.Map;
import lib.map.base.Rectangle;
import lib.exceptions.NullMapException;
import lib.exceptions.MapStateException;

/**
 * @class lib.map.ext
 * 
 * Externs several of BufferedImage methods to work over a mapped
 * image. The class uses a temporary image that keeps the
 * current viewable region. If methods that manage image pixels
 * request pixels inside the area where the current viewable region
 * is, the temporary image is used. Otherwise the viewable region
 * is moved to a position where the x,y requested pixel will reside
 * inside the region 
 * 
 * @author Tihomir Lazarov
 * @date   Jun 2, 2005
 */
public class MappedBufferedImage extends BufferedImage {
	Map m_Map 							= null;
	BufferedImage m_Image 			= null;
	Rectangle m_ViewerRectangle	= null;	
	
	/**
	 * Constructs a MappedBufferedImage using a Map
	 * 
	 * @param map - the map to use
	 */
	public MappedBufferedImage( Map map )
	{
		// Initialize the super object with some fake values
		super( 1, 1, BufferedImage.TYPE_INT_RGB );
		
		// Remember the map and the viewer rectangle
		m_ViewerRectangle = map.getViewerRegion( );
		m_Map 				= map;
		
		try {
			// Make sure that initially we're at the top left corner
			int iZoomLevel = map.getZoomLevels( ) - 1;
			map.zoomTo( iZoomLevel );
			map.translateTo( 0, 0 );
			
			// "cache" the current viewable region image
			m_Image = map.getViewableRegionSnapshot( );
		} catch ( NullMapException e ) {
			System.out.println( "Null map exception: " + e.getMessage( ) ); 
		} catch ( MapStateException e ) {
			System.out.println( "Map state exception: " + e.getMessage( ) );
		}
	}

	/**
	 * A "copy" constructor
	 * 
	 * @param image - the source image to copy
	 */	
	public MappedBufferedImage( BufferedImage image )
	{
		super( image.getWidth( ), image.getHeight( ), BufferedImage.TYPE_INT_RGB );
		
		int iWidth  = image.getWidth( );
		int iHeight = image.getHeight( );
		
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				setRGB( i, j, image.getRGB( i, j ) );
			}
		}
	}
	
	/**
	 * Make a backward-compatible class constructor
	 * 
	 * @param iWidth - the image width
	 * @param iHeight - the image height
	 * @param type - the image type (see BufferedImage documentation)
	 */
	public MappedBufferedImage( int iWidth, int iHeight, int type )
	{
		super( iWidth, iHeight, type );
	}
	
	/**
	 * Returns the large image width
	 */
	public int getWidth( )
	{
		if ( m_Map != null ) {
			return m_Map.getWidth( );
		} else {
			return super.getWidth( );
		}
	}
	
	/**
	 * Returns the large image height
	 */
	public int getHeight( )
	{
		if ( m_Map != null ) {
			return m_Map.getHeight( );
		} else {
			return super.getHeight( );	
		}
	}
	
	/**
	 * Returns the RGB from a given pixel which position is absolute over
	 * the big image
	 * 
	 * @param x - the X position
	 * @param y - the Y position
	 * 
	 * @return the RGB value
	 */
	public int getRGB( int x, int y )
	{
		if ( m_Map != null ) {
			BufferedImage image 	= getImagePart( x, y );
			x -= m_ViewerRectangle.x;
			y -= m_ViewerRectangle.y;
			return image.getRGB( x, y );
		} else {
			return super.getRGB( x, y );
		}
	}
	
	/**
	 * Sets a RGB value on a given pixel position absolute to the
	 * big image
	 * 
	 * @param x - the X position
	 * @param y - the Y position
	 * @param rgb - the RGB color
	 */
	public void setRGB( int x, int y, int rgb )
	{
		if ( m_Map != null ) {
			BufferedImage image = getImagePart( x, y );
			x -= m_ViewerRectangle.x;
			y -= m_ViewerRectangle.y;
			image.setRGB( x, y, rgb );
		} else {
			super.setRGB( x, y, rgb );
		}
	}

	/**
	 * Returns a temporary image from the map where the (x, y) pixel is inside
	 * the temporary image. If the (x, y) pixel is inside the currently viewed
	 * region, the current temporary image is returned. 
	 * 
	 * @param x - the X position that must reside in the viewable region
	 * @param y - the Y position
	 * 
	 * @return a temporary image that's not bigger than the viewed region
	 */	
	private BufferedImage getImagePart( int x, int y )
	{
		// The (x, y) pixel is inside the viewer region
		if ( x >= m_ViewerRectangle.x && x < m_ViewerRectangle.x + m_ViewerRectangle.width &&
		     y >= m_ViewerRectangle.y && y < m_ViewerRectangle.y + m_ViewerRectangle.height ) {
			return m_Image;
		}
		
		// Calculate the X, Y of the viewer region so that the (x, y) pixel 
		// will be located inside
		int iXViewerIndex = x / m_ViewerRectangle.width;
		int iYViewerIndex = y / m_ViewerRectangle.height;
		int iRealX			= iXViewerIndex * m_ViewerRectangle.width;
		int iRealY			= iYViewerIndex * m_ViewerRectangle.height;
		
		// Move the viewer rectangle to that position
		try {
			m_Map.translateTo( iRealX, iRealY );
			m_Image 				= m_Map.getViewableRegionSnapshot( );
			m_ViewerRectangle = m_Map.getViewerRegion( );			 
		} catch ( MapStateException e ) {
			System.out.println( "Map state exception: " + e.getMessage( ) );
		} catch ( NullMapException e ) {
			System.out.println( "Null map exception: " + e.getMessage( ) );
		}
		
		return m_Image;
	}
}
