package lib.road;

import java.util.ArrayList;
import java.util.Hashtable;


/**
 * @class lib.road
 * 
 * Implements a graph of road points
 * 
 * @author Tihomir Lazarov
 * @date   May 11, 2005
 */
public class RoadGraph {
	ArrayList m_Points;
	Hashtable m_PointsMap;
	int       m_iNextID = 0;
	
	// Those below are used for Dijkstra's shortest path algorithm	
	private static final int INFINITY 	= Integer.MAX_VALUE;
	
	/**
	 * Manages graph of points 
	 */
	public RoadGraph( )
	{
		m_Points 	= new ArrayList( );
		m_PointsMap	= new Hashtable( );
	}
	
	/**
	 * Adds a point to the graph
	 * 
	 * @param p - the road point to add
	 */
	public void addPoint( RoadPoint p ) 
	{
		p.id = m_iNextID++;
		m_Points.add( p );
		m_PointsMap.put( new Integer( p.id ), p );
	}
	
	/**
	 * Adds a connection joint between two points. The
	 * connection is two way, i.e. there is a connection
	 * from the first to the second point and vice versa.
	 * 
	 * @param p1 - the first point
	 * @param p2 - the second point
	 */
	public void addConnection( RoadPoint p1, RoadPoint p2 )
	{
		if ( p1 != null && p2 != null ) {
			// Add a 2 way connection
			long lDistance = getDistance( p1, p2 );
			addConnectionPoint( p1, p2.id, lDistance );
			addConnectionPoint( p2, p1.id, lDistance ); 
		}
	}
	
	/**
	 * Adds a directed connection joint between two points
	 * 
	 * @param p1 - the first point
	 * @param p2 - the second point
	 */
	public void addDirectedConnection( RoadPoint p1, RoadPoint p2 )
	{
		if ( p1 != null && p2 != null ) {
			long lDistance = getDistance( p1, p2 ); 
			addConnectionPoint( p1, p2.id, lDistance );
		}
	}
	
	/**
	 * Checks if there's a direction connection between two points
	 * 
	 * @param id1 - the first point
	 * @param id2 - the second point
	 * 
	 * @return true if there's a direction connection. false otherwise
	 */
	public boolean hasDirectConnection( RoadPoint p1, RoadPoint p2 ) 
	{
		if ( p1 == null && p2 == null ) {
			return false;
		}

		// There's no need to check for a connection from
		// p1 to p2 and vice versa as we assume that both
		// connections are available		
		for ( int i = 0; i < p1.connections.size( ); i++ ) {
			ConnectionPoint p = ( ConnectionPoint )p1.connections.get( i );
			if ( p1.id == p2.id ) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Returns the point for a given ID
	 * 
	 * @param id - the ID to search for
	 * 
	 * @return
	 */
	public RoadPoint getPointByID( int id )
	{
		return ( RoadPoint ) m_PointsMap.get( new Integer( id ) );
	}
	
	/**
	 * Finds a road between two points 
	 * 
	 * @param p1 - the source point
	 * @param p2 - the destination point
	 * 
	 * @return an array list of the indexes of
	 *         the points that are on the way
	 */
	public ArrayList findRoad( RoadPoint p1, RoadPoint p2 )
	{
		// Reset all road points
		resetRoadPoints( );
	
		// Find road
		return findRoadEx( p1, p2 );
	}
	
	/**
	 * Finds the shortest road between two points using Dijkstra's
	 * algorithm
	 * 
	 * @param p1 - the source point
	 * @param p2 - the destination point
	 * 
	 * @return a list of points that drive the source point to the
	 *         destination
	 */
	public ArrayList findShortestRoad( RoadPoint p1, RoadPoint p2 )
	{
		// Initialize data structures
		Hashtable distances			= new Hashtable( );
		Hashtable predecessors 		= new Hashtable( );
		ArrayList settledNodes 		= new ArrayList( );
		ArrayList unsettledNodes	= new ArrayList( );
		
		for ( int i = 0; i < m_Points.size( ); i++ ) {
			RoadPoint p  = ( RoadPoint ) m_Points.get( i );
			distances.put( p, new Long( INFINITY ) );
		}
		
		for ( int i = 0; i < m_Points.size( ); i++ ) {
			unsettledNodes.add( m_Points.get( i ) );
		}
		
		// Add the source point to the list of settled nodes
		settledNodes.add( p1 );
		distances.put( p1, new Long( 0 ) );
		
		// Loop until there are unsettled nodes
		while ( unsettledNodes.size( ) > 0 ) {
			// Find the closest point from the unsettled nodes
			RoadPoint minPoint = djkstraExtractMin( unsettledNodes, distances );
			
			// Settle it
			settledNodes.add( minPoint );
			
			if ( minPoint.id == p2.id ) { break; }
			
			// "Relax" neighbours
			dijkstraRelaxNeighbours( minPoint, settledNodes, 
											 unsettledNodes, distances, 
											 predecessors );
		}
		
		RoadPoint pre = ( RoadPoint )predecessors.get( p2 );
		if ( pre == null ) return null;
		
		ArrayList shortestPath = new ArrayList( );
		shortestPath.add( p2 );
		while ( pre != null && pre.id != p1.id ) {
			shortestPath.add( pre );
			pre = ( RoadPoint )predecessors.get( pre );
		}
		
		if ( pre != null ) {
			shortestPath.add( pre );
		}
		
		return shortestPath;
	}
	
	/**
	 * Remove the bi-directional connection between points
	 * 
	 * @param p1 - the first point
	 * @param p2 - the second ponit
	 */
	public void removeConnection( RoadPoint p1, RoadPoint p2 ) 
	{
		removeConnectionPoint( p1, p2.id );
		removeConnectionPoint( p2, p1.id );
	}
	
	/**
	 * Removes a point from the graph that's not referenced
	 * by any other point
	 * 
	 * @param p - the point to remove
	 * 
	 * @return true if the point is sucessfully removed, false
	 *         if there are references 
	 */
	public boolean removePoint( RoadPoint p ) 
	{
		if ( p.references > 0 ) {
			System.out.println( "Will not remove point. Still referenced: " + p.id ); 
			return false; 
		}
		m_Points.remove( p );
		m_PointsMap.remove( new Integer( p.id ) );
		return true;
	}
	
	/** -- PRIVATE METHODS -------------------------------------------------- */
	
	/**
	 * Adds a connection to a given road graph point. Increments the reference
	 * count of the connection point
	 * 
	 * @param p - the source point
	 * @param iConnectionID - the ID of the connection point
	 * @param lDistance - the distance between the points
	 */
	private void addConnectionPoint( RoadPoint p, int iConnectionID, long lDistance )
	{
		p.connections.add( new ConnectionPoint( iConnectionID, lDistance ) );
		
		RoadPoint connection = getPointByID( iConnectionID );
		connection.references++;
	}
	
	/**
	 * Removes a connection from the list of connections for
	 * a given point
	 * 
	 * @param p - the point which connections list will be changed
	 * @param iConnectionID - the connection ID to remove
	 */
	private void removeConnectionPoint( RoadPoint p, int iConnectionID )
	{
		for ( int i = 0; i < p.connections.size( ); i++ ) {
			ConnectionPoint c = ( ConnectionPoint ) p.connections.get( i );
			if ( c.id == iConnectionID ) {
				p.connections.remove( i );
				RoadPoint conn = getPointByID( iConnectionID );
				if ( conn != null ) {
					conn.references--;
				}
				return;
			}
		}
	}

	/**
	 * "Relaxes" the neighbours of a given point. "Relaxing" is in the
	 * terms of Dijkstra's shortest path algorithm. It works like that:
	 * - Gets all neighbours 
	 * - For each neighbour is checked 
	 * 	- if the distance from the source point to the neighbour is bigger
	 *      than the distance to the current point + distance between current
	 *      point and the neighbour.
	 *    - If so the shortest distance from the source point to the neighbour
	 *      point is marked to be the one through the current point
	 *    - The predecessor of the neighbour point is set to be the current point
	 *   
	 * @param p - the current point
	 * @param settledNodes
	 * @param unsettledNodes
	 * @param distances
	 * @param predecessors
	 */	
	private void dijkstraRelaxNeighbours( RoadPoint p, ArrayList settledNodes, 
	                                      ArrayList unsettledNodes, Hashtable distances, 
	                                      Hashtable predecessors ) {
		long pDistance = ( ( Long )distances.get( p ) ).longValue( );
		for ( int i = 0; i < p.connections.size( ); i++ ) {
			ConnectionPoint cp = ( ConnectionPoint ) p.connections.get( i );
			RoadPoint conn = getPointByID( cp.id );
			
			if ( settledNodes.contains( conn ) == false ) {
				long connDistance	= ( ( Long )distances.get( conn ) ).longValue( );
				long ptsDistance	= getDistance( p, conn );
				if ( connDistance > pDistance + ptsDistance ) {
					distances.put( conn, new Long( pDistance + ptsDistance ) );
					predecessors.put( conn, p );
				}
			}
		}
	}
	
	/**
	 * Extracts the "minimum" unsettled node. By "minimum" I mean the
	 * one that has the least "distance" value
	 * 
	 * @param unsettledNodes - a set of unsettled nodes
	 * @param distances - a hashmap of distances to nodes
	 * 
	 * @return the minium node
	 */
	private RoadPoint djkstraExtractMin( ArrayList unsettledNodes, Hashtable distances )
	{
		if ( unsettledNodes == null || unsettledNodes.size( ) == 0 ) return null;
		
		RoadPoint minPoint 	= ( RoadPoint )unsettledNodes.get( 0 );
		long lMinDistance		= ( ( Long )distances.get( minPoint ) ).longValue( );
		int iMinIndex      	= 0;
		
		for ( int i = 1; i < unsettledNodes.size( ); i++ ) {
			RoadPoint thisPoint 	= ( RoadPoint ) unsettledNodes.get( i );
			long lThisDistance		= ( ( Long )distances.get( thisPoint ) ).longValue( );
			if ( lThisDistance < lMinDistance ) {
				iMinIndex		= i;
				lMinDistance 	= lThisDistance;
				minPoint			= thisPoint;
			}
		}
		
		unsettledNodes.remove( iMinIndex );
		
		return minPoint;
	}
	
	/**
	 * Gets the distnace between two points
	 * 
	 * @param p1 - from point
	 * @param p2 - to point
	 * 
	 * @return the distance in pixels
	 */
	public static long getDistance( RoadPoint p1, RoadPoint p2 )
	{
		long a = Math.abs( p1.x - p2.x );
		long b = Math.abs( p1.y - p2.y );
	
		return Math.round( Math.sqrt( a * a + b * b ) );
	}
	
	/**
	 * Calculates the square distance between two points. The
	 * distance is not sqrt-ed, because of performance issues 
	 * 
	 * @param p1 - one of the point
	 * @param p2 - the other point
	 * 
	 * @return the distance in pixels
	 */
	public static long getDistanceSquare( RoadPoint p1, RoadPoint p2 ) 
	{
		int x = p1.x - p2.x;
		int y = p1.y - p2.y;
		return x * x + y * y;
	}
	
	
	/**
	 * Returns all points in the graph with their connections
	 * 
	 * @return an array of points
	 */
	public ArrayList getPoints( )
	{
		return m_Points;
	}
	
	/**
	 * Calculates road length in pixels - the distance between the points
	 * 
	 * @param roadPoints - an array list of road points
	 * 
	 * @return the distance
	 */
	public long getRoadLength( ArrayList roadPoints )
	{
		long lDistance = 0;
		
		if ( roadPoints == null || 
			( roadPoints != null && roadPoints.size( ) <= 1 ) ) {
			return Integer.MAX_VALUE;
		}
		
		RoadPoint lastPoint = ( RoadPoint )roadPoints.get( 0 );
		for ( int i = 1; i < roadPoints.size( ); i++ ) {
			RoadPoint currentPoint = ( RoadPoint )roadPoints.get( i );
			lDistance += getDistance( lastPoint, currentPoint );
			lastPoint = currentPoint;
		}
		
		return lDistance;
	}
	
	/**
	 * Dumps the graph into a string with the following format:
	 * 	<point index1> (<x>, <y>): <connection index1, connection index2, ...>
	 * 	<point index2> (<x>, <y>): <connection index1, connection index2, ...>
	 * 	...
	 */
	public String toString( )
	{
		String szDump = "";
		
		for ( int i = 0; i < m_Points.size( ); i++ ) {
			RoadPoint point = ( RoadPoint ) m_Points.get( i );
			szDump += point.id + " (" + point.x + ", " + point.y + "): ";
			
			for ( int j = 0; j < point.connections.size( ); j++ ) {
				ConnectionPoint connectionPoint = ( ConnectionPoint )point.connections.get( j );
				int iConnectionID = connectionPoint.id;
				szDump += iConnectionID + ",";
			}
			
			if ( point.connections.size( ) > 0 ) {
				// Remove the trailing ","
				szDump = szDump.substring( 0, szDump.length( ) );
			}
			szDump += "\n";
		}
		
		return szDump;
	}
	
	/** -- PRIVATE METHODS -------------------------------------------------- */
	
	/**
	 * Marks all road points as not visited
	 */
	private void resetRoadPoints( )
	{
		for ( int i = 0; i < m_Points.size( ); i++ ) {
			RoadPoint p = ( RoadPoint ) m_Points.get( i );
			p.visited = false;
		}
	}
	
	/**
	 * Finds a road between two points using a stack, not recursively
	 * 
	 * @param p1 - the source point
	 * @param p2 - the destination point
	 * 
	 * @return an array of points that drive id1 to id2
	 */
	private ArrayList findRoadEx( RoadPoint p1, RoadPoint p2 )
	{
		ArrayList pointsStack = new ArrayList( );
		ArrayList roadsStack  = new ArrayList( );
		
		RoadPoint currentPoint = p1;
		ArrayList currentRoad  = new ArrayList( );
		
		pointsStack.add( currentPoint );
		roadsStack.add( currentRoad );
		
		// Put the points we pass through in the points stack. Put the
		// built roads in a stack too, so when we're backtracing to
		// remove the last "misleading" roads
		while ( pointsStack.size( ) > 0 ) {
			int iPointIndex = pointsStack.size( ) - 1;
			int iRoadIndex  = roadsStack.size( ) - 1;
			
			// Pop the last point and last road from the stack
			currentPoint = ( RoadPoint ) pointsStack.get( iPointIndex );
			currentRoad  = ( ArrayList ) roadsStack.get( iRoadIndex );
			pointsStack.remove( iPointIndex );
			roadsStack.remove( iRoadIndex );

			// Skip the already visited points			
			if ( currentPoint.visited ) { continue; }
			currentPoint.visited = true;
			
			// Add the point to the road
			currentRoad.add( currentPoint );
			
			// If there's a direct connection between the last point
			// from the stack and the destination point, then we've reached
			// our goal.			
			if ( hasDirectConnection( currentPoint, p2 ) ) {
				ArrayList tmpRoad = new ArrayList( currentRoad );
				tmpRoad.add( p2 );
				
				RoadPoint goalPoint = p2;
				goalPoint.visited = true;
				currentRoad.add( goalPoint );
				roadsStack.clear( );
				pointsStack.clear( );
				return currentRoad;
			}

			// If there's not direct connection, try to find a direct connection
			// with all of the neighbour points to the direction by pushing them
			// onto the stack. Push the same number of roads on the stack too.
			for ( int i = 0; i < currentPoint.connections.size( ); i++ ) {
				ConnectionPoint nextConnectionPoint = ( ConnectionPoint )currentPoint.connections.get( i );
				RoadPoint nextPoint = getPointByID( nextConnectionPoint.id );
				
				if ( nextPoint.visited == false ) {
					roadsStack.add( currentRoad );
					pointsStack.add( nextPoint );
				}
			}
		}
		
		// No road is found
		return null;
	}
}
