package lib.road;

import java.util.ArrayList;

/**
 * @class lib.road
 * 
 * Optimizes a graph of RoadPoint elements. The optimization procedure
 * is to diminish the graph points by removing 1 of 3 points that are
 * "on the same line" and have direction connections between, 
 * where "on the same line" means that they are not
 * exactly on the same line, but "almost". The angle between the vectors
 * these points define is taken
 * 
 * @author Tihomir Lazarov
 * @date   Jun 4, 2005
 */
public class RoadGraphOptimizer {
	
	double ONE_LINE_COSINE	= 0.90; 

	/**
	 * Optimizes a graph by removing 1 of 3 points that are on the same line
	 * 
	 * @param g - the graph
	 */	
	public void optimizeGraph( RoadGraph g )
	{
		// Loop until there's no optimization. 
		// On each loop take every 3 distinct
		// points and check if they are on one line
		int iAll							= g.getPoints( ).size( );
		while ( optimizeGraphPoints( g ) == true );
		int iRemoved					= iAll - g.getPoints( ).size( );
		double dOptimizationRatio  = ( iRemoved * 100.0 ) / iAll;
		System.out.println( "Optmiized: " + dOptimizationRatio + "% (" + iRemoved + " of " + iAll + " removed)" );
	}
	
	/**
	 * Optimizes the graph points and returns true if there was 
	 * a single optimization
	 * 
	 * @param g - the graph
	 * 
	 * @return true if there was a single optimization, false otherwise
	 */
	private boolean optimizeGraphPoints( RoadGraph g )
	{
		ArrayList points = g.getPoints( );
			
		// Foreach point 
		for ( int i = 0; i < points.size( ); i++ ) {
			RoadPoint p1 = ( RoadPoint ) points.get( i );
	
			// Take the point neighbours
			for ( int j = 0; j < p1.connections.size( ); j++ ) {
				ConnectionPoint connPoint2 = ( ConnectionPoint ) p1.connections.get( j );
				RoadPoint p2 = g.getPointByID( connPoint2.id );
		
				// If this is the "middle" point, then if it will be removed
				// later it must be connected only to 2 other points - p1 and
				// one of its neighbours
				if ( p2.references == 2 ) {
			
					// Yep, this is our point. Take its neighbours two and
					// for each one of them check if they are on the same line
					for ( int k = 0; k < p2.connections.size( ); k++ ) {
						ConnectionPoint connPoint3 = ( ConnectionPoint ) p2.connections.get( k );
						RoadPoint p3 = g.getPointByID( connPoint3.id );
				
						// Ignore the loops...
						if ( p3.id == p1.id ) continue; 
				
						// Check if the points are on the same line
						if ( areOnSingleLine( p1, p2, p3 ) ) {
							// Yes, they are. 
							// We have
							//  o --------- o ----------o
							//  p1          p2         p3
							// Remove the connections between p1 and p2 and
							// between p2 and p3. After that connect
							// p1 with p3 and kill p2.
							g.removeConnection( p1, p2 );
							g.removeConnection( p2, p3 );
							g.removePoint( p2 );
							if ( g.hasDirectConnection( p1, p3 ) == false ) {
								g.addConnection( p1, p3 );
							}
							return true;
						}
					} // for k...
				} // if p2.references..
			} // for j
		} // for i
		return false;		
	}

	private boolean areOnSingleLine( RoadPoint p1, RoadPoint p2, RoadPoint p3 )
	{
		Vector2D v1 = subtractVectors( new Vector2D( p1.x, p1.y ), new Vector2D( p2.x, p2.y ) );
		Vector2D v2 = subtractVectors( new Vector2D( p2.x, p2.y ), new Vector2D( p3.x, p3.y ) );
	
		double dCos = cosBetweenVectors( v1, v2 );
		return ( dCos <= -0.9 );
	}

	/**
	 * @class lib.road
	 * 
	 * Simple 2D vector implementation
	 * 
	 * @author Tihomir Lazarov
	 * @date   Jun 4, 2005
	 */	
	private class Vector2D
	{
		double  m_dX 				= 0;
		double  m_dY 				= 0;
		double  m_dLength			= 0;
		boolean m_bNormalized 	= false;
	
		public Vector2D( double x, double y )
		{
			m_dX			= x;
			m_dY 			= y;
			m_dLength	= Math.sqrt( x * x + y * y );
		}
	
		public double length( )
		{
			return m_dLength;  
		}
	
		public double x( )
		{
			return m_dX;
		}
	
		public double y( )
		{
			return m_dY;
		}
	
		public void normalize( )
		{
			if ( m_bNormalized == true ) {return;}
			double dLength = length( );
			m_dX /= dLength;
			m_dY /= dLength;
		}
	}

	/**
	 * Subtracts two vectors
	 * 
	 * @param v1 - the first vector
	 * @param v2 - the second vector
	 * 
	 * @return the result vector
	 */
	private Vector2D subtractVectors( Vector2D v1, Vector2D v2 )
	{
		return new Vector2D( v2.x() - v1.x(), v2.y() - v1.y() );
	}

	/**
	 * Calculates a dot product between two vectors
	 * 
	 * @param v1 - the first vector
	 * @param v2 - the second vector
	 * 
	 * @return the result vector
	 */
	private double getVectorDotProduct( Vector2D v1, Vector2D v2 )
	{
		return v1.x() * v2.x() + v1.y() * v2.y();
	}

	/**
	 * Calculates the cosine betwen two vectors
	 * 
	 * @param v1 - the first vector
	 * @param v2 - the second vector
	 * 
	 * @return the cosine
	 */
	private double cosBetweenVectors( Vector2D v1, Vector2D v2 )
	{
		// The formula for calculating the cosine between
		// two vectors is:
		//  v1 * v2 = cos(a) * | v1 | * | v2 |
		return getVectorDotProduct( v1, v2 ) / ( v1.length( ) * v2.length( ) );
	}
}
