package lib.road;

import java.awt.Color;
import java.awt.image.BufferedImage;

import java.io.File;
import java.io.IOException;

import java.util.ArrayList;
import javax.imageio.ImageIO;

import lib.map.base.colors.YCrCb;
import lib.utils.ProgressBar;

/**
 * @class lib.road
 * 
 * Recognizes a road in a given HSV color range from a bitmap
 * 
 * @author Tihomir Lazarov
 * @date   May 14, 2005
 */
public class RoadRecognition {
	// This is the size of the matrix that's used to clear
	// the "bad" pixels, i.e. lonely black points on the canvas
	// or lonely white points in the middle of the roads
	int m_iClearPixelsFuzzinessFactor = 3;
	
	// The number of pixels in the matrix with size 
	// iClearPixelsFuzziness factor to assume this is a 
	// "black" spot
	int m_iBlackSpotFuzzinessFactor   = 3;

	// The percentage to assume that a block of black pixels
	// covers fully a part of the road
	int m_iRoadBlockMatchRatio        = 99;

	// The maximum distance between adjacent points in the graph
	int m_iMaxAdjacentPointsDistance  = 50;
	
	// The number of steps to reach from 1 to m_iMaxAdjacentPointsDistance
	int m_iAdjacencySteps				 = 15;
	
	private ProgressBar m_ProgressBar = null;
	
	public static final int CFG_CLEAR_PIXELS_FUZZINESS_FACTOR = 1;
	public static final int CFG_BLACK_SPOT_FUZZINESS_FACTOR   = 2;
	public static final int CFG_ROAD_BLOCK_MATCH_RATIO        = 3;
	public static final int CFG_MAX_ADJACENT_POINTS_DISTANCE  = 4;
	public static final int CFG_ADJACENCY_STEPS					 = 5;
	
	YCrCb m_YCrCbWork	= new YCrCb( );
	
	/**
	 * Contructor by HSV range
	 * 
	 * @param from - HSV from
	 * @param to - HSV to
	 */
	public RoadRecognition( ProgressBar progressBar )
	{
		m_ProgressBar = progressBar;
	}
	
	/**
	 * Set a configuration parameter value
	 * 
	 * @param iParameterType - the type of the parameter. See CFG_* constants
	 * @param iValue - the value of the parameter
	 */
	public void setConfigParameter( int iParameterType, int iValue )
	{
		switch ( iParameterType ) {
			case CFG_BLACK_SPOT_FUZZINESS_FACTOR:
				m_iBlackSpotFuzzinessFactor = iValue;
				break;
				
			case CFG_CLEAR_PIXELS_FUZZINESS_FACTOR:
				m_iClearPixelsFuzzinessFactor = iValue;
				break;
				
			case CFG_MAX_ADJACENT_POINTS_DISTANCE:
				m_iMaxAdjacentPointsDistance = iValue;
				break;
				
			case CFG_ROAD_BLOCK_MATCH_RATIO:
				m_iRoadBlockMatchRatio = iValue;
				break;
			
			case CFG_ADJACENCY_STEPS:
				m_iAdjacencySteps = iValue;
				break;
		}
	}
	
	/**
	 * Finds the points that are in the "core" of the road. These
	 * points can be used to build a graph of the roads
	 * 
	 * @param source - the source iamge
	 * 
	 * @return an array of road points
	 */
	public RoadPoint[] getRoadCorePoints( BufferedImage source, RecognitionInfo info )
	{
		// Step 1:
		// Filter the image by the given range of HSV values.
		// The produced image will be black and white image, 
		// where the black pixels will be the road points
		ByteImage workImage = applyRoughFilter( source, info );
		
		// Step 2:
		// Kill all noise pixels on the image
		workImage = clearNoisePixels( workImage );
		
		// Step 3:
		// Now we have a rough estimate of where the roads are.
		// We need to get only the points that are in the middle
		// of the black regions that look like wider roads
		RoadPoint[] roadPoints = findRoadCorePoints( workImage );
		
		return roadPoints;
	}
	
	/**
	 * Checks if a given color matches a road color. It checks
	 * if the sample pixel colors in the recognition info and the
	 * given color are within the given YCrCb delta of Y, Cr and Cb
	 * values 
	 * 
	 * @param c - the color to check
	 * @param info - recognition info
	 * @return true if the color matches a road color, false otherwise
	 */
	public boolean isRoadColor( Color c, RecognitionInfo info )
	{
		YCrCb sampleYCrCb = new YCrCb( );
		m_YCrCbWork.toYCrCb( c );
		
		// Check if there's any sample pixel that compared
		// to the given color in the YCrCb space is within
		// a given YCrCb delta 
		for ( int k = 0; k < info.samplePixels.size( ); k++ ) {
			Color sampleColor = ( Color ) info.samplePixels.get( k );
			sampleYCrCb.toYCrCb( sampleColor );
			if ( Math.abs( m_YCrCbWork.y  - sampleYCrCb.y )  <= info.ycrcbDelta.y  && 
				  Math.abs( m_YCrCbWork.cr - sampleYCrCb.cr ) <= info.ycrcbDelta.cr &&
				  Math.abs( m_YCrCbWork.cb - sampleYCrCb.cb ) <= info.ycrcbDelta.cb ) {
				return true;	 		
			}
		}
		
		return false;
	}
	
	/**
	 * Builds a graph from "adjacent" road points. We will call
	 * adjacent points those points that are closer than 
	 * m_iMaxAdjacentPointsDistance
	 * 
	 * @param points - an array of road points with unique IDs
	 * 
	 * @return a built graph
	 */
	public RoadGraph buildRoadGraph( RoadPoint[] rawPoints )
	{
		// Go through all points and if two points are "adjacent" 
		// and there's no road between them - add a connection 
		int iAdjacentDistance 	= 1;
		RoadGraph graph 			= new RoadGraph( );
		RoadPoint p1 				= null;
		RoadPoint p2 				= null;
		
		if ( rawPoints == null ) {
			return null;
		}

		// Add the points into the graph and later will add connections		
		for ( int i = 0; i < rawPoints.length; i++ ) {
			graph.addPoint( rawPoints[ i ] );
		}
		
		ArrayList[] neighbours 			= getClosestNeighbours( graph );
		RoadGraphOptimizer optimizer 	= new RoadGraphOptimizer( );
		ArrayList points					= graph.getPoints( );		
		int iAdjacencyStep 				= ( int )Math.ceil( m_iMaxAdjacentPointsDistance / ( float )m_iAdjacencySteps );
		int iSquareMaxAdjacentDistance= m_iMaxAdjacentPointsDistance * m_iMaxAdjacentPointsDistance;
		
		if ( iAdjacencyStep == 0 ) iAdjacencyStep = 1;
		while ( iAdjacentDistance <= m_iMaxAdjacentPointsDistance ) {
			int iSquareAdjacentDistance = iAdjacentDistance * iAdjacentDistance;
			
			m_ProgressBar.init( points.size( ), 1, 
			                    "Building road graph #" + iAdjacentDistance + "/" + m_iMaxAdjacentPointsDistance ); 
			for ( int i = 0; i < points.size( ); i++ ) {
				p1 = ( RoadPoint )points.get( i );
				
				for ( int j = 0; j < neighbours[ i ].size( ); j++ ) {
					// Check if the distance is reasonable so that
					// these points to be called "adjacent".
					// NOTE: Make a little performance trick. Check
					// if the distance on X and on Y separately is 
					// fine first and then calculate the real distance.
					p2 = ( RoadPoint )neighbours[ i ].get( j );
					
					long lCurrentSquareDistance = 0;
					
					if (  Math.abs( p1.x - p2.x ) <= iAdjacentDistance &&
					      Math.abs( p1.y - p2.y ) <= iAdjacentDistance &&
					      graph.hasDirectConnection( p1, p2 ) == false &&
					     ( lCurrentSquareDistance = RoadGraph.getDistanceSquare( p1, p2 ) ) <= iSquareAdjacentDistance ) {
						ArrayList road = graph.findShortestRoad( p1, p2 );
						if ( road == null ) { 
							graph.addConnection( p1, p2 );
						} else {
							long iRoadLength 			= graph.getRoadLength( road );
							long iRoadLengthSquare 	= iRoadLength * iRoadLength;    
							if ( iRoadLengthSquare > lCurrentSquareDistance ) {
								for ( int k = 0; k < road.size( ); k++ ) {
									RoadPoint p = ( RoadPoint )road.get( k );
									if ( RoadGraph.getDistanceSquare( p1, p ) > 2 * iSquareAdjacentDistance ) {
										graph.addConnection( p1, p2 );
										break;
									}
								}
							}
						}
					} // If Math.abs( ... ) < iAdjacent... 
				} // for j...
				
				m_ProgressBar.step( );
			} // for i...
			optimizer.optimizeGraph( graph );
			points = graph.getPoints( );
			neighbours = getClosestNeighbours( graph );
			iAdjacentDistance += iAdjacencyStep;
		}
		
		return graph;
	}
	
	/**
	 * Finds "neighbourhoods" of road points that are used
	 * to make faster operations over neighbourhoods, than over
	 * all points
	 * 
	 * @param g - the road graph
	 * 
	 * @return array of ArrayList elements of neighbourhoods. Each
	 *         index of the array reflects and index in the
	 * 		  returned array list of points by the graph
	 */
	private ArrayList[] getClosestNeighbours( RoadGraph g )
	{
		//	Organize the points into an array of points, each element
		// of which to be an array of "adjacent" points. The adjacent
		// points will be all points that are maximum m_iMaxAdjacentPointsDistance
		// pixels far. This way when running thorugh the points will
		// connect only points from the "adjacent" points array
		ArrayList points			= g.getPoints( );
		ArrayList[] neighbours 	= new ArrayList[ points.size( ) ];
		RoadPoint p1				= null;
		RoadPoint p2				= null;
		for ( int i = 0; i < points.size( ); i++ ) {
			neighbours[ i ] 	= new ArrayList( );
			p1 					= ( RoadPoint )points.get( i );
		
			for ( int j = 0; j < points.size( ); j++ ) {
				if ( i == j ) continue;
				p2 = ( RoadPoint )points.get( j );
				if ( Math.abs( p1.x - p2.x ) <= m_iMaxAdjacentPointsDistance &&
					  Math.abs( p1.y - p2.y ) <= m_iMaxAdjacentPointsDistance ) {
					neighbours[ i ].add( p2 );
				}
			}
		}		
		
		return neighbours;
	}
	
	/**
	 * This method makes all pixels that are matching the recognition info
	 * criteria black and all that are not - white. Uses a 
	 * ByteImage as an output
	 * 
	 * @param source - source image
	 * @param info - recognition info to use for applying the filter
	 * 
	 * @return ByteImage with black and white "pixels"
	 */	
	private ByteImage applyRoughFilter( BufferedImage source, RecognitionInfo info )
	{
		int iWidth  			= source.getWidth( );
		int iHeight 			= source.getHeight( );
		ByteImage destImage 	= new ByteImage( iWidth, iHeight );
		int iFoundPixels 		= 0;
		
		System.out.println( "Applying rough filter..." );

		if ( info.optimizedScanPortion == null ) {
			System.out.println( "Not using optimized scan portion" );
			m_ProgressBar.init( iWidth * iHeight, iHeight, "Rough filter" );
			for ( int i = 0; i < iWidth; i++ ) {
				for ( int j = 0; j < iHeight; j++ ) {
					Color c = new Color( source.getRGB( i, j ) );			
					if ( isRoadColor( c, info ) == true ) {
						iFoundPixels++;
						destImage.set( i, j, ByteImage.COLOR_BLACK );
					} else {
						destImage.set( i, j, ByteImage.COLOR_WHITE );
					}
				}
				m_ProgressBar.step( );
			}
		} else {
			int iHParts 		= ( int )Math.ceil( iWidth / info.optimizedScanPortion.width );
			int iVParts 		= ( int )Math.ceil( iHeight / info.optimizedScanPortion.height );
			int iEachWidth		= info.optimizedScanPortion.width;
			int iEachHeight	= info.optimizedScanPortion.height;
			System.out.println( "Using optimized scan portion: " + 
			                    iHParts + "x" + 
			                    iVParts + " parts, " + 
			                    iEachWidth + "x" + 
			                    iEachHeight + " each" );
			m_ProgressBar.init( iHParts * iVParts, 
			                    1, "Rough filter" );
			for ( int i = 0; i < iHParts; i++ ) {
				for ( int j = 0; j < iVParts; j++ ) {
					int iFromX	= i * iEachWidth;
					int iFromY 	= j * iEachHeight;
					int iToX		= iFromX + iEachWidth;	iToX = iToX >= iWidth ? iWidth : iToX;
					int iToY		= iFromY + iEachHeight; iToY = iToY >= iHeight ? iHeight : iToY;
					for ( int x = iFromX; x < iToX; x++ ) {
						for ( int y = iFromY; y < iToY; y++ ) {
							Color c = new Color( source.getRGB( x, y ) );			
							if ( isRoadColor( c, info ) == true ) {
								iFoundPixels++;
								destImage.set( x, y, ByteImage.COLOR_BLACK );
							} else {
								destImage.set( x, y, ByteImage.COLOR_WHITE );
							}							
						}
					}
					m_ProgressBar.step( );
				}
			}
		}
		
		System.out.println( "Found " + iFoundPixels + " matching pixels" );

		return destImage;
	}
	
	/**
	 * Clears the pixels that do not give us any information, but
	 * more often introduce some noise in the image. These are 
	 * lonely black pixels in the middle of the white, or lonely
	 * white pixels on the black spots.
	 * 
	 * @param source - the source image
	 * 
	 * @return a modified copy of the source image
	 */
	private ByteImage clearNoisePixels( ByteImage source )
	{
		int iWidth  = source.getWidth( );
		int iHeight = source.getHeight( );
		
		ByteImage destImage = new ByteImage( iWidth, iHeight );
		
		m_ProgressBar.init( iWidth * iHeight, iHeight, "Clear noise pixels" );
		
		for ( int i = 0; i < iWidth; i += m_iClearPixelsFuzzinessFactor ) {
			for ( int j = 0; j < iHeight; j += m_iClearPixelsFuzzinessFactor ) {
				byte color = source.get( i, j );
				if ( source.getEnergy( i, j, 
				                m_iClearPixelsFuzzinessFactor, 
				                ByteImage.COLOR_BLACK ) > m_iBlackSpotFuzzinessFactor ) {
					destImage.fill( i, j, 
									 m_iClearPixelsFuzzinessFactor, 
									 m_iClearPixelsFuzzinessFactor, 
					             ByteImage.COLOR_BLACK );
				} else {
					destImage.fill( i, j, 
					          m_iClearPixelsFuzzinessFactor, 
							    m_iClearPixelsFuzzinessFactor, 
								 ByteImage.COLOR_WHITE );
				}
			}
			
			m_ProgressBar.step( );
		}
		
		return destImage;
	}
	
	/**
	 * Finds the core points in the road spots. The state of the
	 * image at that step is assumed to be a black/white image
	 * with areas of black pixels that wind like roads .This
	 * method finds the points that are in the middle of these 
	 * black regions and adds them to a list of points. These
	 * points will be connected later to form a road graph
	 * 
	 * @param source
	 * 
	 * @return an array of "core" points
	 */
	private RoadPoint[] findRoadCorePoints( ByteImage source )
	{
		int iWidth  = source.getWidth( );
		int iHeight = source.getHeight( );
		
		ByteImage tmpImage = new ByteImage( source );
		
		int iAvgBlockWidth1 	= findAverageRoadBlockCoverSize( source );
		int iAvgBlockWidth2 	= findAverageRoadBlockCoverSize( source.getTransposed( ) );
		int iAvgBlockWidth	= ( int )Math.round( Math.ceil( ( iAvgBlockWidth1 + iAvgBlockWidth2 ) / 2.0 ) );
		
		System.out.println( "Avg block size: " + iAvgBlockWidth );

		ArrayList points = new ArrayList( );
		int iPointIndex  = 0;

		if ( iAvgBlockWidth > 0 ) {
			for ( int i = 0; i < iWidth; i++ ) {
				for ( int j = 0; j < iHeight; j++ ) {
					int iEnergy = source.getEnergy( i, j, iAvgBlockWidth, ByteImage.COLOR_BLACK );
					if ( ( iEnergy / ( iAvgBlockWidth * iAvgBlockWidth ) ) * 100 > m_iRoadBlockMatchRatio ) {
						source.fill( i, j, iAvgBlockWidth, iAvgBlockWidth, ByteImage.COLOR_WHITE );
						source.fill( i, j, 1, 1, ByteImage.COLOR_NEUTRAL );
						points.add( new RoadPoint( i, j ) );
					}
				}
			}
		}
		
		System.out.println( "Core points count: " + points.size( ) );
		
		if ( points.size( ) == 0 ) { return null; }
		
		RoadPoint[] roadPoints = new RoadPoint[ points.size( ) ];
		for ( int i = 0; i < points.size( ); i++ ) {
			roadPoints[ i ] = ( RoadPoint )points.get( i );
		}
		
		return roadPoints;
	}
	
	/**
	 * Finds the average size of a square block that covers all
	 * of the road if the center of the block is on the center
	 * of a road spot
	 * 
	 * @param source - the source ByteImage
	 * 
	 * @return the size of the square block
	 */
	private int findAverageRoadBlockCoverSize( ByteImage source ) 
	{
		int iCoverBlocksCount = 0;
		int iBlockAreaSum     = 0;
		int iWidth				 = source.getWidth( );
		int iHeight				 = source.getHeight( );

		m_ProgressBar.init( iWidth * iHeight, iHeight, "Finding road core points" );

		for ( int i = 0; i < iWidth; i++ ) {
			int iStartOfContour   = -1;
			for ( int j = 0; j < iHeight; j++ ) {
				// Find the beginning of a contour
				if ( source.get( i, j ) == ByteImage.COLOR_BLACK ) {
					if ( iStartOfContour == -1 ) {
						iStartOfContour = j;
					}
				} else {
					// We've reached to a white point right after
					// the start of the contour was found
					if ( iStartOfContour >= 0 ) {
						// This is the end of the contour. Go to the middle point and start
						// "getting fat" until the number of black pixels gets less than the last
						// time pixels
						int iLastBlockArea    = 0;
						int iCurrentBlockArea = 0;
						int iMiddle = ( j + iStartOfContour ) / 2;
						int iBlockWidth = 1;
						while ( true ) {
							int iX = i;
							int iY = iMiddle;
				
							iCurrentBlockArea = source.getEnergy( iX, iY, 
																  iBlockWidth, 
																  ByteImage.COLOR_BLACK );
							if ( iCurrentBlockArea < iBlockWidth * iBlockWidth ) { break; }
							iLastBlockArea = iCurrentBlockArea;
				
							iBlockWidth++;
						}
				
						iCoverBlocksCount++;
						iBlockAreaSum += iLastBlockArea;
					}
					iStartOfContour = -1;
				}
			}
	
			m_ProgressBar.step( );
		}

		// Finally the block that covers fine the road is found
		return ( int )Math.ceil( Math.sqrt( ( float )iBlockAreaSum / iCoverBlocksCount ) );		
	}
	
	/**
	 * @class lib.road
	 * 
	 * Implement a simple class that is a matrix of integers
	 * assumed to be a bitmap with few 1 byte colors
	 * 
	 * @author Tihomir Lazarov
	 * @date   May 14, 2005
	 */
	public class ByteImage
	{
		byte [][]m_Image = null;
		int m_iWidth     = 0;
		int m_iHeight    = 0;
		
		public static final byte COLOR_NEUTRAL	= 0;
		public static final byte COLOR_WHITE   = 1;
		public static final byte COLOR_BLACK   = 2;
		public static final byte COLOR_INVALID = -1;
		
		/**
		 * Constructor by dimentions
		 * 
		 * @param iWidth
		 * @param iHeight
		 */
		public ByteImage( int iWidth, int iHeight )
		{
			m_Image = new byte[ iWidth ][ iHeight ];
			for ( int i = 0; i < iWidth; i++ ) {
				for ( int j = 0; j < iHeight; j++ ) {
					m_Image[ i ][ j ] = COLOR_INVALID;
				}
			}
			
			m_iWidth  = iWidth;
			m_iHeight = iHeight;
		}
		
		/**
		 * Copy constructor
		 * @param tmp - the image to copy
		 */
		public ByteImage( ByteImage tmp )
		{
			this( tmp.getWidth( ), tmp.getHeight( ) );
			
			for ( int i = 0; i < tmp.getWidth( ); i++ ) {
				for ( int j = 0; j < tmp.getHeight( ) ; j++ ) {
					set( i, j, tmp.get( i, j ) );
				}
			}
		}
		
		/**
		 * Gets the color of a certain pixel
		 * 
		 * @param x - the x position
		 * @param y - the y position
		 * 
		 * @return the color on the (x, y) position
		 */
		public byte get( int x, int y )
		{
			if ( x >= 0 && x < m_iWidth && y >= 0 && y < m_iHeight )
				return m_Image[ x ][ y ];
			return COLOR_INVALID;
		}
		
		/**
		 * Sets a color on a certain position
		 * 
		 * @param x - position x
		 * @param y - position y
		 * @param c - the color to set
		 */
		public void set( int x, int y, byte c )
		{
			if ( x >= 0 && x < m_iWidth && y >= 0 && y < m_iHeight ) 
				m_Image[ x ][ y ] = c;
		}
		
		/**
		 * Fills a rectangular region from the image with a given color
		 * 
		 * @param x - the left corner of the region
		 * @param y - the top corner of the region
		 * @param w - the width of the region
		 * @param h - the height of the region
		 * @param c - the color to fill with
		 */
		public void fill( int x, int y, int w, int h, byte c )
		{
			if ( x >= 0 && x < m_iWidth && y >= 0 && y < m_iHeight ) {
				int toX = x + w > m_iWidth - 1 ? m_iWidth - 1 : x + w;
				int toY = y + h > m_iHeight - 1 ? m_iHeight - 1 : y + h;
				
				for ( int i = x; i <= toX; i++ ) {
					for ( int j = y; j <= toY; j++ ) {
						set( i, j, c );
					}
				}
			}
		}
		
		/**
		 * Returns the width of the image
		 * 
		 * @return the image width
		 */
		public int getWidth( )
		{
			return m_iWidth;
		}
		
		/**
		 * Returns the height of the image
		 * 
		 * @return the image height
		 */
		public int getHeight( ) 
		{
			return m_iHeight;
		}
		
		/**
		 * Gets the number of pixels in a matrix that are with a given
		 * color. The matrix has a 'size" size
		 * 
		 * @param x - the middle x of the matrix
		 * @param y - the middle y of the matrix
		 * @param size - the the size of the matrix
		 * @param color - the color to search for
		 * 
		 * @return the number of matrix cells with the given color
		 */
		public int getEnergy( int x, int y, int size, byte color )
		{
			int iiBegin = x < 0 ? 0 : x;
			int iiEnd   = x + size >= m_iWidth ? m_iWidth - 1 : x + size;
			int ijBegin = y < 0 ? 0 : y; 
			int ijEnd   = y + size >= m_iHeight ? m_iHeight - 1 : y + size;

			int iSum = 0;
			for ( int i = iiBegin; i <= iiEnd; i++ ) {
				for ( int j = ijBegin; j <= ijEnd; j++ ) {
					if ( get( i, j ) == color ) {
						iSum++;
					}
				}
			}

			return iSum;
		}
		
		/**
		 * Generates a transposed version of the current image and returns it
		 * 
		 * @return the transposed image
		 */
		public ByteImage getTransposed( )
		{
			ByteImage transposed = new ByteImage( m_iHeight, m_iWidth );
			
			for ( int i = 0; i < m_iWidth; i++ ) {
				for ( int j = 0; j < m_iHeight; j++ ) {
					transposed.set( j, i, get( i, j ) );
				}
			}
			
			return transposed;
		}

		/**
		 * Dumps the ByteImage into a file as a black and white image
		 * 
		 * @param szFileName - the file to store the image in
		 */		
		public void dump( String szFileName )
		{
			File file = new File( szFileName );
			BufferedImage img = new BufferedImage( m_iWidth, m_iHeight, BufferedImage.TYPE_INT_RGB );
			for ( int i = 0; i < m_iWidth; i++ ) {
				for ( int j = 0; j < m_iHeight; j++ ) {
					if ( get( i, j ) == COLOR_BLACK ) {
						img.setRGB( i, j, Color.BLACK.getRGB( ) );
					} else {
						img.setRGB( i, j, Color.WHITE.getRGB( ) );
					}
				}
			}
			try {
				String szExt = szFileName.substring( szFileName.lastIndexOf( '.' ) + 1 );
				ImageIO.write( img, szExt, file );
			} catch ( IOException e ) {
			}
		}
	}
}
