package lib.road.ext;

import java.sql.SQLException;
import java.sql.ResultSet;
import java.sql.PreparedStatement;

import java.util.ArrayList;
import java.util.Hashtable;

import lib.road.RoadGraph;
import lib.road.RoadPoint;
import lib.road.ConnectionPoint;
import lib.db.DBManager;

import lib.exceptions.DBManagerException;
import lib.exceptions.DBRoadGraphException;

/**
 * @class lib.road.ext
 * 
 * Extends the road graph functionality by adding storing to and
 * loading from a database 
 * 
 * @author Tihomir Lazarov
 * @date   Jun 3, 2005
 */
public class DBRoadGraph {
	DBManager m_DBManager 	= null;
	
	/**
	 * Constructor.
	 * 
	 * @param db - database manager
	 *
	 * 	 * @throws DBRoadGraphException
	 */
	public DBRoadGraph( DBManager db )
		throws DBRoadGraphException
	{
		m_DBManager		= db;
		
		if ( db == null  ) throw new DBRoadGraphException( "Null database manager", null );
	}
	
	/**
	 * Loads from database all graphs for a given map ID
	 * 
	 * @param iMapID - the map ID
	 * 
	 * @return an array list of RoadGraph elements or null if
	 *         there are no graphs
	 */
	public ArrayList load( int iMapID )
		throws DBRoadGraphException
	{
		try {
			ArrayList graphs = getGraphs( iMapID );
			
			// Check if there are any graphs
			if ( graphs == null ) { return null; }
			
			ArrayList roadGraphs = new ArrayList( );
			// Load each graph
			for ( int i = 0; i < graphs.size( ); i++ ) {
				DBGraph dbGraph		= ( DBGraph ) graphs.get( i );
				RoadGraph roadGraph 	= loadGraph( dbGraph.id );
				if ( roadGraph != null ) {
					roadGraphs.add( roadGraph );
				}
			}
			
			return roadGraphs.size( ) == 0 ? null : roadGraphs;
		} catch ( SQLException e ) {
			throw new DBRoadGraphException( "SQL exception: " + e.getMessage( ), e );
		} catch ( DBManagerException e ) {
			throw new DBRoadGraphException( "DB manager exception: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Loads a given graph from the database
	 * 
	 * @param iGraphID - the road ID
	 * 
	 * @return a RoadGraph element
	 */
	public RoadGraph loadGraph( int iGraphID )
		throws SQLException, DBManagerException
	{
		RoadGraph graph = new RoadGraph( );
		
		// Get the points for the graph
		ArrayList dbPoints = getPoints( iGraphID );
		
		// Check if there are any points in the graph
		if ( dbPoints == null ) { return graph; }
		
		// Remember the relation point DB ID => RoadPoint
		// so that connections may be inserted quicker
		Hashtable pointIDToPoint = new Hashtable( );
		
		// Insert the points in the graph and at the same time
		// get the connections for those points. The connections
		// will be set later
		Hashtable pointConnections = new Hashtable( );
		ArrayList pointIDs = new ArrayList( );
		
		for ( int i = 0; i < dbPoints.size( ); i++ ) {
			DBPoint dbPoint = ( DBPoint ) dbPoints.get( i );
			pointIDToPoint.put( new Integer( dbPoint.id ), dbPoint.point );
			graph.addPoint( dbPoint.point );
			
			// Buffer several point IDs into a group
			if ( pointIDs.size( ) > 100 ) {
				Hashtable currentConnections = getConnections( pointIDs );
				if ( currentConnections != null ) {
					// Add these connections to the list of connections we currently have
					pointConnections.putAll( currentConnections );
				}
				pointIDs.clear( );
			} 
				
			pointIDs.add( new Integer( dbPoint.id ) );
		}
		
		// Get the rest of the connections
		if ( pointIDs.size( ) > 0 ) {
			Hashtable currentConnections = getConnections( pointIDs );
			if ( currentConnections != null ) {
				// Add these connections to the list of connections we currently have
				pointConnections.putAll( currentConnections ); 
			}
		}
		
		// Now we've got a graph with inserted points and a bunch of connections.
		// For each point set the connections in the graphArrayList pointIDs = new ArrayList( );
		for ( int i = 0; i < dbPoints.size( ); i++ ) {
			DBPoint dbPoint = ( DBPoint ) dbPoints.get( i );
			ArrayList connectionPointIDs = ( ArrayList )pointConnections.get( new Integer( dbPoint.id ) );
			if ( connectionPointIDs != null ) {
				for ( int j = 0; j < connectionPointIDs.size( ); j++ ) {
					Integer iConnectionPointID = ( Integer ) connectionPointIDs.get( j );
					RoadPoint connectionPoint	= ( RoadPoint ) pointIDToPoint.get( iConnectionPointID );
					if ( connectionPoint != null ) {
						graph.addDirectedConnection( dbPoint.point, connectionPoint );
					}
				}
			}
		}
		
		return graph;
	}
	
	/**
	 * Saves the road graph to the database
	 * 
	 * @param iMapID - map ID which will own the road graph
	 * @param graphs - an array of road graphs to store
	 * 
	 * @throws DBRoadGraphException
	 */
	public void save( int iMapID, ArrayList graphs )
		throws DBRoadGraphException
	{
		if ( iMapID <= 0 )   throw new DBRoadGraphException( "Invalid map ID passed: " + iMapID, null );
		if ( graphs == null ) throw new DBRoadGraphException( "Null road graphs", null );
		
		try {
			//	Remove all road graphs for the current map
			removeGraphsForMap( iMapID );
			
			// Insert graphs
			for ( int i = 0; i < graphs.size( ); i++ ) {
				// Insert a graph record
				RoadGraph graph  	= ( RoadGraph ) graphs.get( i );
				int iGraphID 		= insertGraph( iMapID, graph );		
				
				// Insert the graph points
				ArrayList points = graph.getPoints( );
				
				// Store the relation point.id => point DB ID
				// so that the connection points can be easily inserted
				Hashtable pointsToID = new Hashtable( );
				
				for ( int j = 0; j < points.size( ); j++ ) {
					// Insert the current graph point
					RoadPoint point = ( RoadPoint ) points.get( j );
					int iPointID = insertPoint( iGraphID, point );
					pointsToID.put( new Integer( point.id ), new Integer( iPointID ) );
				}
				
				for ( int j = 0; j < points.size( ); j++ ) {
					RoadPoint point 	= ( RoadPoint ) points.get( j );
					int iPointID 		= ( ( Integer ) pointsToID.get( new Integer( point.id ) ) ).intValue( );
					
					// Insert the connection points
					for ( int k = 0; k < point.connections.size( ); k++ ) {
						ConnectionPoint connectionPoint = ( ConnectionPoint )point.connections.get( k );
						int iConnectionID  	= ( ( Integer ) pointsToID.get( new Integer( connectionPoint.id ) ) ).intValue( );
						insertConnection( iPointID, iConnectionID );
					}
				}
			}
		} catch ( SQLException e ) {
			throw new DBRoadGraphException( "SQL Exception while saving graph: " + e.getMessage( ), e );
		} catch ( DBManagerException e ) {
			throw new DBRoadGraphException( "DB Manager Exception whlie saving graph: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Removes all graphs for a given map ID
	 * 
	 * @param iMapID - the map ID
	 * 
	 * @throws DBRoadGraphException
	 */
	public void removeForMap( int iMapID )
		throws DBRoadGraphException
	{
		try {
			removeGraphsForMap( iMapID );
		} catch ( SQLException e ) {
			throw new DBRoadGraphException( "SQL Exception while removing graphs: " + e.getMessage( ), e );
		} catch ( DBManagerException e ) {
			throw new DBRoadGraphException( "DB Manager Exception while removing graphs: " + e.getMessage( ), e );
		}
	}
	
	/** -- PRIVATE METHODS -------------------------------------------------- */
	
	/**
	 * Removes the graphs for a given map ID
	 * 
	 * @param iMapID - the map ID
	 * 
	 * @throws SQLException
	 * @throws DBManagerException
	 */
	private void removeGraphs( int iMapID )
		throws SQLException, DBManagerException
	{
		PreparedStatement statement = 
			m_DBManager.getPreparedStatement( "DELETE FROM map_graphs WHERE MapID = ?" );
		statement.setInt( 1, iMapID );
		statement.execute( );
	}
	
	/**
	 * Removes the points for a given graph ID
	 * 
	 * @param iGraphID - the graph ID
	 * 
	 * @throws SQLException
	 * @throws DBManagerException
	 */
	private void removePoints( int iGraphID )
		throws SQLException, DBManagerException
	{
		PreparedStatement statement = 
			m_DBManager.getPreparedStatement( "DELETE FROM graph_points WHERE GraphID = ?" );
		statement.setInt( 1, iGraphID );
		statement.execute( );
	}
	
	/**
	 * Removes the connections for a given array of point IDs
	 * 
	 * @param pointIDs - the point IDs array
	 * 
	 * @throws SQLException
	 * @throws DBManagerException
	 */
	private void removeConnections( ArrayList pointIDs )
		throws SQLException, DBManagerException
	{
		if ( pointIDs == null || pointIDs.size( ) == 0 ) return;
		
		String szPointIDs = "";
		for ( int i = 0; i < pointIDs.size( ); i++ ) {
			szPointIDs += ( ( Integer )pointIDs.get( i ) ).intValue( ) + ","; 
		}
		if ( pointIDs.size( ) > 0 ) {
			// Remove the last ","
			szPointIDs = szPointIDs.substring( 0, szPointIDs.length( ) - 1 );
		}
		
		PreparedStatement statement = 
			m_DBManager.getPreparedStatement( 
			 "DELETE FROM point_connections WHERE PointID IN (" + szPointIDs + ")" 
			);
		statement.execute( );
	}
	
	/**
	 * Gets an array list of graphs for a given map ID
	 * 
	 * @param iMapID - the map ID
	 * 
	 * @return an array list of DBGraph elements. If there are no
	 *         graphs found, null is returned.
	 */
	private ArrayList getGraphs( int iMapID ) 
		throws SQLException, DBManagerException
	{
		PreparedStatement statement = m_DBManager.getPreparedStatement(
			"SELECT * FROM map_graphs WHERE MapID = ?"
		);
		statement.setInt( 1, iMapID );
		ResultSet records = statement.executeQuery( );
		ArrayList graphs  = new ArrayList( );
		
		while ( records.next( ) ) {
			RoadGraph graph 	= new RoadGraph( );
			DBGraph dbGraph 	= new DBGraph( records.getInt( "ID" ), graph );
			graphs.add( dbGraph );
		}
		
		return graphs.size( ) == 0 ? null : graphs;
	}
	
	/**
	 * Gets an array list of points for a given graph ID
	 * 
	 * @param iGraphID - the graph ID
	 * 
	 * @return an array list of DBPoint elements
	 */
	private ArrayList getPoints( int iGraphID )
		throws SQLException, DBManagerException
	{
		PreparedStatement statement = m_DBManager.getPreparedStatement(
			"SELECT * FROM graph_points WHERE GraphID = ? ORDER BY PointID"
		);
		statement.setInt( 1, iGraphID );
		ResultSet records = statement.executeQuery( );
		ArrayList points  = new ArrayList( );
		
		while ( records.next( ) ) {
			RoadPoint point = new RoadPoint( records.getInt( "X" ), 
			                                 records.getInt( "Y" ) );
			DBPoint dbPoint = new DBPoint( records.getInt( "ID" ), point );
			points.add( dbPoint );
		}
		
		return points.size( ) == 0 ? null : points;
	}
	
	/**
	 * Gets the connection points for a given array of point IDs
	 * 
	 * @param pointIDs - the array of point IDs
	 * 
	 * @return a Hashtable of ( key = point ID, value = ArrayList of point IDs 
	 *         which are the connection point IDs )
	 */
	private Hashtable getConnections( ArrayList pointIDs )
		throws SQLException, DBManagerException
	{
		if ( pointIDs == null || pointIDs.size( ) == 0 ) return null;
		
		Hashtable connections = new Hashtable( );
		
		String szPointIDs = "";
		for ( int i = 0; i < pointIDs.size( ); i++ ) {
			szPointIDs += ( ( Integer )pointIDs.get( i )  ).intValue( ) + ","; 
		}
		if ( pointIDs.size( ) > 0 ) {
			// Remove the last ","
			szPointIDs = szPointIDs.substring( 0, szPointIDs.length( ) - 1 );
		}
		
		PreparedStatement statement = m_DBManager.getPreparedStatement(
			"SELECT * FROM point_connections WHERE PointID IN (" + szPointIDs + ")"
		);
		ResultSet records 		= statement.executeQuery( );
		
		while ( records.next( ) ) {
			int iPointID		= records.getInt( "PointID" );
			int iConnectionID	= records.getInt( "ConnectionID" );
			ArrayList conns	= ( ArrayList ) connections.get( new Integer( iPointID ) );
			if ( conns == null ) { conns = new ArrayList( ); }
			conns.add( new Integer( iConnectionID ) );
			connections.put( new Integer( iPointID ), conns );
		}
		
		return connections.size() == 0 ? null : connections;
	}
	
	/**
	 * Inserts a graph record in the database
	 * 
	 * @param iMapID - the map ID to associate the graph with
	 * @param graph - the graph which information will be inserted
	 * 
	 * @return the inserted graph ID
	 */
	private int insertGraph( int iMapID, RoadGraph graph )
		throws SQLException, DBManagerException 
	{
		// Lock
		m_DBManager.execute( "LOCK TABLES map_graphs WRITE" );

		// Insert the record
		PreparedStatement statement = 
				m_DBManager.getPreparedStatement( 
					"INSERT INTO map_graphs ( MapID ) VALUES( ? )" 
				);
		statement.setInt( 1, iMapID ); 
		statement.execute( );
		
		// Get the last insert ID
		ResultSet result = m_DBManager.select( "SELECT LAST_INSERT_ID( ) AS GraphID" ); 
		result.next( );
	   int iGraphID = result.getInt( "GraphID" );
	   
		// Unlock
		m_DBManager.execute( "UNLOCK TABLES" );
		
		return iGraphID;		
	}
	
	/**
	 * Inserts a point record in the database
	 * 
	 * @param iGraphID - the graph ID to associate the point with
	 * @param point    - the point which information will be inserted
	 * 
	 * @return the inserted point ID
	 */
	private int insertPoint( int iGraphID, RoadPoint point )
		throws SQLException, DBManagerException 
	{
		// Lock
		m_DBManager.execute( "LOCK TABLES graph_points WRITE" );
	
		// Insert the record
		PreparedStatement statement = 
				m_DBManager.getPreparedStatement( 
					"INSERT INTO graph_points ( GraphID, X, Y, PointID ) VALUES( ?, ?, ?, ? )" 
				);
		statement.setInt( 1, iGraphID 	); 
		statement.setInt( 2, point.x 		);
		statement.setInt( 3, point.y 		);
		statement.setInt( 4, point.id 	);
		statement.execute( );
	
		// Get the last insert ID
		ResultSet result = m_DBManager.select( "SELECT LAST_INSERT_ID( ) AS PointID" ); 
		result.next( );
	   int iPointID = result.getInt( "PointID" );
	
		// Unlock
		m_DBManager.execute( "UNLOCK TABLES" );
	
		return iPointID;		
	}
	
	/**
	 * Inserts connections for from given point ID to a given connection point ID
	 * 
	 * @param iPointID 				- the source point ID
	 * @param iConnectionPointID 	- the destination point ID
	 * 
	 * @throws SQLException
	 * @throws DBManagerException
	 */
	private void insertConnection( int iPointID, int iConnectionPointID )
		throws SQLException, DBManagerException
	{
		PreparedStatement statement = 
			m_DBManager.getPreparedStatement(
				"INSERT INTO point_connections ( PointID, ConnectionID ) VALUES( ?, ? )"
			);
			
		statement.setInt( 1, iPointID );
		statement.setInt( 2, iConnectionPointID );
		statement.execute( );
	}
	
	/**
	 * Produces graph out of a given array list of points. 
	 * Extracts the connections for each point and creates
	 * a RoadGraph object where points are inserted and
	 * connected.
	 * 
	 * @param points - an array list of DBPoint elements
	 * 
	 * @return an initialized graph object instance
	 */
	private RoadGraph produceGraph( ArrayList points )
	{
		return null;
	}
	
	/**
	 * Removes all graphs, points, connections for a given map ID
	 * 
	 * @param iMapID - the map ID
	 */
	private void removeGraphsForMap( int iMapID )
		throws SQLException, DBManagerException
	{
		//	Get the old graphs. They will be needed for removing the points			
		ArrayList oldGraphs = getGraphs( iMapID );
		
		if ( oldGraphs != null ) {
			for ( int i = 0; i < oldGraphs.size( ); i++ ) {
				DBGraph oldGraph 		= ( DBGraph ) oldGraphs.get( i );
				ArrayList oldPoints  = getPoints( oldGraph.id );
			
				if ( oldPoints != null ) {
					// Remove the connections by groups of points to be faster
					ArrayList pointIDs = new ArrayList( );
					for ( int j = 0; j < oldPoints.size( ); j++ ) {
						if ( pointIDs.size( ) >= 100 ) { // 100 points per group
							removeConnections( pointIDs );
							pointIDs.clear( );
						}
						
						DBPoint point = ( DBPoint ) oldPoints.get( j );
						pointIDs.add( new Integer( point.id ) );
					}
					if ( pointIDs.size( ) > 0 ) {
						removeConnections( pointIDs );
					}
				}
			
				removePoints( oldGraph.id );
			}
		}
	
		//	Remove the graphs
		removeGraphs( iMapID );
	}
	
	
	/** -- PRIVATE CLASSES ------------------------------------------------- */
	
	/**
	 * @class lib.road.ext
	 * 
	 * An object defining a RoadGraph with an ID from the database
	 * 
	 * @author Tihomir Lazarov
	 * @date   Jun 3, 2005
	 */
	private class DBGraph
	{
		public RoadGraph graph;
		public int id;
		
		public DBGraph( int _id, RoadGraph _graph ) 
		{
			graph = _graph;
			id		= _id;
		}
	}
	
	/**
	 * @class lib.road.ext
	 * 
	 * An object defining a RoadPoint with an ID from the database
	 * 
	 * @author Tihomir Lazarov
	 * @date   Jun 3, 2005
	 */
	private class DBPoint
	{
		public RoadPoint point;
		public int id;
		
		public DBPoint( int _id, RoadPoint _point )
		{
			id 	= _id;
			point = _point;
		}
	}
}
