package lib.servlets;

import java.util.ArrayList;
import javax.servlet.*;

import lib.map.dynamic.DynamicMap;
import lib.map.dynamic.Area;
import lib.map.dynamic.Event;
import lib.map.base.Rectangle;
import lib.cfg.Constants;
import lib.exceptions.DynamicMapException;
import lib.exceptions.NullMapException;

/**
 * @class lib.servlets
 * 
 * Adds an area
 * 
 * @author Tihomir Lazarov
 * @date   Aug 22, 2004
 */
public class AddArea extends lib.servlets.ZoomMap {
	public void service( ServletRequest request, ServletResponse response )
			throws ServletException 
	{
		try {
			DynamicMap map = serviceBegin( request, response );
			
			if ( isLogged( request ) ) {
				addArea( request, map, true );
			} else {
				throw new ServletException( "This operation requires an authentication" );
			}
			serviceEnd( request, response, map, map.getSnapshot( ) );
			
		} catch ( NullMapException e ) {
			throw new ServletException( e.getMessage( ) );
		} catch ( DynamicMapException e ) {
			throw new ServletException( e.getMessage( ) );
		}
	}
	
	/**
	 * Adds an area to the current map. The peculiarity here is that the
	 * dimensions of the area are preserved as they are given by the user
	 * 
	 * @param request
	 * @param map
	 * @param bMakeCentered - if to make the area centered
	 * 
	 * @throws ServletException
	 * @throws DynamicMapException
	 */
	protected void addArea( ServletRequest request, DynamicMap map, boolean bMakeCentered ) 
				 throws ServletException, DynamicMapException 
	{
		String szAreaX   		= request.getParameter( Constants.AREA_ATTRIBUTE_X 				);
		String szAreaY   		= request.getParameter( Constants.AREA_ATTRIBUTE_Y 				);
		String szAreaWidth	= request.getParameter( Constants.AREA_ATTRIBUTE_WIDTH 			);
		String szAreaHeight	= request.getParameter( Constants.AREA_ATTRIBUTE_HEIGHT 			);
		String szURL         = request.getParameter( Constants.AREA_ATTRIBUTE_URL 				);
		String szName			= request.getParameter( Constants.AREA_ATTRIBUTE_NAME				);
		String szDescription = request.getParameter( Constants.AREA_ATTRIBUTE_DESCRIPTION 	);
		String szIconID		= request.getParameter( Constants.AREA_ATTRIBUTE_ICON_ID			);
		String szSubMap		= request.getParameter( Constants.AREA_ATTRIBUTE_SUB_MAP			);
	
		if ( szAreaX     == null || szAreaY      == null || 
		     szAreaWidth == null || szAreaHeight == null || 
		     szURL       == null || szIconID     == null || szName == null ) {
			throw new ServletException( "Insufficient arguments to add an area" );
		}
	
		int iAreaX, iAreaY, iAreaWidth, iAreaHeight;
	
		try {
			iAreaX 			= Integer.parseInt( szAreaX );
			iAreaY 			= Integer.parseInt( szAreaY );
			iAreaWidth		= Integer.parseInt( szAreaWidth );
			iAreaHeight		= Integer.parseInt( szAreaHeight );
		} catch ( NumberFormatException e ) { 		
			throw new ServletException( "Invalid position arguments while adding an area: " + e.getMessage( ) );
		}
		
		Rectangle viewer       = map.getViewerRegion( );
		Rectangle realSizeArea = map.scaleToRealSize( new Rectangle( iAreaX + viewer.x, 
		                                                             iAreaY + viewer.y, 
		                                                             iAreaWidth, 
		                                                             iAreaHeight ) );
		                                                             
		if ( bMakeCentered ) {
			realSizeArea.x -= iAreaWidth  / 2;
			realSizeArea.y -= iAreaHeight / 2;
		}
		
		// Preserve the given width/height
		Area area = new Area( realSizeArea.x, realSizeArea.y, iAreaWidth, iAreaHeight, 0 );
		area.addInfo( Constants.AREA_ATTRIBUTE_URL			, szURL 				);
		area.addInfo( Constants.AREA_ATTRIBUTE_NAME        , szName          );
		area.addInfo( Constants.AREA_ATTRIBUTE_DESCRIPTION	, szDescription 	);
		area.addInfo( Constants.AREA_ATTRIBUTE_ICON_ID		, szIconID 			);
		
		if ( szSubMap != null ) 
			area.addInfo( Constants.AREA_ATTRIBUTE_SUB_MAP		, szSubMap			);
		
		if ( szIconID != null ) area.addInfo( Constants.AREA_ATTRIBUTE_ICON_ID, szIconID );
		
		area.addEvent( new Event( Constants.EVENT_ON_CLICK, Constants.ACTION_OPEN_URL ) );
		
		// If there's a sub-map passed, add a 'on max zoom reached' event to the
		// added map and to the main area of the sub-map add an event
		// 'on min zoom reached' to point to the current area and to the current map
		if ( szSubMap != null && szSubMap != "" && szSubMap != "0" ) {
			System.out.println( "ADD AREA: Sub-map passed: " + szSubMap );
			int iSubMapID = Integer.parseInt( szSubMap );
			if ( iSubMapID > 0 ) {
				
				System.out.println( "ADD AREA: Adding a max-zoom-reached event for current area" );
				
				area.addEvent( new Event( Constants.EVENT_ON_MAX_ZOOM_REACHED, "" + iSubMapID ) );
				DynamicMap subMap = new DynamicMap( map.getEnvironment( ), iSubMapID );
				ArrayList areas = subMap.getAreas( );
				
				System.out.println( "ADD AREA: Searching for the 'unremovable' area for the sub-map" );
				
				// Find the area with the policy UNREMOVABLE and add an event to it
				for ( int i = 0; i < areas.size( ); i++ ) {
					Area subArea = ( Area ) areas.get( i );
					if ( subArea.getPolicy( ) == Area.POLICY_UNREMOVABLE ) {
						System.out.println( "ADD AREA: Unremovable area found!" );

						subArea.addEvent( 
							new Event( 
										  Constants.EVENT_ON_MIN_ZOOM_REACHED, 
                                "" + map.getMapID( ) 
                     ) 
						);
						
						System.out.println( "ADD AREA: Replacing with the updated version of the area!" );
						subMap.removeAreaByUniqueID( subArea.getUniqueID( ) );
						subMap.pushArea( subArea );
						
						subMap.save( );
						break;
					}
				}
			}
		}
		
		System.out.println( "Adding area: " + area.toString( ) );
		
		map.pushArea( area );
	}	
}
