package lib.servlets;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import javax.servlet.*;

import lib.map.dynamic.cfg.DynamicMapEnvironment;
import lib.map.dynamic.DynamicMap;
import lib.map.dynamic.Icons;
import lib.map.dynamic.Snapshot;
import lib.map.dynamic.Area;
import lib.map.ext.MapSelector;
import lib.map.base.Rectangle;
import lib.comm.AreaInfo;
import lib.comm.IconInfo;
import lib.comm.GroupInfo;
import lib.comm.MapDefinition;
import lib.comm.MapRequestInfo;
import lib.exceptions.NullMapException;
import lib.cfg.Config;
import lib.cfg.Constants;


/**
 * @class lib.servlets
 * 
 * Gets the current information about the map and sends it as a serializes
 * MapRequestInfo object
 * 
 * @author Tihomir Lazarov
 * @date   Aug 22, 2004
 */
public class GetInfo extends ZoomMap {
	public void service( ServletRequest request, ServletResponse response )
			throws ServletException 
	{
		try {
			DynamicMap map 		= serviceBegin( request, response );
			MapRequestInfo info 	= new MapRequestInfo( );
			Snapshot snapshot		= map.getSnapshot( );
			Rectangle viewer		= map.getViewerRegion( );
			
			String szInfoType		= request.getParameter( Constants.PARAM_INFO_TYPE );
			
			if ( szInfoType == null ) {
				throw new ServletException( "Missing info type parameter" );
			}
			
			int iInfoType	= Integer.parseInt( szInfoType );
			
			switch( iInfoType ) {
				/************************************************
				 * INIT REQUEST
				 ************************************************/
				case Constants.INFO_TYPE_INIT:
					System.out.println( "GetInfo: INIT request" );
					
					DynamicMapEnvironment env 	= map.getEnvironment( );
					Config config					= env.getConfig( );
					
					fillCommonInfo( request, map, snapshot, info );

					info.setZoomLevels( map.getZoomLevels( ) );					
					info.setMapID( map.getMapID( ) );
					
					String szPreferredAreaWidth	= config.getValue( Constants.PARAM_PREFERRED_AREA_WIDTH );
					String szPreferredAreaHeight	= config.getValue( Constants.PARAM_PREFERRED_AREA_HEIGHT );
					if ( szPreferredAreaWidth == null || szPreferredAreaHeight == null ) {
						System.out.println( "Missing preferred area dimensions in config" );
						throw new ServletException( "Failed to find preferred area dimensions in the config" );
					}
					
					info.setPreferredAreaWidth( Integer.parseInt( szPreferredAreaWidth ) );
					info.setPreferredAreaHeight( Integer.parseInt( szPreferredAreaHeight ) );
					
					// Get the current available icons
					ArrayList icons = env.getIcons( ).getIcons( );
					
					for ( int i = 0; i < icons.size( ); i++ ) {
						Icons.Icon icon 		= ( Icons.Icon ) icons.get( i );
						IconInfo iconInfo 	= new IconInfo( );
						iconInfo.id 			= icon.getID( );
						iconInfo.description	= icon.getDescription( );
						info.addIcon( iconInfo );
					}
					
					// Get the icon groups too and put them in a serializable structure
					ArrayList groups = env.getIcons( ).getGroups( );
					for ( int i = 0; i < groups.size( ); i++ ) {
						Icons.Group group 	= ( Icons.Group ) groups.get( i );
						GroupInfo groupInfo	= new GroupInfo( );
						groupInfo.id			= group.getID( );
						groupInfo.name			= group.getName( );
						ArrayList groupIcons	= group.getIcons( );
						for ( int j = 0; j < groupIcons.size( ); j++ ) {
							Icons.Icon groupIcon = ( Icons.Icon ) groupIcons.get( j );
							IconInfo iconInfo 	= new IconInfo( );
							iconInfo.id 			= groupIcon.getID( );
							iconInfo.description	= groupIcon.getDescription( );
							groupInfo.addIcon( iconInfo );
						}
						info.addGroup( groupInfo );
					}
					
					System.out.println( "GetInfo: getting the maps" );
					
					// Get the current available maps
					MapSelector mapSelector 	= new MapSelector( env.getDBManager() );
					ArrayList mapDefinitions 	= mapSelector.getMaps( );
					for ( int i = 0; i < mapDefinitions.size( ); i++ ) {
						MapDefinition mapDefinition = ( MapDefinition ) mapDefinitions.get( i );
						info.addMap( mapDefinition );
					}
				
					break;
				/************************************************
				 * COMMON REQUEST
				 ************************************************/					
				case Constants.INFO_TYPE_COMMON:
					System.out.println( "GetInfo: COMMON request" );
					fillCommonInfo( request, map, snapshot, info );
					break;

				/************************************************
				 * AREA REQUEST
				 ************************************************/					
				case Constants.INFO_TYPE_AREA:
					String szX 	= request.getParameter( Constants.PARAM_X );
					String szY 	= request.getParameter( Constants.PARAM_Y );
					info 			= new MapRequestInfo( );
					if ( szX != null && szY != null ) {
						int iX = Integer.parseInt( szX );
						int iY = Integer.parseInt( szY );
						Area area = map.findAreaByEventIdentifier( iX, iY, Constants.EVENT_ON_CLICK );
						if ( area != null && area.getPolicy( ) != Area.POLICY_UNREMOVABLE ) {
							System.out.println( "GetInfo: Found area on position " + 
							                    "(" + iX + "," + iY + "), " + 
							                    "event " + Constants.EVENT_ON_CLICK );
							info.addArea( convertToAreaInfo( map, area ) );
						} 
					} else {
						throw new ServletException( "Missing (X, Y) parameters when requesting an area" );
					}
					break;
				default:
					throw new ServletException( "Invalid info type passed: " + iInfoType );
			}

			// Serialize the map info object 			
			response.setContentType( "application/unknown" );
			ServletOutputStream os 		= response.getOutputStream( );
			ObjectOutputStream objOut 	= new ObjectOutputStream( os );
			objOut.writeObject( info );
			objOut.close( );
		} catch ( IOException e ) {
			System.out.println( "I/O exception: " + e.getMessage( ) );
			throw new ServletException( "I/O exception: " + e.getMessage( ) );
		} catch ( NullMapException e ) {
			System.out.println( "Null map exception: " + e.getMessage( ) );
			throw new ServletException( "Null map exception: " + e.getMessage( ) );
		}
	}
	
	/**
	 * Fills a MapRequestInfo with the 'common' information, which is the
	 * list of areas on the current map snapshot
	 * 
	 * @param request - the servlet request
	 * @param map - the current map state
	 * @param snapshot - the current snapshot
	 * @param info - the info that will be filled in
	 */
	private void fillCommonInfo( ServletRequest request, DynamicMap map, Snapshot snapshot, MapRequestInfo info ) 
	{
		// Get the current areas in the snapshot but skip the areas
		// that have "unremovable" policy
		if ( snapshot.areas != null && snapshot.areas.length > 0 ) {
			for ( int i = 0; i < snapshot.areas.length; i++ ) {
				Area area = snapshot.areas[ i ];
				if ( area.getPolicy() == Area.POLICY_UNREMOVABLE ) continue;
				info.addArea( convertToAreaInfo( map, area ) );
			}
		}		

		Rectangle viewer = map.getViewerRegion( );
		info.setViewerBounds( viewer.x, viewer.y, viewer.width, viewer.height );
		info.setZoomLevel( map.getCurrentZoomLevel( ) );
		info.setMapDimensions( map.getWidth( ), map.getHeight( ) );
		info.setRoadLength( snapshot.roadLength );
		info.setLogged( isLogged( request ) ); 
	}
	
	/**
	 * Extracts the information from an Area object that will
	 * be serialized later and puts it into an AreaInfo object
	 * which is serializable
	 * 
	 * @param map - the current map
	 * @param area - the area to convert
	 * 
	 * @return an AreaInfo object
	 */
	private AreaInfo convertToAreaInfo( DynamicMap map, Area area )
	{
		AreaInfo areaInfo 	= new AreaInfo( );
		Rectangle v				= map.getViewerRegion( );
		Rectangle a				= area.getRectangle( );
		Rectangle r				= map.scaleToZoomLevel( a );
		areaInfo.x				= r.x - v.x;
		areaInfo.y				= r.y - v.y;	
		areaInfo.width			= r.width; 	
		areaInfo.height		= r.height;	
		areaInfo.realWidth	= a.width;  // Preserve the original width
		areaInfo.realHeight	= a.height; // Preserve the original height
		areaInfo.id				= area.getUniqueID( );
		areaInfo.info			= area.getInfo( );
		return areaInfo;
	}
}
