package lib.servlets;

import java.util.ArrayList;

import javax.servlet.*;
import javax.servlet.http.*;

import lib.cfg.Constants;
import lib.map.base.Map;
import lib.map.base.Rectangle;
import lib.map.dynamic.DynamicMap;
import lib.map.dynamic.Snapshot;
import lib.map.dynamic.Area;
import lib.map.dynamic.Event;
import lib.map.dynamic.cfg.DynamicMapEnvironment;

import lib.exceptions.DynamicMapException;
import lib.exceptions.NullMapException;
import lib.exceptions.MapStateException;


/**
 * @class lib.servlets
 * 
 * Shows a portion of the current map
 * 
 * @author Tihomir Lazarov
 * @date   Aug 22, 2004
 */
public class ShowMap extends lib.servlets.ZoomMap {
	
	/**
	 * The method that's invoked on a request
	 * 
	 * @param request	- the request object
	 * @param response - the response object
	 */
	public void service( ServletRequest request, ServletResponse response )
		throws ServletException 
	{
		DynamicMap map		= super.serviceBegin( request, response );
		Snapshot snapshot	= null;
		
		String szEvent 	= request.getParameter( Constants.PARAM_EVENT );

		if ( szEvent != null ) {
			String szX	= request.getParameter( Constants.PARAM_X );
			String szY	= request.getParameter( Constants.PARAM_Y );
			int x       = szX == null ? 0 : Integer.parseInt( szX );
			int y 		= szY == null ? 0 : Integer.parseInt( szY );
			Area area 	= map.findAreaByEventIdentifier( x, y, szEvent );
	
			if ( area != null ) {
				System.out.println( "Area found for Event '" + szEvent + "': " + area.toString( ) );
				Event event = area.findEvent( szEvent );
				try {
					map 		= doMapAction( request, response, map, x, y, szEvent, area );
					snapshot	= map.getSnapshot( );
				} catch ( NullMapException e ) {
					throw new ServletException( e.getMessage( ) );
				} catch ( DynamicMapException e ) {
					throw new ServletException( e.getMessage( ) );
				} catch ( MapStateException e ) {
					throw new ServletException( e.getMessage( ) );
				}
			}
		}
		
		super.serviceEnd( request, response, map, snapshot );
	}
	
	
	/**
	 * Executes a map action, i.e. "move" or "zoom" over
	 * the map.
	 * 
	 * @param request	 - the ServletRequest object
	 * @param response - the ServletResponse
	 * @param x        - the X position of the event
	 * @param y        - the Y position of the event
	 * @param szEvent  - the event that takes place
	 * @param area     - the area that handles the event
	 * 
	 * @return a snapshot from the result after the
	 *         action over the map
	 * 
	 * @throws Map.MapStateException
	 * @throws Map.NullMapImageException
	 * @throws IOException
	 */
	public DynamicMap doMapAction(
								 ServletRequest request, 
								 ServletResponse response, 
	                      DynamicMap map,
								 int x, 
								 int y,
								 String szEvent, 
								 Area area ) 
			throws DynamicMapException,
			       MapStateException,
				    NullMapException {
	
		HttpServletResponse re 	= ( HttpServletResponse )response;
		Event event            	= area.findEvent( szEvent );
		
		if ( event != null ) {
			String szAction = event.getAction( );
			
			System.out.println( "-----------------------------------------------------" );
			System.out.println( "Executing action '" + szAction + "'" );
			
			Rectangle viewerRectangle = map.getViewerRegion( );
		
			if ( szAction.equals( Constants.ACTION_ZOOM_IN ) ) {
				x += viewerRectangle.width / 2;
				y += viewerRectangle.height / 2;
				System.out.println( "ACTION: Zooming in" );
				if ( !map.isZoomableIn( ) ) {
					System.out.println( "Max zoom reached" );
					/* Check if there's some action on reaching the max zoom */
					map = processMaxZoomReachedEvent( map, 
					                                  ( ( HttpServletRequest )request ).getSession( ), 
					                                  x, y );
				} else {
					map.zoom( Map.ZOOM_IN );
				}
			} else if ( szAction.equals( Constants.ACTION_ZOOM_OUT ) ) {
				x += viewerRectangle.width / 2;
				y += viewerRectangle.height / 2;
				
				System.out.println( "ACTION: Zooming out" );
				if ( !map.isZoomableOut( ) ) {
					System.out.println( "Min zoom reached" );
					map = processMinZoomReachedEvent( map,
					                                  ( ( HttpServletRequest ) request ).getSession( ),
					                                  x, y );
				} else {
					System.out.println( "===> Just zooming out" );
					map.zoom( Map.ZOOM_OUT );
				}
			} else if ( szAction.equals( Constants.ACTION_ZOOM_TO ) ) {
				String szZoomLevel = request.getParameter( Constants.PARAM_ZOOM_LEVEL );
				if ( szZoomLevel != null ) {
					int iZoomLevel = Integer.parseInt( szZoomLevel );
					map.zoomTo( iZoomLevel );
				}	
			} else if ( szAction.equals( Constants.ACTION_MOVE_LEFT ) ) {
				System.out.println( "ACTION: Moving LEFT" );
				map.translate( Map.TRANS_DIR_LEFT );
			} else if ( szAction.equals( Constants.ACTION_MOVE_UP ) ) {
				System.out.println( "ACTION: Moving UP" );
				map.translate( Map.TRANS_DIR_UP );
			} else if ( szAction.equals( Constants.ACTION_MOVE_RIGHT ) ) {
				System.out.println( "ACTION: Moving RIGHT" );
				map.translate( Map.TRANS_DIR_RIGHT );
			} else if ( szAction.equals( Constants.ACTION_MOVE_DOWN ) ) {
				System.out.println( "ACTION: Moving DOWN" );
				map.translate( Map.TRANS_DIR_DOWN );
			} else if ( szAction.equals( Constants.ACTION_ZOOM_IN_MOVE ) || 
			            szAction.equals( Constants.ACTION_ZOOM_OUT_MOVE ) ) {
				System.out.println( "ACTION: Zoom In/Out & Move" );
				
				Rectangle viewer = map.getViewerRegion( );
				int iOffsetX = x - viewer.width  / 2;
				int iOffsetY = y - viewer.height / 2;
				System.out.println( "Translating to offset (" + iOffsetX + ", " + iOffsetY + ") = (" + x + ", " + y + ") - (" + ( viewer.width / 2 ) + ", " + ( viewer.height / 2 ) + ")" );
				
				if ( szAction.equals( Constants.ACTION_ZOOM_OUT_MOVE ) ) {
					if ( !map.isZoomableOut( ) ) {
						System.out.println( "Min zoom reached" );
						map = processMinZoomReachedEvent( map,
														  ( ( HttpServletRequest ) request ).getSession( ),
													  		x, y );
					} else {
						map.translateWithOffset( iOffsetX, iOffsetY );
						map.zoom( Map.ZOOM_OUT );
					}				
			   } else {
					if ( !map.isZoomableIn( ) ) {
						System.out.println( "Max zoom reached" );
						/* Check if there's some action on reaching the max zoom */
						map = processMaxZoomReachedEvent( map, 
														  ( ( HttpServletRequest )request ).getSession( ), 
														  x, y );
					} else {
						map.translateWithOffset( iOffsetX, iOffsetY );
						map.zoom( Map.ZOOM_IN );
					}					
				}
			} else if ( szAction.equals( "OpenURL" ) ) {
				System.out.println( "ACTION: Open URL" );
			}
		}
	
		return map;
	}
			
			
	/**
	 * This method assumes that the max zoom level has been reached and checks if there's
	 * an area with an event 'on max zoom reached' that points to the next map ID to 
	 * be loaded and loads it.
	 * 
	 * @param map - the map
	 * @param session - the HTTP session object
	 * @param x - the X coordinate we're on
	 * @param y - the Y coordinate
	 * 
	 * @return the result map
	 */
	private DynamicMap processMaxZoomReachedEvent( DynamicMap map, HttpSession session, int x, int y )
		throws DynamicMapException, MapStateException
	{
		Area specialArea 	= map.findAreaByEventIdentifier( x, y, Constants.EVENT_ON_MAX_ZOOM_REACHED );
		if ( specialArea != null ) {
			Event event 		= specialArea.findEvent( Constants.EVENT_ON_MAX_ZOOM_REACHED );
			String szAction 	= event.getAction( );
	
			System.out.println( "*** Max zoom is reached. Loading map ID " + szAction );
	
			// Store the last environment
			DynamicMapEnvironment env = map.getEnvironment( );
			resetMap( session, map );
			map = new DynamicMap( env, Integer.parseInt( szAction ) );
		} else {
			System.out.println( "*** No areas with OnMaxZoomReached event were found on position (" + x + ", " + y + "). Just moving to position" );
			Rectangle viewer = map.getViewerRegion( );
			int iOffsetX = x - viewer.width  / 2;
			int iOffsetY = y - viewer.height / 2;
			map.translateWithOffset( iOffsetX, iOffsetY );			
		}		
		
		return map;
	}
	

	/**
	 * This method assumes that the min zoom level has been reached and checks if there's
	 * an area with an event 'on min zoom reached' that points to an 'upper' level map
	 * and a coordinate to translate to.
	 * 
	 * @param map - the map
	 * @param session - the HTTP session object
	 * @param x - the X coordinate we're on
	 * @param y - the Y coordinate
	 * 
	 * @return the result map
	 */
	private DynamicMap processMinZoomReachedEvent( DynamicMap map, HttpSession session, int x, int y )
		throws DynamicMapException, 
				 MapStateException,
	          NullMapException
	{
		Area specialArea 	= map.findAreaByEventIdentifier( x, y, Constants.EVENT_ON_MIN_ZOOM_REACHED );
		if ( specialArea != null ) {
			Event event 		= specialArea.findEvent( Constants.EVENT_ON_MIN_ZOOM_REACHED );
			String szAction 	= event.getAction( );
			System.out.println( "*** Min zoom is reached. Opening " + szAction );
	
			/* Get the file name of the definition of the map */
			int iMapID = map.getMapID( );

			// Remember the old environment settings
			DynamicMapEnvironment env = map.getEnvironment( );
			resetMap( session, map );	
			map = new DynamicMap( m_Env, Integer.parseInt( szAction ) );
			
			/* Zoom the map to the biggest zoom. Find the area
			 * that leads to the former map. Move the viewer
			 * region in the middle of that area.
			 */
			int iMaxZoomLevel = map.getZoomLevels( ) - 1;
			System.out.println( "Zooming to the max zoom level of '" + szAction + "': " + iMaxZoomLevel );
			map.zoomTo( iMaxZoomLevel );
			
			Rectangle viewerRectangle = map.getViewerRegion( );
	
			System.out.println( "Getting the area that leads to the former map..." );
			ArrayList allAreas = map.getAreas( );
			Area zoomSpotArea  = null; 
	
			for ( int i = 0; i < allAreas.size( ); i++ ) {
				Area currentArea = ( Area )allAreas.get( i );
				Event currentEvent = currentArea.findEvent( Constants.EVENT_ON_MAX_ZOOM_REACHED );
				if ( currentEvent != null && currentEvent.getAction( ).equals( "" + iMapID ) ) {
					/* That's the area */
					zoomSpotArea = currentArea;
					System.out.println( "Area found! " + zoomSpotArea.toString( ) );
					break;
				}
			}
	
			if ( zoomSpotArea != null ) {
				Rectangle zoomSpotAreaRectangle = zoomSpotArea.getRectangle( );
				System.out.println( "Translating to (" + zoomSpotAreaRectangle.x + ", " + zoomSpotAreaRectangle.y + ")" );
				map.translateTo( zoomSpotAreaRectangle.x + zoomSpotAreaRectangle.width  / 2 - viewerRectangle.width  / 2, 
								     zoomSpotAreaRectangle.y + zoomSpotAreaRectangle.height / 2 - viewerRectangle.height / 2 );
			}
		} else {
			Rectangle viewer = map.getViewerRegion( );
			int iOffsetX = x - viewer.width  / 2;
			int iOffsetY = y - viewer.height / 2;
			map.translateWithOffset( iOffsetX, iOffsetY );			
		}
		
		return map;		
	}
}
