package lib.user;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import lib.db.DBManager;
import lib.cfg.Constants;
import lib.exceptions.DBManagerException;
import lib.exceptions.AuthenticationException;

/**
 * @class lib.user
 * 
 * A module for authenticating users
 * 
 * @author Tihomir Lazarov
 * @date   Jun 11, 2005
 */
public class Authentication 
{
	DBManager m_DBManager = null;
	
	/**
	 * Constructs a DBManager object
	 * 
	 * @param db - a reference to a DBManager object
	 */
	public Authentication( DBManager db )
	{
		m_DBManager = db;
	}
	
	/**
	 * Authenticates a given login-password pair 
	 * 
	 * @param szLogin - the login
	 * @param szPassword - the password
	 * 
	 * @return true if the authentication has been successful. 
	 *         False otherwise
	 */
	public boolean authenticate( String szLogin, String szPassword )
		throws AuthenticationException
	{
		try {
			PreparedStatement statement = 
				m_DBManager.getPreparedStatement( 
					"SELECT * FROM users WHERE Login = ? AND Password = PASSWORD( ? )"
				);
			statement.setString( 1, szLogin 		);
			statement.setString( 2, szPassword 	);
			ResultSet records = statement.executeQuery( );
			return records.next( ); 
		} catch ( DBManagerException e ) {
			throw new AuthenticationException( "Authenticate: DBManager exception: " + e.getMessage( ), e );
		} catch ( SQLException e ) {
			throw new AuthenticationException( "Authenticate: SQL exception: " + e.getMessage( ), e );
		}
	}
	
	/**
	 * Sets a record in the session identifying the owner of the
	 * session as successfully authenticated
	 * 
	 * @param request - the servlet request object
	 */
	public void setAuthenticated( ServletRequest request )
	{
		HttpServletRequest req 	= ( HttpServletRequest )request;
		HttpSession session    	= req.getSession( );
		session.setAttribute( Constants.SESSION_PARAM_AUTH, "1" );
	}
	
	/**
	 * Sets the session state to "logged out"
	 * 
	 * @param request - the servlet request parameter
	 */
	public void setLoggedOut( ServletRequest request )
	{
		HttpServletRequest req 	= ( HttpServletRequest )request;
		HttpSession session    	= req.getSession( );
		session.removeAttribute( Constants.SESSION_PARAM_AUTH );
	}

	/**
	 * Checks if there's a session parameter SESSION_PARAM_AUTH 
	 * and if so, this means that the user has been already successfully
	 * authenticated
	 * 
	 * @param request - the servlet request object
	 * 
	 * @return true if the user is successfully authenticated, false otherwise
	 */	
	public boolean isAuthenticated( ServletRequest request )
	{
		HttpServletRequest req 	= ( HttpServletRequest )request;
		HttpSession session    	= req.getSession( );
		if ( session == null ) return false;
		return ( session.getAttribute( Constants.SESSION_PARAM_AUTH ) != null );
	}
}
