package tools.RoadSelector;

import javax.swing.*;

import java.awt.event.ActionListener;
import java.awt.event.MouseListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.awt.Color;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Label;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.awt.FlowLayout;
import java.awt.BorderLayout;
import java.awt.image.*;

import java.util.ArrayList;


import lib.map.base.colors.YCrCb;

import lib.db.DBManager;
import lib.exceptions.ConfigException;
import lib.exceptions.DBManagerException;
import lib.exceptions.DBRoadGraphException;
import lib.exceptions.MapStateException;
import lib.exceptions.NullMapException;
import lib.cfg.Config;
import lib.comm.MapDefinition;

import lib.map.base.Map;
import lib.map.ext.MapSelector;
import lib.map.ext.MappedBufferedImage;

import lib.road.RoadRecognition;
import lib.road.RoadPoint;
import lib.road.ConnectionPoint;
import lib.road.RoadGraph;
import lib.road.RecognitionInfo;
import lib.road.ext.DBRoadGraph;

import lib.utils.Profiler;
import lib.utils.ProgressBar;

/**
 * @class tools
 * 
 * A user interface for getting the color of the road
 * on a bitmap and generating a graph of all the roads
 * on the map. This graph is stored in the database.
 * 
 * @author Tihomir Lazarov
 * @date   Feb 27, 2005
 */
public class RoadSelectorApplet extends JApplet 
										  implements ActionListener, 
										             MouseListener
{
	// The source image
	BufferedImage 	m_SourceImage 			= null;
	
	// The information needed to start recognition
	RecognitionInfo m_RecognitionInfo   = null;
	
	// The offset values are used for navigating over the original map
	int            m_iHOffset           = 0;
	int            m_iVOffset           = 0;
	
	// The array of all road points    
	RoadPoint[] m_TotalRoadPoints       = null;
	RoadPoint[] m_LastRoadPoints        = null;
	
	// Progress bar
	ProgressBar m_ProgressBar				= null;
	
	// The step operations that are about to take place
	static final int STEP_SELECT_ROAD_COLOR      = 2;
	static final int STEP_PREVIEW                = 3;
	static final int STEP_FINISH                 = 4;
	
	// The current step in the operations
	int m_iStep = STEP_SELECT_ROAD_COLOR;

	// Recognition process buttons	
	JButton m_SelectRoadButton		 		= new JButton( "Select road" );
	JButton m_PreviewOKButton           = new JButton( "Preview OK" );
	JButton m_PreviewFailedButton       = new JButton( "Preview FAILED" );
	JButton m_MoreButton				      = new JButton( "More roads..." );
	JButton m_DoneButton                = new JButton( "Done" );
	JButton m_ResetButton               = new JButton( "Reset" );
	
	JButton m_ZoomInButton					= new JButton( "Zoom In (+)" );
	JButton m_ZoomOutButton					= new JButton( "Zoom Out (-)" );
	
	// Navigation buttons
	JButton m_UpButton                   = new JButton( "Up" );
	JButton m_DownButton                 = new JButton( "Down" );
	JButton m_LeftButton                 = new JButton( "Left" );
	JButton m_RightButton                = new JButton( "Right" );
	
	// Canvas
	Label m_Canvas                      = new Label( "" );
	
	// Maps list
	JComboBox m_MapsListChoice				= new JComboBox( );
	JButton m_LoadMapButton					= new JButton( "Load Map" );
	
	// DB manager and config
	DBManager m_DBManager					= null;
	Config m_Config							= null;
	
	Map m_Map									= null;
	
	ArrayList m_MapsList						= null;
	
	String m_szDBHost							= "";
	String m_szDBName							= "";
	String m_szDBUser							= "";
	String m_szDBPass							= "";
	
	
	/**
	 * Initializes the applet and arranges the controls
	 */
	public void init() {
		super.init();
		setSize( 640, 480 );
		
		m_szDBHost = getParameter( "DBHost" );
		m_szDBName = getParameter( "DBName" );
		m_szDBUser = getParameter( "DBUser" );
		m_szDBPass = getParameter( "DBPass" );
		
		if ( m_szDBHost == null ) m_szDBHost = "";
		if ( m_szDBName == null ) m_szDBName = "";
		if ( m_szDBUser == null ) m_szDBUser = "";
		if ( m_szDBPass == null ) m_szDBPass = "";
		
		notifyUser( m_szDBHost + ":" + m_szDBName + ":" + m_szDBUser + ":" + m_szDBPass );
		
		// Set the following layout:
		// 
		//        [...map name.. |V] [Load map]
		//
		// [step1] [OK] [FAILED] [more] [done] [reset]
		// 
		//  [ +  ]       [ - ]
		//         [up]
		// [left]   img        [right]
		//         [down]
		//
		getContentPane( ).setLayout(new BorderLayout( ) );
		
		// The panel with buttons for managing the process of recognition
	   JPanel recognitionPanel     = new JPanel(new FlowLayout());
	   
	   // The navigation panel
	   JPanel navigationPanel = new JPanel( new BorderLayout( ) );
	   
	   // The zoom panel
	   JPanel zoomPanel = new JPanel( new FlowLayout( ) );
	   
	   // The maps panel
	   JPanel mapsPanel = new JPanel( new FlowLayout( ));

		// The base control panel where the navigation, zoom, maps
		// panels stay
	   JPanel controlsPanel = new JPanel( new BorderLayout( ) );
			   
		// Add the buttons for the recognition management
		recognitionPanel.add( m_SelectRoadButton );
		recognitionPanel.add( m_PreviewOKButton );
		recognitionPanel.add( m_PreviewFailedButton );
		recognitionPanel.add( m_MoreButton );
		recognitionPanel.add( m_DoneButton );
		recognitionPanel.add( m_ResetButton );
		
		// Add the navigation buttons
		navigationPanel.add( "North", m_UpButton );
		navigationPanel.add( "South", m_DownButton );
		navigationPanel.add( "East" , m_RightButton );
		navigationPanel.add( "West" , m_LeftButton );
		
		// Add the canvas
		navigationPanel.add( "Center", m_Canvas );
		
		// Add the zoom buttons
		zoomPanel.add( "West", m_ZoomInButton );
		zoomPanel.add( m_ZoomOutButton );
		
		// Organize the controls in the wizard and the map selector
		mapsPanel.add( m_MapsListChoice );
		mapsPanel.add( m_LoadMapButton );
		
		controlsPanel.add( "North" , mapsPanel );
		controlsPanel.add( "Center", recognitionPanel );
		controlsPanel.add( "South" , zoomPanel );
		
	   // Add the panels to the applet
	   getContentPane( ).add( "North", controlsPanel );
	   getContentPane( ).add( navigationPanel );
		
		
		// Add action listeners on the buttons and over the canvas
		m_SelectRoadButton.addActionListener( this );
		m_MoreButton.addActionListener( this );
		m_PreviewOKButton.addActionListener( this );
		m_PreviewFailedButton.addActionListener( this );
		m_DoneButton.addActionListener( this );
		m_ResetButton.addActionListener( this );
		
		// Navigatoin buttons
		m_UpButton.addActionListener( this );
		m_DownButton.addActionListener( this );
		m_RightButton.addActionListener( this );
		m_LeftButton.addActionListener( this );
		
		// Zoom buttons
		m_ZoomInButton.addActionListener( this );
		m_ZoomOutButton.addActionListener( this );
		
		m_LoadMapButton.addActionListener( this );
		
		m_Canvas.addMouseListener( this );

		// Reset all operations and controls
		resetOperations( );
		
		// Initialize the progress bar
		m_ProgressBar = new ProgressBar( this );
		
		System.out.println( "Loading maps..." );
	   loadMaps( );
	   System.out.println( "Loaded." );
	   
	   setVisible( true );
	}
	
	public void start( )
	{
		repaint( );
	}
	
	/**
	 * Loads the available maps from the database
	 */
	private void loadMaps( )
	{
		try {
			m_DBManager = new DBManager( m_szDBHost, m_szDBName, m_szDBUser, m_szDBPass );
			m_Config		= new Config( m_DBManager );
			
			MapSelector mapSelector = new MapSelector( m_DBManager );
			m_MapsList = mapSelector.getMaps( );
			
			if ( m_MapsList != null ) {
				m_MapsListChoice.removeAllItems( );
				for ( int i = 0; i < m_MapsList.size( ); i++ ) {
					MapDefinition mapDefinition = ( MapDefinition ) m_MapsList.get( i );
					m_MapsListChoice.addItem( mapDefinition );
				}
			} else {
				notifyUser( "Empty list of maps returned" );
			}
		} catch ( DBManagerException e ) {
			notifyUser( "DB exception: " + e.getMessage( ) ); 		
		} catch ( ConfigException e ) {
			notifyUser( "Configuratoin exception: " + e.getMessage( ) );
		}
	}
	
	/**
	 * Loads a map that's selected in the choice component where
	 * the list of maps stay
	 */
	private void loadSelectedMap( )
	{
		int iSelectedMapID 				= 0;
		MapDefinition mapDefinition	= ( MapDefinition ) m_MapsListChoice.getSelectedItem( );
		
		iSelectedMapID = mapDefinition.id;
		
		// Check if the selected map ID is OK
		if ( iSelectedMapID > 0 ) {
			m_Map = new Map( );
			try {
				System.out.println( "Trying to load map..." );
				m_Map.load( m_DBManager, m_Config, iSelectedMapID, 
				            m_Canvas.getWidth( ), m_Canvas.getHeight( ) );
				m_Map.zoomTo( m_Map.getZoomLevels( ) - 1 );
				m_SourceImage = m_Map.getViewableRegionSnapshot( );
				System.out.println( "Loaded" );
			} catch ( MapStateException e ) {
				notifyUser( "Map exception: " + e.getMessage( ) );
			} catch ( NullMapException e ) {
				notifyUser( "Null map exception: " + e.getMessage( ) );
			}
		} else {
			notifyUser( "Invalid map selected" );
		}
	}
	
	/**
	 * Handles actions over buttons
	 * 
	 * @param event - the event that's performed
	 */
	public void actionPerformed( ActionEvent event )
	{
		try {
			// UP button click
			if ( event.getSource( ) == m_UpButton ) {
				m_Map.translate( Map.TRANS_DIR_UP );
				m_SourceImage = m_Map.getViewableRegionSnapshot( ); 
				repaint( );
			//	DOWN button click
			} else if ( event.getSource( ) == m_DownButton ) {
				m_Map.translate( Map.TRANS_DIR_DOWN );
				m_SourceImage = m_Map.getViewableRegionSnapshot( ); 
				repaint( );
			//	RIGHT button click
			} else if ( event.getSource( ) == m_RightButton ) {
				m_Map.translate( Map.TRANS_DIR_RIGHT );
				m_SourceImage = m_Map.getViewableRegionSnapshot( ); 
				repaint( );
			//	LEFT button click			
			} else if ( event.getSource( ) == m_LeftButton ) {
				m_Map.translate( Map.TRANS_DIR_LEFT );
				m_SourceImage = m_Map.getViewableRegionSnapshot( ); 
				repaint( );
			// ZOOM IN button click
			} else if ( event.getSource( ) == m_ZoomInButton ) {
				m_Map.zoom( Map.ZOOM_IN );
				m_SourceImage = m_Map.getViewableRegionSnapshot( ); 
				repaint( );
			// ZOOM OUT button click
			} else if ( event.getSource( ) == m_ZoomOutButton ) {
				m_Map.zoom( Map.ZOOM_OUT );
				m_SourceImage = m_Map.getViewableRegionSnapshot( ); 
				repaint( );
				
			// SELECT ROAD button click
			} else if ( event.getSource( ) == m_SelectRoadButton ) {
				// Process the preview image and show the graph with recognized points
				RoadPoint[] points = processImage( new MappedBufferedImage( m_SourceImage ), m_RecognitionInfo );
				m_SourceImage = visualizeGraph( points, m_SourceImage );
				m_LastRoadPoints = points;
				m_iStep = STEP_PREVIEW;
				setControlsState( );
				repaint( );
			// PREVIEW OK button click
			} else if ( event.getSource( ) == m_PreviewOKButton ) {
				// Add the last road points to all road points
				if ( m_LastRoadPoints != null ) {
					if ( m_TotalRoadPoints == null ) {
						m_TotalRoadPoints = new RoadPoint[ m_LastRoadPoints.length ];
						for ( int i = 0; i < m_LastRoadPoints.length; i++ ) {
							m_TotalRoadPoints[ i ] = m_LastRoadPoints[ i ];
						}
					} else {
						RoadPoint[] tmp = m_TotalRoadPoints;
						m_TotalRoadPoints = new RoadPoint[ tmp.length + m_LastRoadPoints.length ];
						for ( int i = 0; i < tmp.length; i++ ) {
							m_TotalRoadPoints[ i ] = tmp[ i ];
						}
						for ( int j = 0; j < m_LastRoadPoints.length; j++ ) {
							m_TotalRoadPoints[ tmp.length + j ] = m_LastRoadPoints[ j ];
						}
						m_SourceImage = visualizeGraph( m_TotalRoadPoints, m_Map.getViewableRegionSnapshot() );
					}
				}
				
				m_LastRoadPoints = null;				
				
				notifyUser( "Visualizing image..." );
				
				m_iStep = STEP_FINISH;
				setControlsState( );
				repaint( );
				notifyUser( "Done." );
			// PREVIEW FAILED button click
			} else if ( event.getSource( ) == m_PreviewFailedButton ) {
				// Let the user try again with selecting the road
				m_TotalRoadPoints = m_LastRoadPoints;
				m_iStep           = STEP_SELECT_ROAD_COLOR;
				setControlsState( );
				repaint( );
			// MORE button click
			} else if ( event.getSource( ) == m_MoreButton ) {
				// Add more roads
				m_SourceImage        = m_Map.getViewableRegionSnapshot();
				m_iStep 					= STEP_SELECT_ROAD_COLOR;
				m_RecognitionInfo.reset( );
				setControlsState();
				repaint( );
			// DONE button click
			} else if ( event.getSource( ) == m_DoneButton ) {
				// Set the size of the sub-image in the big image so the
				// scanning will be faster
				m_RecognitionInfo.optimizedScanPortion = m_Map.getSubImageSize( );
		
				RoadPoint[] points = processImage( new MappedBufferedImage( m_Map ), m_RecognitionInfo );
				RoadRecognition rr = new RoadRecognition( m_ProgressBar );
				RoadGraph graph    = rr.buildRoadGraph( points );

				DBRoadGraph dbGraph = new DBRoadGraph( m_DBManager );
				
				ArrayList allGraphs = new ArrayList( );
				allGraphs.add( graph );
				dbGraph.save( m_Map.getMapID( ), allGraphs );
				
				destroy();
			// RESET button click
			} else if ( event.getSource( ) == m_ResetButton ) {
				resetOperations( );
				repaint( );
			} else if ( event.getSource( ) == m_LoadMapButton ) {
				loadSelectedMap( );
				repaint( );
			}
		} catch ( MapStateException e ) {
			notifyUser( "Exception caught: " + e.getMessage( ) );
		} catch ( NullMapException e ) {
			notifyUser( "Exception caught: " + e.getMessage( ) );
		} catch ( DBRoadGraphException e ) {
			notifyUser( "DB road graph exception: " + e.getMessage( ) );
		}
	}
	
	/** 
	 * Handles the pressing of the mouse left button
	 * 
	 * @param event - the event that occurs
	 */
	public void mousePressed( MouseEvent event ){}
	public void mouseReleased( MouseEvent event ){ }
	
	/**
	 * Handles the releasing of the mouse left button
	 * 
	 * @param event - the event that occurs
	 */
	public void mouseClicked( MouseEvent event )
	{
		switch ( m_iStep ) {
			case STEP_SELECT_ROAD_COLOR: {
				int x = event.getX( );
				int y = event.getY( );

				// Get the currently clicked pixel and add it to the list
				// of pixels that must be matched. Also set the recognition
				// information that will be needed when processing the image				
				Color color = new Color( m_SourceImage.getRGB( x, y ) );
				int red        = color.getRed( );
				int green      = color.getGreen( );
				int blue       = color.getBlue( );
											  
				if ( event.getButton( ) != MouseEvent.BUTTON1 ) {
					System.out.println( "Removing the last added color" );
					if ( m_RecognitionInfo.samplePixels.size( ) > 0 ) 
						m_RecognitionInfo.samplePixels.remove( m_RecognitionInfo.samplePixels.size( ) - 1 );
				} else {
					m_RecognitionInfo.samplePixels.add( color );
				}
				
				YCrCb yy = new YCrCb( );
				yy.toYCrCb( color );
				notifyUser( "RGB (" + red + ", " + green + ", " + blue + ") = YCbCr(" + yy.y + ", " + yy.cb + ", " + yy.cr + ")" );

				m_RecognitionInfo.ycrcbDelta.y  = 20;
				m_RecognitionInfo.ycrcbDelta.cr = 10;
				m_RecognitionInfo.ycrcbDelta.cb = 10;
				
				repaint( );
				break;
			}
		}
	}
	
	public void mouseEntered( MouseEvent event ){}
	public void mouseExited( MouseEvent event ){}
	
	/**
	 * Shows a message on the status bar
	 * 
	 * @param szMessage
	 */
	private void notifyUser( String szMessage )
	{
		System.out.println( "*** " + szMessage );
		showStatus( szMessage );
	}

	/** 
	 * Resets all operations and goes back to step 1
	 */	
	private void resetOperations( )
	{
		try {
			m_SourceImage        = m_Map == null ? null : m_Map.getViewableRegionSnapshot( );
			m_iStep           	= STEP_SELECT_ROAD_COLOR;
			m_iHOffset           = 0;
			m_iVOffset           = 0;
			m_TotalRoadPoints    = null;
			m_LastRoadPoints     = null;
			m_RecognitionInfo		= new RecognitionInfo( );
		
			setControlsState( );
		} catch ( NullMapException e ) {
			notifyUser( "Failed resetting operations: " + e.getMessage( ) );
		}
	}
	
	/**
	 * Sets the states of the controls on the screen depending on 
	 * the current step
	 */
	private void setControlsState()
	{
		switch ( m_iStep )
		{
			// During the pass of selecting a road color make only 
			// the second button enabled
			case STEP_SELECT_ROAD_COLOR: {
				m_SelectRoadButton.setEnabled( true );
				m_MoreButton.setEnabled( false );
				m_PreviewOKButton.setEnabled( false );
				m_PreviewFailedButton.setEnabled( false );
				m_DoneButton.setEnabled( false );
				m_ResetButton.setEnabled( true );
				
				m_UpButton.setEnabled( true );
				m_DownButton.setEnabled( true );
				m_LeftButton.setEnabled( true );
				m_RightButton.setEnabled( true );
				
				break;
			}
			// Allow the user just to say if the preview is OK for him 
			// or not
			case STEP_PREVIEW: {
				m_SelectRoadButton.setEnabled( false );
				m_MoreButton.setEnabled( false );
				m_PreviewOKButton.setEnabled( true );
				m_PreviewFailedButton.setEnabled( true );
				m_DoneButton.setEnabled( false );
				m_ResetButton.setEnabled( true );
				
				m_UpButton.setEnabled( false );
				m_DownButton.setEnabled( false );
				m_LeftButton.setEnabled( false );
				m_RightButton.setEnabled( false );
				
				break; 
			}
			// Store the recognized points
			case STEP_FINISH: {
				m_SelectRoadButton.setEnabled( false );
				m_MoreButton.setEnabled( true );
				m_PreviewOKButton.setEnabled( false );
				m_PreviewFailedButton.setEnabled( false );
				m_DoneButton.setEnabled( true );
				m_ResetButton.setEnabled( true );
				
				m_UpButton.setEnabled( true );
				m_DownButton.setEnabled( true );
				m_LeftButton.setEnabled( true );
				m_RightButton.setEnabled( true );
				
				break;
			}
		}
	}

	/**
	 * Loads a given image by its path name
	 * 
	 * @param szFileName - the path to the image
	 * 
	 * @return the loaded image
	 */	
	private BufferedImage load( String szFileName )
	{
		BufferedImage image  = null; // the returned image
		MediaTracker tracker = new MediaTracker( new Container( ) );
		Image tmp            = Toolkit.getDefaultToolkit( ).createImage( szFileName );
		tracker.addImage( tmp, 0 );
		
		try { tracker.waitForAll( ); } catch ( InterruptedException e ) {
			System.err.println( "InterruptedException: " + e.getMessage( ) );
			return null;
		}
	
		image      = new BufferedImage( tmp.getWidth( null ), tmp.getHeight( null ), BufferedImage.TYPE_INT_RGB );
		Graphics g = image.getGraphics( );
		g.drawImage( tmp, 0, 0, null );
		
		return image;
	}
	
	/**
	 * Draw whatever that has to be drawn
	 * 
	 * @param g - a reference to the applet's Graphics object
	 */
	public void paint( Graphics g ) 
	{
		super.paint( g );
		
		switch ( m_iStep ) {
			/**
			 * STEP FINISH
			 */
			case STEP_FINISH: {
				BufferedImage img = m_SourceImage;
				if ( img != null ) {
					// Get the part of the image that must be drawn. We have the center X and 
					// center Y position and  offsets - horizontal and vertical. Calculate
					// the x, y position of the left top corner
					int iImageWidth   = img.getWidth( );
					int iImageHeight  = img.getHeight( );
					int iCanvasWidth  = m_Canvas.getWidth( );
					int iCanvasHeight = m_Canvas.getHeight( );
					 
					int iCenterX 		= iImageWidth  / 2;
					int iCenterY 		= iImageHeight / 2;
					int x 				= iCenterX + m_iHOffset * 100 - iCanvasWidth / 2;
					int y 				= iCenterY + m_iVOffset * 100 - iCanvasHeight / 2;
					
					if ( x + iCanvasWidth > iImageWidth   ) { x = iImageWidth - iCanvasWidth; m_iHOffset--; }
					if ( x < 0                            ) { x = 0; } 
					
					if ( y + iCanvasHeight > iImageHeight ) { y = iImageHeight - iCanvasHeight; m_iVOffset--; }
					if ( y < 0                            ) { y = 0; }
		
					int iCroppedImageWidth 	= Math.min( iCanvasWidth, iImageWidth );
					int iCroppedImageHeight	= Math.min( iCanvasHeight, iImageHeight );
					Graphics gg = m_Canvas.getGraphics( );
					gg.drawImage( m_SourceImage, 0, 0, iCroppedImageWidth, iCroppedImageHeight, this );
				}
		
				break;
			}
			/**
			 * STEP 1 - SELECT ROAD COLOR
			 */
			case STEP_SELECT_ROAD_COLOR: {
				// Draw the pattern image that waits for road color selection
				if ( m_SourceImage == null ) { return; }
				notifyUser( "Drawing image..." );				
				BufferedImage roadsPreview = copyImage( m_SourceImage );
				ProgressBar progress = new ProgressBar( this );
				RoadRecognition rr = new RoadRecognition( m_ProgressBar );
				if ( m_RecognitionInfo != null ) {
					progress.init( m_SourceImage.getWidth( ) * m_SourceImage.getHeight( ), 1, "Road points: " );
					for ( int i = 0; i < m_SourceImage.getWidth( ); i++ ) {
						for ( int j = 0; j < m_SourceImage.getHeight( ); j++ ) {
							Color c = new Color( m_SourceImage.getRGB( i, j ) );
							if ( rr.isRoadColor( c, m_RecognitionInfo ) ) {
								roadsPreview.setRGB( i, j, Color.WHITE.getRGB( ) );
								if ( i < m_SourceImage.getWidth( ) - 1 ) {
									roadsPreview.setRGB( i + 1, j, Color.BLACK.getRGB( ) );
								}
							}
							progress.step( );
						}
					}
				}
				
				Graphics gg = m_Canvas.getGraphics( );
				gg.drawImage( roadsPreview, 0, 0, this );
				notifyUser( "Done" );
				break;
			}
			/**
			 * STEP PREVIEW
			 */
			case STEP_PREVIEW: {
				// Draw the pattern image that waits for road color selection
				Graphics gg = m_Canvas.getGraphics( );
				gg.drawImage( m_SourceImage, 0, 0, 
				              m_SourceImage.getWidth( ), m_SourceImage.getHeight( ), this );
				break;
			}
		}
	}
	
	public void repaint( )
	{
		paint( getGraphics( ) );
	}
	
	/**
	 * Copies a given image
	 * 
	 * @param srcImage - the original image
	 * 
	 * @return a copy of the original image
	 */
	private BufferedImage copyImage( BufferedImage srcImage )
	{
		int iWidth  = srcImage.getWidth( );
		int iHeight = srcImage.getHeight( );
		
		BufferedImage imageCopy = new BufferedImage( iWidth, iHeight, BufferedImage.TYPE_INT_RGB );
		
		for ( int i = 0; i < iWidth; i++ ) {
			for ( int j = 0; j < iHeight; j++ ) {
				imageCopy.setRGB( i, j, srcImage.getRGB( i, j ) );
			}
		}
		
		return imageCopy;
	}

	/**
	 * Does image preprocessing
	 * 
	 * @param srcImage - the source image
	 * 
	 * @return a copied and preprocessed version of the source image
	 */	
	private BufferedImage preprocessImage( BufferedImage srcImage )
	{
		BufferedImage image = copyImage( srcImage );
		return image;
	}
	
	/**
	 * Processes the image and extracts the points from the road graph
	 * 
	 * @param srcImage - source image
	 * @param info - recognition info object
	 * 
	 * @return returns an array of recognized road points	 
	 */	
	private RoadPoint[] processImage( MappedBufferedImage srcImage, RecognitionInfo info )
	{
		Profiler profiler  = new Profiler( );
		profiler.start( );
		
		// Get the road points first
		RoadRecognition rr   = new RoadRecognition( m_ProgressBar );
		RoadPoint[] points   = rr.getRoadCorePoints( srcImage, info );
		
		if ( points == null ) {
			notifyUser( "No roads found" );
			return null;
		}
		
		System.out.println( "Road points recognition finished: " + profiler.format( profiler.stop( ) ) );
		
		return points;		
	}		
	
	/**
	 * Draws the graph generated from the given array of points
	 * 
	 * @param points - the points that form a graph
	 * @param image - the image to use as a background (the image is not modified)
	 * 
	 * @return the visualized image
	 */
	private BufferedImage visualizeGraph( RoadPoint[] points, BufferedImage bg )
	{
		if ( points == null ) {
			System.out.println( "No points to draw" );
			return bg;
		}
		//	Build a graph from the points (just to draw a graph
		// to the user
		Profiler profiler  = new Profiler( ); profiler.start( );
		RoadRecognition rr = new RoadRecognition( m_ProgressBar );
		RoadGraph graph    = rr.buildRoadGraph( points );
		System.out.println( "Built graph from " + points.length + " points: " + profiler.format( profiler.stop( ) ) );
		profiler.start( );
	
		// Draw the graph
		ArrayList roadPoints = graph.getPoints( );
	
		BufferedImage image 	= copyImage( bg );
		Graphics2D g 			= ( Graphics2D ) image.getGraphics( );
	
		System.out.println( "Drawing " + roadPoints.size( ) + " points..." );
	
		for ( int i = 0; i < roadPoints.size( ); i++ ) {
			RoadPoint point = ( RoadPoint )roadPoints.get( i );
			g.setColor( Color.CYAN );
			g.fillRect( point.x, point.y, 1, 1 );
			/*
			if ( point.x < image.getWidth( ) - 1 && point.y < image.getHeight( ) - 1 ) {
				g.setColor( Color.CYAN );
				g.fillRect( point.x + 1, point.y + 1, 1, 1 );
			}*/
			
			Stroke stroke = new BasicStroke( 2 );
			g.setStroke( stroke );
			for ( int j = 0; j < point.connections.size( ); j++ ) {
				int iConnID = ( ( ConnectionPoint )point.connections.get( j ) ).id;
				RoadPoint conn = graph.getPointByID( iConnID );

				g.setColor( Color.CYAN );
				Line2D.Double line = new Line2D.Double( point.x, point.y, conn.x, conn.y );
				g.draw( stroke.createStrokedShape( line ) );
				
				/*
				if ( point.x < image.getWidth( ) - 2 && point.y < image.getHeight( ) - 2 &&
				     conn.x < image.getWidth( ) - 2 && conn.y < image.getHeight( ) - 2 ) {
					g.setColor( Color.CYAN );
					line = new Line2D.Double( point.x + 2, point.y + 2, conn.x + 2, conn.y + 2 );
					g.draw( stroke.createStrokedShape( line ) );
				}
				*/
			}
		}
	
		System.out.println( roadPoints.size( ) + " points are drawn: " + profiler.format( profiler.stop( ) ) );
		
		return image;
	}
}
